local util = include( "modules/util" )
local simunit = include( "sim/simunit" )
local simdefs = include( "sim/simdefs" )
local simfactory = include( "sim/simfactory" )
local inventory = include( "sim/inventory" )
local speechdefs = include( "sim/speechdefs" )
-----------------------------------------------------
-- Grenade item

local W93_snare_grenade = { ClassType = "W93_snare_grenade" }

function W93_snare_grenade:throw( throwingUnit, targetCell )
	local sim = self:getSim()
	local player = throwingUnit:getPlayerOwner()
	local x0, y0 = throwingUnit:getLocation()

    	assert( player )
	self:setPlayerOwner(player)
	sim:dispatchEvent( simdefs.EV_UNIT_THROWN, { unit = self, x=targetCell.x, y=targetCell.y } )

	if x0 ~= targetCell.x or y0 ~= targetCell.y then
		inventory.useItem( sim, throwingUnit, self )
		sim:warpUnit(self, targetCell)
	end

    	self:getTraits().throwingUnit = throwingUnit:getID()
     	self:activate( throwingUnit, targetCell )
end

function W93_snare_grenade:activate( throwingUnit, targetCell )
	assert( not self:getTraits().deployed )
	local sim = self:getSim()
	local cell = sim:getCell( throwingUnit:getLocation() )

	sim:addTrigger( simdefs.TRG_UNIT_PICKEDUP, self )
	self:getTraits().deployed = true

	if throwingUnit and targetCell and targetCell.impass > 0 then
		sim:dispatchEvent( simdefs.EV_TELEPORT, { units={self}, warpOut = true } )
		sim:warpUnit( self, cell )
		sim:dispatchEvent( simdefs.EV_TELEPORT, { units={self}, warpOut = false } )
		inventory.pickupItem( sim, throwingUnit, self )
		return
	end

	for i, cellUnit in pairs(targetCell.units) do
		if cellUnit:getTraits().isAgent and cellUnit:getPlayerOwner() == sim:getNPC() then
			self:onExplode(cell)
		end
	end
end

function W93_snare_grenade:deactivate()
	self:getTraits().deployed = nil
	local sim = self:getSim()

	if self:getTraits().explodes == nil then
		sim:removeTrigger( simdefs.TRG_UNIT_PICKEDUP, self )
	elseif self:getTraits().explodes > 0 then
		sim:removeTrigger( simdefs.TRG_START_TURN, self )
		sim:removeTrigger( simdefs.TRG_UNIT_PICKEDUP, self )
	end

	sim:dispatchEvent( simdefs.EV_UNIT_DEACTIVATE, { unit = self } )
end

function W93_snare_grenade:onExplode( targetCell )
	local sim = self:getSim()
	local x0, y0 = self:getLocation()
	local grenadeCell = sim:getCell( x0, y0 )
	local throwingUnit = sim:getUnit(self:getTraits().throwingUnit)
	local targetUnit

	for i, cellUnit in pairs(grenadeCell.units) do
		if cellUnit:getTraits().isAgent and cellUnit:getPlayerOwner() == sim:getNPC() then
			targetUnit = cellUnit
			break
		end
	end

	for i, exit in pairs(targetCell.exits) do
		if (not exit.door) and exit.cell.impass <= 0 and not sim:getQuery().checkDynamicImpass(sim, exit.cell) then
			local witnesses = {}
			for i, witness in pairs(sim:getNPC():getUnits()) do
				if witness ~= targetUnit and witness:getTraits().isGuard and sim:canUnitSeeUnit(witness, targetUnit) then
					table.insert(witnesses, witness)
				end
			end

			targetUnit:interruptMove( sim, targetUnit )
			sim:dispatchEvent( simdefs.EV_UNIT_STOP_WALKING, { unit = targetUnit } )
			sim:dispatchEvent( simdefs.EV_UNIT_APPEARED, { unitID = targetUnit:getID(), noSightingFx=true } )
			sim:dispatchEvent( simdefs.EV_TELEPORT, { units={targetUnit}, warpOut = true } )
			sim:warpUnit( targetUnit, exit.cell )
			sim:warpUnit( self, exit.cell )
			sim:dispatchEvent( simdefs.EV_UNIT_APPEARED, { unitID = targetUnit:getID(), noSightingFx=true } )
			sim:dispatchEvent( simdefs.EV_TELEPORT, { units={targetUnit}, warpOut = false } )
			sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = targetUnit } )
			sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = self } )

			if targetUnit:isValid() and (not targetUnit:isDown()) then
				targetUnit:setAlerted(true)
    	        		local x1, y1 = targetUnit:getLocation()
	            		sim:getNPC():spawnInterest(x1, y1, simdefs.SENSE_RADIO, simdefs.REASON_HUNTING, targetUnit)
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = targetUnit } )
			end

			for i, witness in pairs(witnesses) do
				witness:setAlerted(true)
	            		sim:getNPC():spawnInterest(x0, y0, simdefs.SENSE_RADIO, simdefs.REASON_HUNTING, witness)
				sim:dispatchEvent( simdefs.EV_UNIT_ALERTED, { unitID = witness:getID() } )
				sim:emitSpeech( witness, speechdefs.HUNT_NOISE)
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = witness } )
			end
			break
		end
	end
end

function W93_snare_grenade:onWarp( sim, oldcell, cell )
	if not oldcell and cell then
		sim:addTrigger( simdefs.TRG_UNIT_WARP, self )
	elseif not cell and oldcell then
		sim:removeTrigger( simdefs.TRG_UNIT_WARP, self )
	end
end

function W93_snare_grenade:onTrigger( sim, evType, evData )
	if evType == simdefs.TRG_UNIT_PICKEDUP and evData.item == self then
		self:deactivate()
	elseif evType == simdefs.TRG_UNIT_WARP and evData.unit ~= self and self:getTraits().deployed == true then
		local x0,y0 = self:getLocation()

		if evData.to_cell == sim:getCell(self:getLocation()) and evData.unit:getTraits().isAgent and evData.unit:getPlayerOwner() == sim:getNPC() then
			self:onExplode(sim:getCell(sim:getUnit(self:getTraits().throwingUnit):getLocation()))
		end
	end
end
-----------------------------------------------------
-- Interface functions

local function createGrenade( unitData, sim )
	return simunit.createUnit( unitData, sim, W93_snare_grenade )
end

simfactory.register( createGrenade )

return
{
	createGrenade = createGrenade,
}


