local abilityutil = include("sim/abilities/abilityutil")
local inventory = include("sim/inventory")
local mathutil = include("modules/mathutil")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local util = include("modules/util")

-- TODO Tooltips for Knockback

local _M = {
    name = STRINGS.HEATSIGNATURE.ABILITIES.FACEBREAK,
    profile_icon = "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png",
    usesAction = true
}

function _M:getName(sim, unit, userUnit)
    return self.name
end

function _M:createToolTip(sim, unit, targetCell)
    return abilityutil.formatToolTip(
        STRINGS.HEATSIGNATURE.ABILITIES.FACEBREAK,
        STRINGS.HEATSIGNATURE.ABILITIES.FACEBREAK_DESC,
        1
    )
end

local function validTargetCellFn(sim, unit, targetCell)
    local x0, y0 = unit:getLocation()
    local x1, y1 = targetCell.x, targetCell.y
    if x0 == x1 and y0 == y1 then
        return false
    end

    if targetCell.impass > 0 then
        return false
    end

    if mathutil.dist2d(x0, y0, x1, y1) < 2 then
        return false, STRINGS.HEATSIGNATURE.REASON.BAD_TARGET_TOO_CLOSE
    end

    for i, targetUnit in ipairs(targetCell.units) do
        if simquery.isAgent(targetUnit) and not targetUnit:isDown() and not targetUnit:getTraits().isDrone then
            return true
        end
    end

    return false, STRINGS.HEATSIGNATURE.REASON.BAD_TARGET_NO_FACE
end

function _M:acquireTargets(targets, game, sim, itemUnit, unit)
    if not self:canUseAbility(sim, itemUnit, unit) then
        return nil
    end
    local lineTarget = targets.lineKnockbackTarget(game, sim, unit)
    lineTarget.maxDist = unit:getMP() or 8
    lineTarget.range = -1
    lineTarget.ignoreDynamicImpass = false
    lineTarget.ignoreStaticImpass = false
    lineTarget.validTargetCellFn = validTargetCellFn
    return lineTarget
end

function _M:canUseAbility(sim, itemUnit, unit, targetCell)
    if unit:getTraits().movingBody then
        return false, STRINGS.UI.REASON.DROP_BODY_TO_USE
    end

    if itemUnit:getTraits().cooldown and itemUnit:getTraits().cooldown > 0 then
        return false, util.sformat(STRINGS.UI.REASON.COOLDOWN, itemUnit:getTraits().cooldown)
    end

    if unit:getMP() < 1 then
        return false, string.format(STRINGS.UI.REASON.REQUIRES_AP, 1)
    end

    return true
end


function _M:executeAbility(sim, itemUnit, userUnit, targetCoords)
    local x0, y0 = userUnit:getLocation()
    local x1, y1 = unpack(targetCoords)

    userUnit:setDisguise(false)

    local facing = simquery.getDirectionFromDelta(x1 - x0, y1 - y0)
    simquery.suggestAgentFacing(userUnit, facing)
    if userUnit:getBrain() then
        sim:refreshUnitLOS(userUnit)
        sim:processReactions(userUnit)
    end

    if userUnit:isValid() and not userUnit:getTraits().interrupted then
        local line = mathutil.getCellsBetween(x1, y1, x0, y0)
        local amtTravelledCells = #line / 2

        userUnit:turnToFace(x1, y1)
        sim:dispatchEvent(simdefs.EV_UNIT_PLAY_ANIM, {unit = userUnit, anim = "pin_pre"})
        sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = userUnit})

        if #line > 0 then
            userUnit:changeKanim("leap_saderfiasco") -- Activate leap/dash anim
            sim:dispatchEvent(simdefs.EV_UNIT_START_WALKING, {unit = userUnit, reverse = false})

            -- Stop just before the actual targetCell.
            sim:warpUnit(userUnit, sim:getCell(line[1], line[2]))

            userUnit:changeKanim(nil) -- Deactivate leap/dash anim
        end
        sim:dispatchEvent(simdefs.EV_UNIT_STOP_WALKING, {unit = userUnit}) -- Does regular "grind to a halt" anim

        sim:startTrackerQueue(true)
        sim:startDaemonQueue()
        sim:dispatchEvent(simdefs.EV_KO_GROUP, true)

        local hitUnits = dashabilityutil.applyKnockback(sim, x1, y1, x1 * 2 - x0, y1 * 2 - y0)

        if hitUnits[1] then
            local unit = hitUnits[1].unit
            if unit:isValid() and unit:getTraits().canKO then
                local damageMod = amtTravelledCells - 1  -- assumes base damage 1
                itemUnit:getTraits().damage = itemUnit:getTraits().damage + damageMod
                local meleeDamage = simquery.calculateMeleeDamage(sim, itemUnit, unit)
                itemUnit:getTraits().damage = itemUnit:getTraits().damage - damageMod
                if meleeDamage > 0 then
                    unit:setKO(sim, meleeDamage)
                end
            end
        end

        for _, hit in pairs(hitUnits) do
            if hit.unit:isValid() and not hit.unit:isDown() then
                -- theoretically this also needs lastHit and lastAttack like in simengine, but practically those are irrelevant for all possible cases here
                sim:triggerEvent( simdefs.TRG_UNIT_HIT, {
                    targetUnit = hit.unit,
                    sourceUnit = userUnit,
                    x = x1,
                    y = y1,
                    noTargetAlert = false
                })
                sim:dispatchEvent(simdefs.EV_NONBLOCKING_EVENT, {
                    unit = hit.unit,
                    sourceUnit = userUnit,
                    actualEventType = simdefs.EV_UNIT_HIT
                })
            end
            sim:processReactions(hit.unit)
        end

        sim:dispatchEvent(simdefs.EV_KO_GROUP, false)
        sim:startTrackerQueue(false)
        sim:processDaemonQueue()

        -- Did we cross a deadly lazer?
        dashabilityutil.processLasersOnCells(sim, userUnit, line)

        local newFacing = (math.floor(userUnit:getFacing() * 0.5) * 2)
        userUnit:setFacing(newFacing)

        sim:dispatchEvent(simdefs.EV_UNIT_APPEARED, {unitID = userUnit:getID(), noSightingFx = true})
        sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = userUnit})

        userUnit:stopHacking(sim)

        -- Guard vision along the path glimpses/tracks the agent.
        dashabilityutil.guardTrackMovementToCell(sim, userUnit, x0, y0, x1, y1)

        inventory.useItem(sim, userUnit, itemUnit)
        userUnit:setInvisible(false)
        userUnit:useMP(amtTravelledCells, sim)
    end
end

return _M
