local util = include("modules/util")
local mathutil = include("modules/mathutil")
local simdefs = include("sim/simdefs")
local inventory = include("sim/inventory")
local abilityutil = include("sim/abilities/abilityutil")

local _M = {
    name = STRINGS.HEATSIGNATURE.ABILITIES.GLITCHTRAP_DEPLOY,
    profile_icon = "gui/icons/action_icons/Action_icon_Small/icon-item_hijack_small.png",
    usesAction = true,
    alwaysShow = true
}

function _M:getName(sim, unit, userUnit)
    return self.name
end

function _M:createToolTip(sim, unit, targetCell)
    return abilityutil.formatToolTip(
        STRINGS.HEATSIGNATURE.ABILITIES.GLITCHTRAP_DEPLOY,
        STRINGS.HEATSIGNATURE.ABILITIES.GLITCHTRAP_DEPLOY_DESC,
        1
    )
end

local function isValidTargetLoc(sim, x, y, data)
    local targetCell = sim:getCell(x, y)
    if not targetCell or data.player:getLastKnownCell(sim, x, y) == nil then
        return false, nil
    end
    if targetCell.impass > 0 then
        return false, STRINGS.HEATSIGNATURE.REASON.BAD_TARGET_IMPASSABLE
    end
    if not data.ignoreLOS and simquery.isCellWatched(sim, data.player, x, y) == simdefs.CELL_HIDDEN then
        return false, STRINGS.UI.REASON.BLOCKED
    end
    if data.range and mathutil.dist2d(x, y, unpack(data.origin)) > data.range then
        return false, STRINGS.UI.REASON.BAD_TARGET_RANGE
    end
    return true, nil
end

function _M:acquireTargets(targets, game, sim, itemUnit, userUnit)
    if not self:canUseAbility(sim, itemUnit, userUnit) then
        return nil
    end
    local x, y = itemUnit:getLocation()
    if x == nil then
        x, y = userUnit:getLocation()
    end
    return targets.areaTargetDelegatable(
        game,
        0,
        sim,
        isValidTargetLoc,
        {
            player = itemUnit:getPlayerOwner(),
            origin = {x, y},
            range = itemUnit:getTraits().range,
            ignoreLOS = itemUnit:getTraits().targeting_ignoreLOS
        }
    )
end

local function shouldTrash(itemUnit)
    return itemUnit:getTraits().should_trash_after_warp or (itemUnit:getTraits().usesLeft or 1) <= 0
end

function _M:canUseAbility(sim, itemUnit, userUnit, targetCell)
    if shouldTrash(itemUnit) then
        return false, nil
    end

    if userUnit and userUnit:getTraits().movingBody then
        return false, STRINGS.UI.REASON.DROP_BODY_TO_USE
    end

    if (itemUnit:getTraits().cooldown or 0) > 0 then
        return false, util.sformat(STRINGS.UI.REASON.COOLDOWN, itemUnit:getTraits().cooldown)
    end

    if itemUnit:getTraits().usesCharges and itemUnit:getTraits().charges <= 0 then
        return false, util.sformat(STRINGS.UI.REASON.CHARGES, itemUnit:getTraits().charges)
    end

    return true
end

function _M:deactivate(sim, itemUnit)
    sim:removeTrigger(simdefs.TRG_UNIT_PICKEDUP, self)
    sim:removeTrigger(simdefs.TRG_START_TURN, self)
    sim:removeTrigger(simdefs.TRG_UNIT_WARP, self)

    if itemUnit:getTraits().mainframe_item then
        itemUnit:getTraits().mainframe_status = "off"
    end

    itemUnit:getTraits().deployed = false
    itemUnit:getTraits().hidden_at_distance = false

    sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = itemUnit})
end

function _M:activate(sim, itemUnit)
    sim:addTrigger(simdefs.TRG_UNIT_PICKEDUP, self, itemUnit)
    sim:addTrigger(simdefs.TRG_START_TURN, self, itemUnit)
    sim:addTrigger(simdefs.TRG_UNIT_WARP, self, itemUnit)

    if itemUnit:getTraits().mainframe_item then
        itemUnit:getTraits().mainframe_status = "active"
    end

    itemUnit:getTraits().deployed = true
    itemUnit:getTraits().hidden_at_distance = true

    sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = itemUnit})

    self:applyTrap(sim, itemUnit, sim:getCell(itemUnit:getLocation()).units)
end

function _M:applyTrap(sim, itemUnit, targetUnits)
    assert(type(targetUnits) == "table")
    assert(itemUnit:getTraits().glitchtrap_targetCell ~= nil)

    local x0, y0 = itemUnit:getLocation()
    local x1, y1 = unpack(itemUnit:getTraits().glitchtrap_targetCell)
    local targetCell = sim:getCell(x1, y1)

    if not self:canUseAbility(sim, itemUnit) or not targetCell then
        return
    end

    local validTargets = {}

    for _, u in pairs(targetUnits) do
        if u:isValid() and u:getPlayerOwner() ~= itemUnit:getPlayerOwner() and u:getTraits().isAgent then
            table.insert(validTargets, u)
        end
    end

    if #validTargets < 1 then
        return
    end

    -- TODO what to do if already occupied and a dynamicimpass warps in?

    for _, u in pairs(validTargets) do
        sim:dispatchEvent(simdefs.EV_UNIT_STOP_WALKING, {unit = u})
        -- u:interruptMove( sim, itemUnit )
    end

    sim:dispatchEvent(simdefs.EV_PLAY_SOUND, {sound = "SpySociety/Actions/Teleport_strike_inbound", x = x0, y = y0})
    local _, firstUnit = next(validTargets) -- only spawn fx once
    sim:dispatchEvent(
        simdefs.EV_UNIT_ADD_FX,
        {
            unit = firstUnit,
            kanim = "fx/glitch_fx",
            anim = "in",
            symbol = "effect",
            x = x0,
            y = y0
        }
    )

    for _, u in pairs(validTargets) do
        sim:warpUnit(u, targetCell)
        u:getTraits().interrupted = true
        if u:getBrain() then
            u:getBrain():spawnInterest(x1, y1, simdefs.SENSE_HIT, simdefs.REASON_FOUNDOBJECT, u)
        end
    end
    
    sim:dispatchEvent(
        simdefs.EV_UNIT_ADD_FX,
        {
            unit = firstUnit,
            kanim = "fx/glitch_fx",
            anim = "out",
            symbol = "effect",
            x = x1,
            y = y1
        }
    )

    sim:processReactions(itemUnit)

    inventory.useItem(sim, itemUnit, itemUnit)  -- does not remove disposables (good, it'd crash)
    itemUnit:getTraits().should_trash_after_warp = itemUnit:getTraits().disposable
end

function _M:onWarp(sim, evData, itemUnit)
    if evData.unit == itemUnit and evData.to_cell == nil then
        self:deactivate(sim, itemUnit)
        return
    end

    if itemUnit:getTraits().mainframe_status == "inactive" then
        return
    end

    local x0, y0 = itemUnit:getLocation()
    local x1, y1 = evData.unit:getLocation()
    if x0 == x1 and y0 == y1 then
        self:applyTrap(sim, itemUnit, {evData.unit})
    end

    local cell = sim:getCell(x0, y0)
    if evData.to_cell == cell or evData.from_cell == cell then
        if evData.unit:getTraits().isAgent then
            sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = itemUnit}) -- not strictly necessary but unit ghosts can be weird
        end
    end
end

function _M:onStartTurn(sim, evData, itemUnit)
    -- cool down when not held
    if (itemUnit:getTraits().cooldown or 0) > 0 and itemUnit:getUnitOwner() == nil then
        itemUnit:getTraits().cooldown = itemUnit:getTraits().cooldown - 1
    end

    if shouldTrash(itemUnit) then
        sim:warpUnit(itemUnit)
        sim:despawnUnit(itemUnit)  -- patch disposable / usesLeft outside of warp
    end
end

function _M:onSelfPickedUp(sim, evData, itemUnit)
    itemUnit:setPlayerOwner(nil)
    itemUnit:getTraits().glitchtrap_targetCell = nil
    self:deactivate(sim, itemUnit)
end

function _M:onTrigger(sim, evType, evData, itemUnit)
    if evType == simdefs.TRG_UNIT_PICKEDUP and evData.item == itemUnit then
        self:onSelfPickedUp(sim, evData, itemUnit)
    elseif evType == simdefs.TRG_UNIT_WARP then
        self:onWarp(sim, evData, itemUnit)
    elseif evType == simdefs.TRG_START_TURN and evData:isPC() then
        self:onStartTurn(sim, evData, itemUnit)
    end
end

function _M:executeAbility(sim, itemUnit, userUnit, targetCell)
    local player = userUnit:getPlayerOwner()
    local x0, y0 = userUnit:getLocation()
    local x1, y1 = unpack(targetCell)

    inventory.dropItem(sim, userUnit, itemUnit)
    sim:dispatchEvent(simdefs.EV_UNIT_PICKUP, {unitID = userUnit:getID()})

    itemUnit:setPlayerOwner(player)
    itemUnit:getTraits().glitchtrap_targetCell = targetCell
    self:activate(sim, itemUnit)

    sim:processReactions(userUnit)
end

return _M
