local util = include( "modules/util" )
local simdefs = include("sim/simdefs")
local abilityutil = include( "sim/abilities/abilityutil" )
local simquery = include("sim/simquery")
local store = include( "sim/units/store" )
local simability = include( "sim/simability" )
local mainframe_abilities = include( "sim/abilities/mainframe_abilities" )
local simfactory = include( "sim/simfactory" )

local SELECTABLE_PROGRAM_COUNT = {2, 2, 3, 3, 4}

local function createProgramUnitData( abilityName, sim )

    local abilityDef = mainframe_abilities[ abilityName ]

	assert( abilityDef )
	assert( sim )

	if abilityDef.abilityOverride then
    		if sim:isVersion(abilityDef.abilityOverride[2]) then
			abilityDef =  mainframe_abilities[abilityDef.abilityOverride[1]]
		end
	end

	return
	{
		type = "simunit",
		name = abilityDef.name,
		program = true,
		traits =
		{
			mainframe_program = abilityName,
		},
		onTooltip = function( tooltip, unit )

			-- FIXME: should delegate to ability, but the parameters are different! (No hud access here)
			tooltip:addLine( abilityDef.name, util.sformat( STRINGS.PROPS.STORE_PROGRAM_TOOLTIP, abilityDef:getCpuCost() ))
			if abilityDef.maxCooldown and abilityDef.maxCooldown > 0  then
				tooltip:addLine( util.sformat( STRINGS.PROGRAMS.COOLDOWN, abilityDef.maxCooldown )  )
			end
			tooltip:addAbility( abilityDef.shortdesc, abilityDef.desc, "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png" )

		end,

		profile_icon_100 = abilityDef.icon_100,
		value = abilityDef.value
	}
end

local function showDialog( sim, headerTxt, bodyTxt, options, t, result )
	sim._choiceCount = sim._choiceCount + 1
	local choice = sim._choices[ sim._choiceCount ]
	options = options or { "OK" }

	if not choice then
		-- Choice has not yet been made.
		sim:dispatchChoiceEvent( "MAA_EV_PROGRAM_DIALOG", util.extend( { headerTxt = headerTxt, bodyTxt = bodyTxt, options = options }) ( t ), true )
		choice = sim._choices[ sim._choiceCount ]
	end

	return choice
end

local maa_craft_program =
{
	name = STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.NAME,
	profile_icon = "gui/icons/item_icons/items_icon_small/icon-item_generic_head_small2.png",
	proxy = true,

	createToolTip = function( self, sim, unit )
		return abilityutil.formatToolTip( STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.NAME, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.DESC )
	end,

	getName = function( self, sim, unit )
		return self.name
	end,

	isTarget = function( self, sim, userUnit, targetUnit )
		if not targetUnit:getTraits().mainframe_console then
			return false
		end

		if targetUnit:getTraits().mainframe_status ~= "active" then
			return false
		end

		return true
	end,

	acquireTargets = function( self, targets, game, sim, abilityOwner, abilityUser )
		local cell = sim:getCell( abilityUser:getLocation() )
		local units = {}
		for dir, exit in pairs(cell.exits) do
			for _, cellUnit in ipairs( exit.cell.units ) do
				if self:isTarget( sim, abilityUser, cellUnit ) then
					if simquery.canUnitReach( sim, abilityUser, exit.cell.x, exit.cell.y ) then
						table.insert( units, cellUnit )
					end
				end
			end
		end

		return targets.unitTarget( game, units, self, abilityOwner, abilityUser )
	end,

	canUseAbility = function( self, sim, unit, userUnit, targetUnitID )
		local player = sim:getPC()
		if not unit:getTraits().installed then
			return false
		end
		if unit:getTraits().movingBody then
				return false, STRINGS.UI.REASON.DROP_BODY_TO_USE
		end
		if unit:getTraits().ammo and unit:getTraits().ammo <= 0 then
			return false, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.NO_CHARGE
		end

		local targetUnit = sim:getUnit( targetUnitID )
		if targetUnit then
			local dir = targetUnit:getFacing()
			local x0, y0 = targetUnit:getLocation()
			local x1, y1 = simquery.getDeltaFromDirection(dir)
			local consoleFront = sim:getCell( x0 + x1, y0 + y1 )
			if consoleFront ~= sim:getCell( userUnit:getLocation() ) then
				return false, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.DIRECTION
			end
		end

		return true
	end,

	executeAbility = function( self, sim, unit, userUnit, targetUnitID )
		local targetUnit = sim:getUnit( targetUnitID )

		if userUnit:getTraits().data_hacking then
			local target = sim:getUnit(userUnit:getTraits().data_hacking)
			if target then
				target:getTraits().hacker = nil
			end
			userUnit:getTraits().data_hacking = nil
			userUnit:getSounds().spot = nil
			sim:dispatchEvent( simdefs.EV_UNIT_TINKER_END, { unit = userUnit } )
			sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit })
		else
			local x0,y0 = userUnit:getLocation()
			local x1,y1 = targetUnit:getLocation()
			local tempfacing = simquery.getDirectionFromDelta( x1 - x0, y1 - y0 )
			userUnit:setFacing(tempfacing)
			sim:dispatchEvent( simdefs.EV_UNIT_USECOMP, { unitID = userUnit:getID(), facing = tempfacing, sound = "SpySociety/Actions/monst3r_jackin" , soundFrame = 16, useTinkerMonst3r=true } )


			local player = sim:getPC()
			local program_options = {}
			table.insert( program_options, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.MODAL_CANCEL )

			for i=SELECTABLE_PROGRAM_COUNT[ (userUnit:getSkillLevel( "hacking" ) or 0) ], 1, -1 do
				if unit.programs[i] and unit.programs[i]:getTraits().mainframe_program and not player:hasMainframeAbility( unit.programs[i]:getTraits().mainframe_program ) then
					table.insert( program_options, {unit:getID(), i} )
				end
			end

			local option = showDialog( sim, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.MODAL_TITLE, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.CRAFT_PROGRAM.MODAL_TEXT, program_options )

			if option > 1 then
				self.add_program = unit.programs[ program_options[option][2] ]:getTraits().mainframe_program

				userUnit:getTraits().data_hacking = unit:getID()
				userUnit:getTraits().data_hacking_spotsound = true
				unit:getTraits().progress = 0
				userUnit:getSounds().spot = "SpySociety/Actions/monst3r_hacking"
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit })
			else
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit })
				sim:dispatchEvent( simdefs.EV_UNIT_TINKER_END, { unit = userUnit } )
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit })
			end
		end

	end,

	onSpawnAbility = function( self, sim, unit )
		unit:getTraits().ammo = unit:getTraits().maxAmmo
		sim:addTrigger( simdefs.TRG_START_TURN, self, unit )

		local player = unit:getPlayerOwner()
		local virusUser = unit:getUnitOwner()

		local progList = util.tcopy( store.STORE_ITEM.progList )
		progList:addList( store.STORE_ITEM.progList_17_5 )
		local nobreakList = util.tcopy( store.STORE_ITEM.noBreakerProgList )
		nobreakList:addList( store.STORE_ITEM.noBreakerProgList_17_5 )

		local progList_unweighted = {} -- I don't know how to remove elements out of a weighted list without making a big mess
		for i = 1, #progList, 2 do
			progList_unweighted[ progList[i+1] ] = progList[i]
		end
		for i = 1, #nobreakList, 2 do
			progList_unweighted[ nobreakList[i+1] ] = nobreakList[i]
		end

		for program, weight in pairs( progList_unweighted ) do
			if player:hasMainframeAbility( program ) then
				progList_unweighted[program] = nil
			end
		end

		progList = util.weighted_list( progList_unweighted )
		unit.programs = {}
		for i=1, 4, 1 do
			local w = sim:nextRand(1, progList:getTotalWeight() )
			local programName = progList:removeChoice( w )
			local programData = createProgramUnitData( programName, sim )
			local programUnit = simfactory.createUnit( programData, sim )
			table.insert( unit.programs, programUnit )
		end
	end,

	onDespawnAbility = function( self, sim, unit )
		sim:removeTrigger( simdefs.TRG_START_TURN, self, unit )

		if self.program then
			local player = sim:getPC()
			player:removeAbility(sim, self.program )
		end
	end,

	onTrigger = function( self, sim, evType, evData, unit )
		local player = sim:getPC()
		local userUnit = unit:getUnitOwner()
		if evType == simdefs.TRG_START_TURN then
			if evData:isPC() then
				if userUnit:getTraits().data_hacking and userUnit:getTraits().data_hacking == unit:getID() and unit:getTraits().progress and unit:getTraits().progress >= (unit:getTraits().maa_hacking_turns or 1) then
					if self.add_program then
						local ability = simability.create( self.add_program )
						table.insert( player._mainframeAbilities, ability )
						ability.maa_temporary_program = true
						--special case for abacus
						local start_pwr_before = player:getTraits().extraStartingPWR or 0
						ability:spawnAbility( sim, player )
						if (player:getTraits().extraStartingPWR  or 0) > start_pwr_before then
							player:addCPUs( (player:getTraits().extraStartingPWR  or 0) - start_pwr_before )
						end

						unit:getTraits().ammo = unit:getTraits().ammo - 1

						self.program = ability
						self.add_program = nil

						local target = sim:getUnit(userUnit:getTraits().data_hacking)
						if target then
							target:getTraits().hacker = nil
						end
						userUnit:getTraits().data_hacking = nil
						userUnit:getSounds().spot = nil
						sim:dispatchEvent( simdefs.EV_UNIT_TINKER_END, { unit = userUnit } )
						sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit })
					end
				end
			end
		end
	end,
}

return maa_craft_program
