local util = include( "modules/util" )
local simdefs = include("sim/simdefs")
local abilityutil = include( "sim/abilities/abilityutil" )
local array = include( "modules/array" )
local simquery = include("sim/simquery")

local function enterMainframe( ability, sim, abilityOwner, userUnit, target )
	--ko the unit and flag it for inside the mainframe
	userUnit:setKO( sim, 2 )
	userUnit:getTraits().maa_inside_mainframe = true

	--listen to the unit getting warped or revived, so we can cancel the connection if so
	sim:addTrigger( simdefs.TRG_UNIT_WARP, ability, abilityOwner )
	sim:addTrigger( simdefs.TRG_UNIT_KO, ability, abilityOwner )

	--no inventory while in the mainframe (so we can set all proxy abilities on a mainframe device as useable but can't actually buy/take items)
	ability.oldInventory = userUnit:getTraits().inventoryMaxSize
	ability.oldAugment = userUnit:getTraits().augmentMaxSize
	userUnit:getTraits().augmentMaxSize = nil
	userUnit:getTraits().inventoryMaxSize = nil

	--create an eyeball, so we can see the units own tile. this is necessary, since a koed agent you don't have vision over is considered a ghost.
	local eyeball = include( "sim/units/eyeball" )
	ability.eyeballUnit = eyeball.createEyeball( sim )
	ability.eyeballUnit:setPlayerOwner( userUnit:getPlayerOwner() )
	ability.eyeballUnit:getTraits().LOSarc = math.pi * 2
	ability.eyeballUnit:getTraits().LOSrange = 1
	local cell = sim:getCell( userUnit:getLocation() )
	sim:spawnUnit( ability.eyeballUnit )
	sim:warpUnit( ability.eyeballUnit, cell)

	--flag the used console, so we only show the abilities on that one
	local targetUnit = sim:getUnit( target )
	targetUnit:getTraits().maa_mainframe_entrance = true
	ability.enteredConsole = targetUnit
end

local function exitMainframe( ability, sim, abilityOwner, userUnit, target )
	--only run if the user is still in the mainframe
	if userUnit:getTraits().maa_inside_mainframe then
		--remove the inside mainframe flag and then check if the units cell is free an we can remove the KO
		userUnit:getTraits().maa_inside_mainframe = nil
		if not simquery.isUnitCellFull( sim, userUnit ) and not simquery.isUnitDragged( sim, userUnit ) then
			userUnit:setKO( sim, nil )
		end
		sim:removeTrigger( simdefs.TRG_UNIT_WARP, ability, abilityOwner )
		sim:removeTrigger( simdefs.TRG_UNIT_KO, ability, abilityOwner )

		userUnit:getTraits().inventoryMaxSize = ability.oldInventory
		userUnit:getTraits().augmentMaxSize = ability.oldAugment

		if ability.eyeballUnit then
			sim:warpUnit( ability.eyeballUnit )
			sim:despawnUnit( ability.eyeballUnit )
			ability.eyeballUnit = nil
		end

		ability.enteredConsole:getTraits().maa_mainframe_entrance = nil
		ability.enteredConsole = nil

		--remove control over controlled device
		if userUnit:getTraits().mma_controled_device then
			--drones have a specific way to lose control
			if userUnit:getTraits().mma_controled_device.loseControl then
				--grab the ice
				local ice_before = userUnit:getTraits().mma_controled_device:getTraits().mainframe_ice
				
				userUnit:getTraits().mma_controled_device:loseControl( sim )
				
				--reset the ice
				userUnit:getTraits().mma_controled_device:getTraits().mainframe_ice = ice_before
				--skip drone reboot
				userUnit:getTraits().mma_controled_device:setKO( sim, nil )
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = userUnit:getTraits().mma_controled_device } )
			else
				userUnit:getTraits().mma_controled_device:takeControl( sim:getNPC() )
			end
			sim:getPC():glimpseUnit( sim, userUnit:getTraits().mma_controled_device:getID() )
			userUnit:getTraits().mma_controled_device = nil
		end

		userUnit:getTraits().maa_turns_inside_mainframe = nil
	end
end

local function applyPenalty( ability, sim, unit, userUnit )
	--apply penalties for being inside the mainframe
	--power penalty
	if unit:getTraits().pwr_penalty_per_turn and unit:getTraits().pwr_penalty_per_turn > 0 then
		local player = unit:getPlayerOwner()
		local cpus = unit:getTraits().pwr_penalty_per_turn
		if player:getCpus() < cpus then
			exitMainframe( ability, sim, unit, userUnit )
		else
			player:addCPUs( -cpus )
			local x0, y0 = userUnit:getLocation()
			sim:dispatchEvent( simdefs.EV_UNIT_FLY_TXT, {txt=util.sformat(STRINGS.UI.FLY_TXT.MINUS_PWR, cpus), x=x0,y=y0, color={r=163/255,g=243/255,b=248/255,a=1},} )
		end
	end
	--mp penalty
	if unit:getTraits().maxMP_penalty_per_turn and unit:getTraits().maxMP_penalty_per_turn > 0 then
		if userUnit:getMPMax() <= unit:getTraits().maxMP_penalty_threshold then
			--force exit mainframe
			exitMainframe( ability, sim, unit, userUnit )
		else
			userUnit:addMPMax( -unit:getTraits().maxMP_penalty_per_turn )
			userUnit:addMP( -unit:getTraits().maxMP_penalty_per_turn )
			if not ability.mpPenalties then
				ability.mpPenalties = 0
			end
			ability.mpPenalties = ability.mpPenalties + unit:getTraits().maxMP_penalty_per_turn
			local x0, y0 = userUnit:getLocation()
			sim:dispatchEvent( simdefs.EV_UNIT_FLOAT_TXT, {txt=util.sformat( STRINGS.MORE_ARCHIVE_AGENTS.UI.FLY_TXT.MP_MPMAX_PENALTY, unit:getTraits().maxMP_penalty_per_turn ), x=x0, y=y0, color={r=163/255,g=0,b=0,a=1}} )
		end
	end
end

local maa_mainframe_possession =
{
	name = STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.NAME,
	profile_icon = "gui/icons/item_icons/items_icon_small/icon-item_oravirus_small.png",

	createToolTip = function( self, sim, unit )
		local userUnit = unit:getUnitOwner()
		if userUnit:getTraits().maa_inside_mainframe then
			return abilityutil.formatToolTip( STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.EXIT_NAME, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.EXIT_DESC, 0 )
		end
		return abilityutil.formatToolTip( STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.ENTER_NAME, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.ENTER_DESC, 0 )
	end,

	proxy = true,
	maa_mp_ability = true,

	getName = function( self, sim, unit )
		return self.name
	end,

	isTarget = function( self, sim, userUnit, targetUnit )
		if userUnit:getTraits().maa_inside_mainframe then
			if targetUnit:getTraits().maa_mainframe_entrance then
				return true
			end
		else
			if not targetUnit:getTraits().mainframe_console then
				return false
			end

			if targetUnit:getTraits().mainframe_status ~= "active" then
				return false
			end

			return true
		end
	end,

	acquireTargets = function( self, targets, game, sim, abilityOwner, abilityUser )
		local cell = sim:getCell( abilityUser:getLocation() )
		local units = {}
		for dir, exit in pairs(cell.exits) do
			for _, cellUnit in ipairs( exit.cell.units ) do
				if self:isTarget( sim, abilityUser, cellUnit ) then
					if simquery.canUnitReach( sim, abilityUser, exit.cell.x, exit.cell.y ) then
						table.insert( units, cellUnit )
					end
				end
			end
		end

		return targets.unitTarget( game, units, self, abilityOwner, abilityUser )
	end,

	canUseAbility = function( self, sim, unit, userUnit )
		if not unit:getTraits().installed then
			return false
		end

		if not userUnit:getTraits().maa_inside_mainframe then
			if unit:getTraits().cd_mp and unit:getTraits().cd_mp > 0 then
				return false, util.sformat(STRINGS.UI.REASON.COOLDOWN, unit:getTraits().cd_mp)
			end

			if unit:getTraits().maxMP_penalty_threshold and userUnit:getMPMax() <= unit:getTraits().maxMP_penalty_threshold + 1 then
				return false, STRINGS.MORE_ARCHIVE_AGENTS.ABILITIES.POSSESS_MAINFRAME.FATIGUED
			end
		end

		return true
	end,

	executeAbility = function( self, sim, unit, userUnit, target )
		if userUnit:getTraits().maa_inside_mainframe then
			exitMainframe( self, sim, unit, userUnit, target )
		else
			enterMainframe( self, sim, unit, userUnit, target )
			applyPenalty( self, sim, unit, userUnit )
			if unit:getTraits().cdMax_mp then
				unit:getTraits().cd_mp = unit:getTraits().cdMax_mp
			end
		end
	end,

	onSpawnAbility = function( self, sim, unit )
		sim:addTrigger( simdefs.TRG_START_TURN, self, unit )
	end,

	onDespawnAbility = function( self, sim, unit )
		sim:removeTrigger( simdefs.TRG_START_TURN, self, unit )
	end,

	onTrigger = function( self, sim, evType, evData, unit )
		local userUnit = unit:getUnitOwner()
		if evType == simdefs.TRG_START_TURN then
			if evData:isPC() then
				if userUnit:getTraits().maa_inside_mainframe then
					--refresh the ko
					userUnit:setKO( sim, 2 )
					if not userUnit:getTraits().maa_turns_inside_mainframe then
						userUnit:getTraits().maa_turns_inside_mainframe = 0
					end
					userUnit:getTraits().maa_turns_inside_mainframe = userUnit:getTraits().maa_turns_inside_mainframe + 1

					applyPenalty( self, sim, unit, userUnit )
				else
					--regenerate penalties
					if self.mpPenalties and self.mpPenalties > 0 and unit:getTraits().maxMP_regeneration_per_turn then
						local count = 0
						for i=1, unit:getTraits().maxMP_regeneration_per_turn, 1 do
							userUnit:addMPMax( 1 )
							userUnit:addMP( 1 )
							self.mpPenalties = self.mpPenalties - 1
							count = count + 1
							if self.mpPenalties <= 0 then
								break
							end
						end
						if count > 0 then
							local x0, y0 = userUnit:getLocation()
							self._sim:dispatchEvent( simdefs.EV_UNIT_FLOAT_TXT, {txt=util.sformat( STRINGS.MORE_ARCHIVE_AGENTS.UI.FLY_TXT.MP_MPMAX_REGENERATION, count ), x=x0, y=y0, color={r=1,g=1,b=41/255,a=1}} )
						end
					end
				end
				if unit:getTraits().cd_advance then
					unit:getTraits().cd_advance = math.max( unit:getTraits().cd_advance - 1, 0 )
				end
				if unit:getTraits().cd_mp then
					unit:getTraits().cd_mp = math.max( unit:getTraits().cd_mp - 1, 0 )
				end
				if unit:getTraits().cd_control then
					unit:getTraits().cd_control = math.max( unit:getTraits().cd_control - 1, 0 )
				end
			else
				--keep drones under control
				if userUnit:getTraits().maa_inside_mainframe then
					if userUnit:getTraits().mma_controled_device and userUnit:getTraits().mma_controled_device:getTraits().controlTimer then
						userUnit:getTraits().mma_controled_device:getTraits().controlTimer = userUnit:getTraits().mma_controled_device:getTraits().controlTimer + 1
					end
				end
			end
		elseif evType == simdefs.TRG_UNIT_WARP then
			if evData.unit == userUnit then
				exitMainframe( self, sim, unit, userUnit )
			end
		elseif evType == simdefs.TRG_UNIT_KO then
			if evData.unit == userUnit and evData.ticks == nil then
				exitMainframe( self, sim, unit, userUnit )
			end
		end
	end,
}

return maa_mainframe_possession
