local array = include("modules/array")
local abilityutil = include("sim/abilities/abilityutil")
local inventory = include("sim/inventory")
local mathutil = include("modules/mathutil")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local util = include("modules/util")

local _M = {
    name = STRINGS.HEATSIGNATURE.ABILITIES.SWAPPER,
    profile_icon = "gui/icons/item_icons/items_icon_small/icon-item_swapper_small.png"
}

--TODO make this not usable on self or without target

function _M:getName(sim, unit, userUnit)
    return self.name
end

function _M:createToolTip(sim, unit, targetCell)
    return abilityutil.formatToolTip(
        STRINGS.HEATSIGNATURE.ABILITIES.SWAPPER,
        STRINGS.HEATSIGNATURE.ABILITIES.SWAPPER_DESC,
        1
    )
end

function _M:isTarget(targetUnit)
    return simquery.isAgent(targetUnit) and not targetUnit:isDown() and not targetUnit:getTraits().isDrone and
        targetUnit:getUnitData().id ~= "spider_mod"
end

function _M:acquireTargets(targets, game, sim, itemUnit, userUnit)
    if not self:canUseAbility(sim, itemUnit, userUnit) then
        return nil
    end

    local x0, y0 = userUnit:getLocation()
    local units = {}
    for _, targetUnit in pairs(sim:getAllUnits()) do
        local x1, y1 = targetUnit:getLocation()
        if x1 and y1 and self:isTarget(targetUnit) and targetUnit ~= userUnit then
            local range = mathutil.dist2d(x0, y0, x1, y1)
            if range > 0 and range <= itemUnit:getTraits().range then
                local targetGhost = userUnit:getPlayerOwner():getLastKnownUnit(sim, targetUnit:getID())
                local x2, y2 = targetGhost:getLocation()
                if
                    (targetGhost:isGhost() and x1 == x2 and y1 == y2) or
                        sim:canPlayerSeeUnit(userUnit:getPlayerOwner(), targetUnit)
                 then
                    table.insert(units, targetUnit)
                end
            end
        end
    end

    return targets.unitTarget(game, units, self, itemUnit, userUnit)
end

function _M:canUseAbility(sim, itemUnit, userUnit, targetCell)
    if itemUnit ~= userUnit and itemUnit:getUnitOwner() ~= userUnit then
        return false
    end

    if userUnit:getTraits().movingBody then
        return false, STRINGS.UI.REASON.DROP_BODY_TO_USE
    end

    if itemUnit:getTraits().cooldown and itemUnit:getTraits().cooldown > 0 then
        return false, util.sformat(STRINGS.UI.REASON.COOLDOWN, itemUnit:getTraits().cooldown)
    end

    if itemUnit:getTraits().usesCharges and itemUnit:getTraits().charges < 1 then
        return false, util.sformat(STRINGS.UI.REASON.CHARGES)
    end

    return abilityutil.checkRequirements(itemUnit, userUnit)
end

function _M:executeAbility(sim, itemUnit, userUnit, targetUnitID)
    local x0, y0 = userUnit:getLocation()
    assert(x0 and y0, "No user location: " .. tostring(x0) .. "/" .. tostring(y0))
    local cell0 = sim:getCell(x0, y0)
    local targetUnit = sim:getUnit(targetUnitID)
    assert(targetUnit, "No target : " .. tostring(targetUnitID))
    local x1, y1 = targetUnit:getLocation()
    assert(x1 and y1, "No target location: " .. tostring(x1) .. "/" .. tostring(y1))
    local cell1 = sim:getCell(x1, y1)
    local userPlayer = userUnit and userUnit:getPlayerOwner()

    -- swapping with an enemy should have potential for friendly fire (for the NPC)
    local meleeOverwatch
    local shootOverwatch
    if userPlayer:isPC() and targetUnit:isNPC() then
        meleeOverwatch = {}
        shootOverwatch = {}
        for _, otherUnit in pairs(sim:getAllUnits()) do
            if not otherUnit:isPC() then
                if otherUnit:getTraits().isMeleeAiming then
                    local meleeAbility = otherUnit:hasAbility("melee")
                    if meleeAbility and otherUnit:canUseAbility(sim, otherUnit, userUnit) then
                        table.insert(meleeOverwatch, otherUnit)
                    end
                elseif otherUnit:isAiming() or otherUnit:getTraits().isArmed then
                    local weaponUnit = simquery.getEquippedGun(otherUnit)
                    local shot = simquery.calculateShotSuccess(sim, otherUnit, userUnit, weaponUnit)
                    if shot and not shot.armorBlocked and array.find(otherUnit:getSeenUnits(), userUnit) then
                        table.insert(shootOverwatch, otherUnit)
                    end
                end
            end
        end
    end

    -- special case if the swappee himself holds user at gunpoint
    local targetShootsSelf =
        targetUnit:isAiming() and simquery.isEnemyAgent(userPlayer, targetUnit) and
        array.find(targetUnit:getSeenUnits(), userUnit)

    if targetShootsSelf then
        local oldFacing = targetUnit:getFacing()
        local newFacing = simquery.getDirectionFromDelta(x0 - x1, y0 - y1)
        sim:dispatchEvent(
            simdefs.EV_UNIT_START_SHOOTING,
            {
                unitID = targetUnit:getID(),
                newFacing = newFacing,
                oldFacing = oldFacing,
                overwatch = true,
                targetUnitID = userUnit:getID()
            }
        )
        local newFacing = simquery.getDirectionFromDelta(x0 - x1, y0 - y1)
        sim:dispatchEvent(simdefs.EV_UNIT_STOP_SHOOTING, {unitID = targetUnit:getID(), facing = newFacing})
    end

    -- fancy stuff
    sim:dispatchEvent(simdefs.EV_PLAY_SOUND, {sound = "SpySociety/Actions/Teleport_strike_inbound", x = x0, y = y0})
    sim:dispatchEvent(
        simdefs.EV_UNIT_ADD_FX,
        {unit = userUnit, kanim = "fx/glitch_fx", anim = "out", symbol = "effect", x = x0, y = y0}
    )
    sim:dispatchEvent(
        simdefs.EV_UNIT_ADD_FX,
        {unit = targetUnit, kanim = "fx/glitch_fx", anim = "out", symbol = "effect", x = x1, y = y1}
    )

    sim:warpUnit(userUnit, cell1)
    sim:warpUnit(targetUnit, cell0)

    targetUnit:setAlerted(true)

    sim:dispatchEvent(simdefs.EV_UNIT_APPEARED, {unitID = userUnit:getID(), noSightingFx = true})
    sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = userUnit})
    sim:dispatchEvent(simdefs.EV_UNIT_APPEARED, {unitID = targetUnitID, noSightingFx = true})
    sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = targetUnit})

    if targetShootsSelf then
        local weapon = simquery.getEquippedGun(targetUnit)
        local wouldBeDmgt = simquery.calculateShotSuccess(sim, targetUnit, userUnit, weapon)
        if weapon and wouldBeDmgt and not wouldBeDmgt.armorBlocked then
            targetUnit:setInvisible(false)
            targetUnit:setDisguise(false, nil, true)
            local dmgt = simquery.calculateShotSuccess(sim, targetUnit, targetUnit, weapon)
            sim:hitUnit(targetUnit, targetUnit, dmgt)
            sim:triggerEvent(simdefs.TRG_UNIT_SHOT, {targetUnit = targetUnit, sourceUnit = targetUnit})
            if not targetUnit:getTraits().silencer then
                sim:emitSound(simdefs.SOUND_MED, x0, y0, targetUnit)
            end
        end
    end

    if meleeOverwatch then
        for _, otherUnit in pairs(meleeOverwatch) do
            local meleeAbility = otherUnit:hasAbility("melee")
            if meleeAbility and meleeAbility:canUseAbility(sim, otherUnit, targetUnit) then
                local meleeOverwatchAbility = otherUnit:hasAbility("meleeOverwatch")
                if meleeOverwatchAbility then
                    meleeOverwatchAbility:executeAbility(sim, otherUnit, targetUnit)
                else
                    meleeAbility:executeAbility(sim, otherUnit, otherUnit, targetUnitID)
                end
            end
            otherUnit:resetAllAiming()
        end
    end
    if shootOverwatch then
        for _, otherUnit in pairs(shootOverwatch) do
            local shootAbility = otherUnit:hasAbility("shootOverwatch")
            if shootAbility and shootAbility:canUseAbility(sim, otherUnit, targetUnit) and not targetUnit:isDown() then
                local weaponUnit = simquery.getEquippedGun(otherUnit)
                shootAbility:executeAbility(sim, weaponUnit, otherUnit, targetUnit)
            end
            otherUnit:resetAllAiming()
        end
    end

    sim:processReactions(userUnit)
    sim:processReactions(targetUnit)

    if shootOverwatch then
        for _, otherUnit in pairs(shootOverwatch) do
            if otherUnit.armTurret then
                otherUnit:armTurret(sim)
            end
        end
    end

    userUnit:stopHacking(sim)

    inventory.useItem(sim, userUnit, itemUnit)
end

return _M
