local array = include("modules/array")
local mathutil = include("modules/mathutil")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local util = include("modules/util")
local line_of_sight = include("sim/line_of_sight")

local function guardTrackMovementToCell(sim, userUnit, x0, y0, x1, y1)
    if userUnit:getTraits().invisible then
        return
    end

    local targetCell = sim:getCell(x1, y1)
    local line = mathutil.getCellsBetween(x1, y1, x0, y0)
    table.insert(line, x0)
    table.insert(line, y0)

    local notified = {}
    for i, x2, y2 in util.xypairs(line) do
        local seerUnits = sim._los:getSeers(x2, y2)
        if seerUnits then
            for _, unitID in ipairs(seerUnits) do
                if not notified[unitID] then
                    local seerUnit = sim:getUnit(unitID)
                    if seerUnit and seerUnit:getTraits().isGuard and seerUnit:getBrain() then
                        if
                            seerUnit:getBrain():getTarget() == userUnit and
                                simquery.couldUnitSee(sim, seerUnit, userUnit) and
                                simquery.couldUnitSeeCell(sim, seerUnit, targetCell)
                            then
                            log:write("[Knockback] Agent tracked by " .. seerUnit:getID())
                            seerUnit:turnToFace(x1, y1)
                        elseif
                            seerUnit:getBrain():getSenses():hasTarget(userUnit) and
                                not array.find(seerUnit:getSeenUnits(), userUnit)
                            then
                            log:write("[Knockback] Agent lost by " .. seerUnit:getID())
                            seerUnit:getBrain():getSenses():removeTarget(userUnit)
                            seerUnit:getBrain():getSenses():addInterest(
                                x2,
                                y2,
                                simdefs.SENSE_SIGHT,
                                simdefs.REASON_LOSTTARGET,
                                userUnit
                            )
                        elseif not array.find(seerUnit:getSeenUnits(), userUnit) then
                            log:write("[Knockback] Agent noticed by " .. seerUnit:getID())
                            seerUnit:getBrain():getSenses():addInterest(
                                x2,
                                y2,
                                simdefs.SENSE_PERIPHERAL,
                                simdefs.REASON_NOTICED,
                                userUnit
                            )
                        end
                    end
                    notified[unitID] = true
                end
            end
        end
    end
end

local function processLasersOnCells(sim, unit, cellCoords)
    for i, x, y in util.xypairs(cellCoords) do
        for i, targetUnit in ipairs(sim:getCell(x, y).units) do
            if targetUnit.tripLaser then
                targetUnit:tripLaser(sim, x, y, unit)
                if targetUnit:getTraits().emitterID then
                    local emitterUnit = sim:getUnit(targetUnit:getTraits().emitterID)
                    if emitterUnit and emitterUnit:getTraits().empWhenTripped then
                        emitterUnit:processEMP(emitterUnit:getTraits().empWhenTripped)
                    end
                end
            end
        end
    end
end

local function knockUnitToCell(sim, unit, cell)
    local x0, y0 = unit:getLocation()
    if cell and (cell.x ~= x0 or cell.y ~= y0) then
        sim:warpUnit(unit, cell)
        local line = mathutil.getCellsBetween(x0, y0, cell.x, cell.y)
        table.insert(line, 1, cell.x)
        table.insert(line, 2, cell.y)
        processLasersOnCells(sim, unit, line)
    end
end

local function applyKnockback(sim, x0, y0, x1, y1)
    local line = mathutil.getCellsBetween(x0, y0, x1, y1)
    table.insert(line, 1, x0)
    table.insert(line, 2, y0)

    -- TODO handle smoke & inconspicuous ficus correctly
    local los = line_of_sight(sim) -- custom LoS that *only* uses walls
    -- raycast is float, normalise to first blocked cell
    local xr, yr = los:raycast(x0, y0, x1, y1)
    local xrLim = mathutil.getCloserSideOfWall(xr, x1)
    local yrLim = mathutil.getCloserSideOfWall(yr, y1)

    local hitUnits = {}
    local currentUnit = nil
    local prevCell = nil

    for i, x2, y2 in util.xypairs(line) do
        log:write("[Knockback] Knocking into cell... ".. x2 .."/".. y2)
        local cell = sim:getCell(x2, y2)
        if not cell or cell.impass > 0 or (xrLim and x2 == xrLim) or (yrLim and y2 == yrLim) then
            log:write("[Knockback] Unit hit impass: ".. currentUnit:getName() or "nil")
            knockUnitToCell(sim, currentUnit, prevCell)
            hitUnits[#hitUnits].impactDist = i
            currentUnit = nil
            break
        end
        for u, unit in ipairs(cell.units) do
            if unit:getTraits().dynamicImpass and simquery.isAgent(unit) then
                log:write("[Knockback] New affected unit: ".. unit:getName() or "nil")
                if currentUnit then
                    knockUnitToCell(sim, currentUnit, prevCell)
                    hitUnits[#hitUnits].impactDist = i
                end
                currentUnit = unit
                table.insert(hitUnits, {unit = unit, startDist = i})
                break
            end
        end
        prevCell = cell
    end

    if currentUnit then
        log:write("[Knockback] Unit hit nothing: ".. currentUnit:getName() or "nil")
        knockUnitToCell(sim, currentUnit, prevCell)
    end

    return hitUnits
end

return {
    processLasersOnCells = processLasersOnCells,
    guardTrackMovementToCell = guardTrackMovementToCell,
    applyKnockback = applyKnockback,
}
