--------------------------------------------------------------------
-- Master file for dragAgents. In charge of making sure everything
-- is set up properly for this feature. modinit is responsible for
-- passing relevant campaign generation options.
--
-- Feature description:
-- Guards carry agents to a guard elevator (NOT the exit elevator
-- that agents use) and teleport the agent out.
-- Teleported agent can only be retrieved by doing a detention
-- center mission.
-- Guards should remain in the level after teleporting the agent
-- and will resume hunting.
-- Guards should stop to shoot any agent that gets in their way
-- as they head to the guard elevator.

local abilitydefs = include("sim/abilitydefs")
local actions = include("sim/btree/actions")
local modifiers = include("sim/modifiers")
local senses = include("sim/btree/senses")
local simability = include("sim/simability")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local simunit = include("sim/simunit")
-- local util = include("modules/util")

--------------------------------------------------------------------
-- Mod campaign option information, set by modinit
local DRAG_AGENTS = false
local IGNORE_MISSION_CRITICAL_AGENTS = true
local function setModInfo(dragAgent, ignoreMissionCriticalAgents)
	DRAG_AGENTS = dragAgent or false
	IGNORE_MISSION_CRITICAL_AGENTS = ignoreMissionCriticalAgents or true
end

local function lateLoad(modApi, options)
	-- Mark items that should never be removed from the level by a guard.
	-- These items are required to complete the mission.
	local itemdefs = include("sim/unitdefs/itemdefs")
	local propdefs = include("sim/unitdefs/propdefs") -- Some items are in propdefs
	-- Flag important items
	-- Vanilla items:
	if itemdefs.augment_final_level ~= nil then -- augment that gets installed in monst3r during the final level. Since this isn't droppable marking it as nct_noFlyList isn't really necessary.
		itemdefs.augment_final_level.traits.nct_noFlyList = true
	end
	if itemdefs.quest_material ~= nil then -- no idea what this is
		itemdefs.quest_material.traits.nct_noFlyList = true
	end
	if itemdefs.special_exit_passcard ~= nil then -- passcard required to reach the teleporter
		itemdefs.special_exit_passcard.traits.nct_noFlyList = true
	end
	if itemdefs.item_incognita ~= nil then -- incognita as an item. Carried by Central in the final mission.
		itemdefs.item_incognita.traits.nct_noFlyList = true
	end
	-- Contingency Plan DLC items:
	if propdefs.item_incognita_case ~= nil then -- special container that must be collected during the mid mission.
		propdefs.item_incognita_case.traits.nct_noFlyList = true
	end
	if propdefs.item_power_cell ~= nil then -- power cell that must be retrieved each mission (after the mid mission) to keep incognita alive.
		propdefs.item_power_cell.traits.nct_noFlyList = true
	end
end

--------------------------------------------------------------------
-- helper functions

-- Returns true if this agent is Central on the final mission or Monst3r
-- on the mid or final missions
local function isMissionCriticalAgent(agent)
	local situationName = agent:getSim():getParams().situationName
	if agent:getTraits().central == true and situationName == "ending_1" then
		return true
	elseif agent:getTraits().monst3r == true and (situationName == "mid_1"
	                                           or situationName == "mid_2"
	                                           or situationName == "ending_1") then
		return true
	end
	return false
end

-- Returns the first player agent on this cell or nil if no agents are on this cell
-- ignoreMissionCriticalAgents -> optional parameter. If true, then Monst3r and Central
-- will not be returned during the mid (DLC) and final missions.
local function cellContainsAgent(cell, ignoreMissionCriticalAgents)
	local agent = nil
	for _, checkUnit in ipairs(cell.units) do
		if checkUnit.getTraits ~= nil and checkUnit:isDown() == true and checkUnit:getTraits().isAgent == true and checkUnit:getBrain() == nil and checkUnit:getTraits().isDrone == nil then
			if not (ignoreMissionCriticalAgents == true and isMissionCriticalAgent(checkUnit) == true) then
				agent = checkUnit
				break
			end
		end
	end
	return agent
end

-- Functions to control guard focus. Used to ensure a guard does not deviate
-- from picking up an agent and dragging them to the elevator
local function isFocusedOnAgent(guardUnit, agentUnit)
	if agentUnit.nct_escortGuard and agentUnit.nct_escortGuard:isValid() and
		 agentUnit.nct_escortGuard:getID() == guardUnit:getID() then
		return true
	end
	return false
end

local function focusOnAgent(guardUnit, agentUnit)
	if not isFocusedOnAgent(guardUnit, agentUnit) then
		if agentUnit.nct_escortGuard then
			agentUnit.nct_escortGuard:getModifiers():remove("nct_escortGuard", "noInterestDistraction")
		end
		agentUnit.nct_escortGuard = guardUnit
		guardUnit:getModifiers():add("noInterestDistraction", "nct_escortGuard", modifiers.SET, true)
	end
end

local function unfocusAgent(guardUnit, agentUnit)
	agentUnit.nct_escortGuard = nil
	guardUnit:getModifiers():remove("nct_escortGuard", "noInterestDistraction")
end

--------------------------------------------------------------------
-- actions helper Functions
local function teleportAgent(sim, unit)
	local cell = sim:getCell(unit:getLocation())
	if not cell then
		return
	end

	if not simquery.cellHasTag(sim, cell, "guard_spawn") then
		return
	end

	-- Drop the agent
	local agentUnit = unit:getTraits().movingBody
	local moveBody = unit:hasAbility("moveBody")
	moveBody:executeAbility(sim, unit, unit, agentUnit:getID())
	-- Ransack the agent for any items that are required to complete the mission
	for _,itemUnit in ipairs(agentUnit:getChildren()) do
		if itemUnit.getTraits and itemUnit:getTraits().nct_noFlyList == true then
			-- Remove from agent inventory
			agentUnit:removeChild(itemUnit)
			if itemUnit:getTraits().drop_dropdisguise then
				agentUnit:setDisguise(false)
			end
			if itemUnit:getTraits().equipped then
				itemUnit:getTraits().equipped = nil
			end
			-- Add to guards inventory
			if itemUnit:getTraits().pickup_trigger then
				sim:triggerEvent(itemUnit:getTraits().pickup_trigger)
			end
			sim:dispatchEvent(simdefs.EV_UNIT_PICKEDUP, {unit=itemUnit})
			sim:triggerEvent(simdefs.TRG_UNIT_PICKEDUP, {item=itemUnit, unit=unit})
			unit:addChild(itemUnit)
			sim:triggerEvent("agentGotItem", {item=itemUnit})
			sim:dispatchEvent(simdefs.EV_PLAY_SOUND, "SpySociety/HUD/gameplay/HUD_ItemStorage_PutIn")
			sim:dispatchEvent(simdefs.EV_UNIT_GET_ITEM, {item = itemUnit})
		end
	end
	-- Resume our normal routinue (should be hunting since we should be alerted)
	unfocusAgent(unit, agentUnit)
	-- Prepare the agent for removal from the mission.
	-- Note a whole host of things can go wrong in vanilla code if it interacts with an agent with no location. So do this stuff before warping the unit to nil.
	-- I wanted to fully despawn the agent, thus removing it from the list of active agents in the bottom left of the in-mission hud.
	-- However:
	-- 1. If a red ghost is left of the agent the agentrig will expect the agent to still exist.
	-- 2. The agents current inventory only gets saved if the agent is still valid at the end of the mission (has an _id and _sim and is still stored in the master _units table).
	-- The main reason to despawn the unit is to remove the unit from the HUD and to avoid a "are you sure you want to leave?" message when escaping with your remaining agents.
	-- A hackish way to do that is to make the agent no longer owned by the player.
	-- I would prefer to set the player owner to nil but the agent has to have an owner to avoid agentrig errors.
	-- Thankfully, since the agent doesn't have a brain, the NPC won't try to do something silly with the agent.
	agentUnit:setPlayerOwner(sim:getNPC())
	agentUnit:setKO(sim, nil) -- Stop the agent from trying to wake up (game freezes if a K.O.ed agent is removed from the level without this)
	-- Transport the agent
	sim:dispatchEvent(simdefs.EV_TELEPORT, {units = {agentUnit}, warpOut = true})
	sim:warpUnit(agentUnit, nil)
end

local function headToGuardElevator(sim, guardUnit, agentUnit)
	local succeeded = true
	-- Setup an interest at the nearest guard teleporter, also make sure that our interest does not change while enroute
	local destination = actions.MoveToNearestExit.getDestination({unit = guardUnit}) -- Note that this function call WILL fail if UI tweaks is not installed, as we are assuming that guards can manipulate doors while dragging bodies (which they cannot do in vanilla game).
	if destination == nil then
		succeeded = false
		log:write("New Corporate Tactics: Error > Could not find destination to drag agent to. Trying again next turn.")
		guardUnit.nct_destinationFailed = true
		destination = sim:getCell(guardUnit:getLocation())
	end
	local interest = guardUnit:getBrain():getSenses():addInterest(destination.x, destination.y, simdefs.SENSE_RADIO, simdefs.REASON_HUNTING)--, sourceUnit, ignoreDisguies)
	interest.remember = true -- if we somehow get distracted, make sure we go back to this interest.
	focusOnAgent(guardUnit, agentUnit) -- Ignore requests for new interests (will still shoot agents that get in line of sight).
	return succeeded
end

--------------------------------------------------------------------
-- Extend actions to respond appropriately to drag situations
local old_MarkInterestInvestigated = actions.MarkInterestInvestigated
actions.MarkInterestInvestigated = function(sim, unit)
	local returnValue = old_MarkInterestInvestigated(sim, unit)
	if DRAG_AGENTS == true then
		if returnValue == simdefs.BSTATE_COMPLETE then
			if unit:getTraits().movingBody ~= nil then
				if simquery.cellHasTag(sim, sim:getCell(unit:getLocation()), "guard_spawn") == true then
					teleportAgent(sim, unit)
				elseif unit.nct_destinationFailed == true then
					unfocusAgent(unit, unit:getTraits().movingBody)
					if headToGuardElevator(sim, unit, unit:getTraits().movingBody) == false then
						returnValue = simdefs.BSTATE_FAILED
					end
				else
					log:write("New Corporate Tactics: Error > Absolutely no idea what happened while trying to drag body. Please report to developer.")
				end
			else
				-- Drones cannot carry agents
				if unit:getTraits().isDrone ~= true and unit:getTraits().movingBody == nil then
					-- Is there a downed agent (make sure it's not a guard!) at this interest spot?
					local cell = unit:getSim():getCell(unit:getLocation())
					local agent = cellContainsAgent(cell, IGNORE_MISSION_CRITICAL_AGENTS)
					if agent ~= nil and agent ~= unit then
						-- Pick this agent up and stop checking for more agents. We can't carry more than one agent at a time.
						if not unit:hasAbility("moveBody") then
							if unit._abilities then
								table.insert(unit._abilities, simability.create("moveBody"))
							else
								unit._abilities = {simability.create("moveBody")}
							end
						end
						local moveBody = unit:hasAbility("moveBody")
						moveBody:executeAbility(sim, unit, unit, agent:getID())
						-- Setup an interest at the nearest guard teleporter, also make sure that our interest does not change while enroute
						if headToGuardElevator(sim, unit, agent) == false then
							returnValue = simdefs.BSTATE_FAILED
						end
						-- Wound the agent, lowering their MP for the rest of the mission
						if agent:getTraits().woundMe then
							local currentWounds = agent:getTraits().wounded or 0
							local reduceMPby = agent:getTraits().woundMe
							agent:getTraits().wounded = currentWounds + reduceMPby
							agent:addMP(-reduceMPby)
							agent:addMPMax(-reduceMPby)
							agent:getTraits().woundMe = nil
						end
					end
				end
			end
		end
	end

	return returnValue
end

local old_ReactToInterest = actions.ReactToTarget
actions.ReactToTarget = function(sim, unit)
	if DRAG_AGENTS == true then
		local target = unit:getBrain():getTarget()
		local x1,y1 = target:getLocation()
		if x1 and y1 and unit:isValid() and target and target:isValid() and target:getLocation() then
			if unit:getTraits().movingBody then
				local agentUnit = unit:getTraits().movingBody
				local moveBody = unit:hasAbility("moveBody")
				moveBody:executeAbility(sim, unit, unit, agentUnit:getID())
				local interest = unit:getBrain():getSenses():getCurrentInterest()
				interest.x, interest.y = unit:getLocation()
			end
		end
	end
	return old_ReactToInterest(sim,unit) -- Must come AFTER dropping the body to avoid visiual glitches
end

--------------------------------------------------------------------
-- Extend simunit to respond appropriately to drag situations

-- Agents dragged by guards should count as neutralized
local old_isNeutralized = simunit.isNeutralized
simunit.isNeutralized = function(self)
	local returnValue = old_isNeutralized(self)
	if DRAG_AGENTS == true then
		if returnValue ~= true and simquery.isUnitDragged(self:getSim(), self) == true then
			returnValue = true
		end
	end
	return returnValue
end

-- Need to unfocus guards if an agent moves before the guard can get to them.
local old_onWarp = simunit.onWarp
simunit.onWarp = function(self, sim, oldCell, newCell)
	old_onWarp(self, sim, oldCell, newCell)
	if DRAG_AGENTS == true then
		local escortGuard = self.nct_escortGuard
		if escortGuard ~= nil then
			if escortGuard:getTraits().movingBody == nil or escortGuard:getTraits().movingBody:getID() ~= self:getID() then
				unfocusAgent(escortGuard, self)
			end
		end
	end
end


--------------------------------------------------------------------
-- Extend senses to stop guards from just walking away from downed agents

local old_addInterest = senses.addInterest
senses.addInterest = function(self, x, y, sense, reason, sourceUnit, ignoreDisguies)

	local interest = old_addInterest(self, x, y, sense, reason, sourceUnit, ignoreDisguies)
	if DRAG_AGENTS == true and interest ~= nil then
		if self.unit:getTraits().isDrone ~= true then -- Drones can't drag agents, so there is no reason for them to focus on downed agents.
			if reason == simdefs.REASON_FOUNDCORPSE and sourceUnit:isValid() == true and sourceUnit:getTraits().isAgent == true and sourceUnit:getBrain() == nil then
				local escortGuard = sourceUnit.nct_escortGuard
				if (escortGuard ~= nil and escortGuard:isDown()) or escortGuard == nil then
					focusOnAgent(self.unit, sourceUnit)
				end
			end
		end
	end
	return interest
end

--------------------------------------------------------------------
-- Extend simquery so that guards can drag bodies through doors
local old_canModifyExit = simquery.canModifyExit
-- Below code is taken from osheroff's UI Tweaks mod and modified to
-- affect only guards and to allow guards to move through doors while
-- dragging regardless of whether or not the body can be dropped at
-- that location.
-- The primary reason this code was copied was because osheroff's code
-- called moveBody:canUseAbility(). However it would ONLY call the Vanilla
-- version of moveBody:canUseAbility(). This means that if a mod overwrites
-- the moveBody ability (like function library does) that my changes to
-- moveBody:canUseAbility (found further below) were not guaranteed to be
-- used by osheroff's code.
simquery.canModifyExit = function(unit, exitop, cell, dir)
	local canModify, reason = old_canModifyExit(unit, exitop, cell, dir)

	if canModify == false and reason == STRINGS.UI.DOORS.DROP_BODY then
		if unit:getTraits().isGuard then
			local body = unit:getTraits().movingBody
			unit:getTraits().movingBody = nil
			canModify, reason = old_canModifyExit(unit, exitop, cell, dir)
			unit:getTraits().movingBody = body
		end
	end

	return canModify, reason
end


--------------------------------------------------------------------
-- Extend the moveBody ability so that guards can always drop bodies,
--   even when there is already a body on the same cell.
-- This allows guards to find the exit even when picking up an agent
--   off a stack of multiple bodies.
local moveBody = abilitydefs.lookupAbility("moveBody")
local old_moveBody_canUseAbility = moveBody.canUseAbility
moveBody.canUseAbility = function(self, sim, unit, userUnit, targetID)
	local returnValue, returnReason = old_moveBody_canUseAbility(self, sim, unit, userUnit, targetID)
	if DRAG_AGENTS == true then
		if returnValue == false and returnReason == STRINGS.UI.REASON.CANT_STACK_BODIES then
			if unit:getTraits().isGuard == true then
				returnValue, returnReason = true, nil
			end
		end
	end
	return returnValue, returnReason
end

return {
	setModInfo = setModInfo,
	lateLoad = lateLoad
}
