--------------------------------------------------------------------
-- Master file for eliteCameras. In charge of making sure everything
-- is set up properly for this feature. modinit is responsible for
-- passing relevant campaign generation options.
--
-- This file contains the code for all camera related features that
-- this mod provides, they are as follows:
--
-- TRACKING CAMERAS:
-- When a camera spots an agent, it reports that
-- agents last seen location to a nearby guard. This differs from
-- vanilla game where the camera reports the first tile the agent
-- was spotted on.
--
-- CAMERAS DETECT DOWNED GUARDS:
-- If a guard is downed, the camera will raise the
-- alarm, regardless of whether or not it could see the agent.

-- CAMERA ACTIVATION COUNT:
-- Cameras can raise the alarm level multiple times in one turn
-- depending on what this value is set to.

local simcamera = include("sim/units/simcamera")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
-- local util = include("modules/util")

--------------------------------------------------------------------
-- Mod campaign option information, set by modinit
local TRACKING_CAMERAS = false
local CAMERAS_DETECT_DOWNED_GUARDS = false
local CAMERA_ACTIVATION_COUNT = 1
local function setModInfo(trackingCameras, camerasDetectDownedGuards, cameraActivationCount)
	TRACKING_CAMERAS = trackingCameras or false
	CAMERAS_DETECT_DOWNED_GUARDS = camerasDetectDownedGuards or false
	CAMERA_ACTIVATION_COUNT = cameraActivationCount or 1
end

--------------------------------------------------------------------
-- extend simcamera

-- Returns true if:
-- 1. The camera is controlled by the corporation.
-- 2. targetUnit is a conscious agent belonging to the player OR targetUnit is a downed guard.
-- 3. The camera can see targetUnit.
local function nct_canTrackTarget(cameraUnit, targetUnit)
	if not cameraUnit:isPC() and targetUnit then
		if (simquery.isEnemyAgent(cameraUnit:getPlayerOwner(), targetUnit) and not targetUnit:isKO())
		or (CAMERAS_DETECT_DOWNED_GUARDS == true and not simquery.isEnemyAgent(cameraUnit:getPlayerOwner(), targetUnit) and targetUnit:isDown()) then
			if cameraUnit._sim:canUnitSeeUnit(cameraUnit, targetUnit) then
				return true
			end
		end
	end
	return false
end

local old_onSpawn = simcamera.onSpawn
simcamera.onSpawn = function(self, sim)
	if old_onSpawn then
		old_onSpawn(self, sim)
	end
	if CAMERAS_DETECT_DOWNED_GUARDS == true then
		sim:addTrigger(simdefs.TRG_UNIT_HIT, self)
	end
	if CAMERA_ACTIVATION_COUNT > 0 then
		self:getTraits().nct_activationsLeft = CAMERA_ACTIVATION_COUNT
		self:getTraits().nct_alreadySeenAgents = {}
	end
end

local old_onTrigger = simcamera.onTrigger
simcamera.onTrigger = function(self, sim, evType, evData)
	if CAMERA_ACTIVATION_COUNT > 0 then
		if self:isValid() and not (self:isDown() or self:getTraits().mainframe_status ~= "active" or self:isPC()) then -- No need to do stuff if not under active corporate control.
			local target = nil

			if evType == simdefs.TRG_START_TURN then
				if evData ~= nil and not evData:isNPC() then
					self:getTraits().nct_activationsLeft = CAMERA_ACTIVATION_COUNT
					self:getTraits().nct_alreadySeenAgents = {}
				end
			elseif evType == simdefs.TRG_OVERWATCH then
				target = evData
			elseif evType == simdefs.TRG_UNIT_HIT then
				target = evData.targetUnit
			end

			if target then
				if nct_canTrackTarget(self, target) then
					if self:getTraits().nct_activationsLeft > 0 and not self:getTraits().nct_alreadySeenAgents[target:getID()] then
						target:interruptMove(sim, self)
						self:performTracking(sim, target)
						self:getTraits().nct_alreadySeenAgents[target:getID()] = true
						self:getTraits().nct_activationsLeft = self:getTraits().nct_activationsLeft - 1
					elseif TRACKING_CAMERAS == true then
						local x0, y0 = self:getLocation()
						local x1, y1 = target:getLocation()
						if not self:getTraits().camera_drone then
							sim:triggerEvent( simdefs.TRG_NEW_INTEREST, { x = x0, y = y0, range = simdefs.SOUND_RANGE_3, interest = { x= x1, y = y1, reason=simdefs.REASON_CAMERA} })
						end
					end
				end
			end
		end
	end

	old_onTrigger(self, sim, evType, evData)
end

return {
	setModInfo = setModInfo
}
