--------------------------------------------------------------------
-- Master file for guardsUseCover. In charge of making sure everything
-- is set up properly for this feature. modinit is responsible for
-- passing relevant campaign generation options.
--
-- Feature Description:
--  This feature allows guards to hide behind cover when alerted, making them hard to spot by your agents.
--  It has two generation options to control it.
--  They are:
--  1. GUARDS USE COVER
--    When enabled, alerted guards will be invisible if they happen to be behind cover. Note that if an agent can see behind said cover, they will be visible.
--    Guards cannot hide even they are currently targeting an agent.
--  2. GUARDS PREFER COVER
--    When enabled this option changes guard pathing (when alerted) to prefer movements behind cover.
--    This is done by changing the calculated move cost of tiles adjacent to cover so that guards will think moving around cover is cheaper than moving in the open.
--    Actual move cost is not changed.
--  Note that each feature can be enabled independently of each other.
--  Also, both options by default are disabled because they are ridiculously sadistic.

-- TODO
-- * Guards need to use a hiding animation when beside cover (like agents).
-- * Need to remove the tooltip that appears when mousing over a square with a guard
--   ... it needs to not appear if a guard is hiding behind cover and not visible
--   > While were at it need to make sure that pressing shift on a square with a nonvisible guard does not reveal their LOS

-- I think that playing with this feature is scary and probably not fun.

local simunit = include("sim/simunit")
local simquery = include("sim/simquery")
-- local util = include("modules/util")

--------------------------------------------------------------------
-- Mod campaign option information, set by modinit
local GUARDS_HIDE_ON_ALERT = false
local GUARDS_PREFER_COVER = false
local function setModInfo(guardsHideOnAlert, guardsPreferCover)
	GUARDS_HIDE_ON_ALERT = guardsHideOnAlert or false
	GUARDS_PREFER_COVER = guardsPreferCover or false
end

--------------------------------------------------------------------
-- extend simquery

-- Need to modify vision checks to return guards hiding behind cover as invisible

-- LOS-independent sight check; should return true if targetUnit could be seen assuming a LOS-check succeeds
-- unit: The unit doing the sight check
-- targetUnit: The target unit; may be nil if only wish to check whether unit can see (at all)
-- ignoreCover: if false, do not check cover status of targetUnit.
local old_couldUnitSee = simquery.couldUnitSee

simquery.couldUnitSee = function(sim, unit, targetUnit, ignoreCover, targetCell)
	if GUARDS_HIDE_ON_ALERT == true then
		if targetUnit then
			-- Add check for alerted guards behind cover, then let the old function handle everything
			if targetUnit:getTraits().alerted == true then
				-- Guards on overwatch don't hide, make sure guard is not on overwatch
				if targetUnit:getBrain() and not targetUnit:getBrain():getTarget() then
					if not targetCell then
						targetCell = sim:getCell(targetUnit:getLocation())
					end
					-- NOTE may need to not hide guards if ignoreCover is true
					if simquery.checkIfCellNextToCover(sim, targetCell) and simquery.checkCover(sim, unit, targetCell.x, targetCell.y) then
						return false
					end
				end
			end
		end
	end
	return old_couldUnitSee(sim, unit, targetUnit, ignoreCover, targetCell)
end

--------------------------------------------------------------------
-- extend simunit
local old_setAlerted = simunit.setAlerted
simunit.setAlerted = function(self, alerted)
	if GUARDS_HIDE_ON_ALERT == true then
		if self._sim._resultTable.guards[self:getID()] then
			self:getTraits().hidesInCover = true
		end
	end
	old_setAlerted(self, alerted)
end

return {
	setModInfo = setModInfo
}
