--------------------------------------------------------------------
-- Master file for spotLootedSafes. In charge of making sure everything
-- is set up properly for this feature. modinit is responsible for
-- passing relevant campaign generation options.
--
-- Feature description:
-- If a guards sees a looted safe, he will be prompted to investigate.
-- Optionally, he may even go into 'alert' mode (default is 'investigate').
-- Guards should be prompted during an agents turn if a safe is looted in
-- their sight.

local senses = include("sim/btree/senses")
local simdefs = include("sim/simdefs")
-- local util = include("modules/util")

--------------------------------------------------------------------
-- Mod campaign option information, set by modinit
local RESPONSE_VALUES =
{
	NONE = 0,
	INVESTIGATE = 1,
	ALERT = 2
}
local SPOT_SAFES_WHEN_LOOTED = RESPONSE_VALUES.NONE
local SPOT_LOOTED_SAFES = RESPONSE_VALUES.NONE
local function setModInfo(spotSafesWhenLooted, spotLootedSafes)
	SPOT_SAFES_WHEN_LOOTED = spotSafesWhenLooted or RESPONSE_VALUES.NONE
	SPOT_LOOTED_SAFES = spotLootedSafes or RESPONSE_VALUES.NONE
end

--------------------------------------------------------------------
-- Helper functions
local function safeNotAlreadySeen(guardUnit, safeUnit)
	if guardUnit.nct_safesAlreadySeen == nil then
		return true
	end
	return guardUnit.nct_safesAlreadySeen[safeUnit:getID()] == nil
end

-- Mark the safe as seen to avoid repeat investigations
local function spotSafe(guardUnit, safeUnit)
	if guardUnit.nct_safesAlreadySeen == nil then
		guardUnit.nct_safesAlreadySeen = {[safeUnit:getID()] = true}
	else
		guardUnit.nct_safesAlreadySeen[safeUnit:getID()] = true
	end
end

-- Determines the guards response level to a looted safe
local function getSafeSpottedReason(responseValue)
	if responseValue == RESPONSE_VALUES.INVESTIGATE then
		return simdefs.REASON_NOTICED
	elseif responseValue == RESPONSE_VALUES.ALERT then
		return simdefs.REASON_FOUNDOBJECT
	end
end

--------------------------------------------------------------------
-- Extend senses

-- Controls a guards direct line of sight (when the guard moves)
local old_processAppearedTrigger = senses.processAppearedTrigger
senses.processAppearedTrigger = function(self, sim, evData)
	old_processAppearedTrigger(self, sim, evData)
	if SPOT_LOOTED_SAFES ~= RESPONSE_VALUES.NONE then
		local seerUnit = sim:getUnit(evData.seerID)
		if seerUnit ~= self.unit then
			return -- the seer isn't us
		end

		if self.unit:isKO() or self.unit:isDead() then
			return
		end

		local x,y = evData.unit:getLocation()

		if evData.unit:getTraits().safeUnit == true and evData.unit:getTraits().open == true and safeNotAlreadySeen(self.unit, evData.unit) then
			spotSafe(self.unit, evData.unit)
			self:addInterest(x, y, simdefs.SENSE_SIGHT, getSafeSpottedReason(SPOT_LOOTED_SAFES))
		end
	end
end

-- Controls a guards peripheral line of sight (when the guard moves)
local old_checkPeripheralVision = senses.checkPeripheralVision
senses.checkPeripheralVision = function(self)
	old_checkPeripheralVision(self)
	if SPOT_LOOTED_SAFES ~= RESPONSE_VALUES.NONE then
		local losCoords, cells = {}, {}
		local sim = self.unit:getSim()
		sim:getLOS():getPeripheralVizCells( self.unit:getID(), losCoords )
		for i = 1, #losCoords, 2 do
			local x, y = losCoords[i], losCoords[i+1]
			table.insert(cells, sim:getCell(x, y))
		end
		for _, cell in ipairs(cells) do
			for _, unit in ipairs(cell.units) do
				local canSee, canSense = sim:canUnitSeeUnit(self.unit, unit)
				if not canSee and canSense then
					if unit:getTraits().safeUnit == true and unit:getTraits().open == true and safeNotAlreadySeen(self.unit, unit) then
						spotSafe(self.unit, unit)
						self:addInterest(cell.x, cell.y, simdefs.SENSE_PERIPHERAL, getSafeSpottedReason(SPOT_LOOTED_SAFES))
					end
				end
			end
		end
	end
end

-- Handle trigger responses. Need to respond to safes getting looted in our vision
local old_onTrigger = senses.onTrigger
senses.onTrigger = function(self, sim, evType, evData)
	if self.unit == nil or self:checkDisabled() then
		return
	end
	old_onTrigger(self, sim, evType, evData)

	if SPOT_SAFES_WHEN_LOOTED ~= RESPONSE_VALUES.NONE then
		if evType == simdefs.TRG_SAFE_LOOTED then
			local canSee, canSense = sim:canUnitSeeUnit(self.unit, evData.targetUnit)
			local senseType = nil
			if canSee == true then
				senseType = simdefs.SENSE_SIGHT
			elseif canSense == true then
				senseType = simdefs.SENSE_PERIPHERAL
			end
			if senseType ~= nil then
				spotSafe(self.unit, evData.targetUnit)
				local x, y = evData.unit:getLocation()
				self:addInterest(x, y, senseType, getSafeSpottedReason(SPOT_SAFES_WHEN_LOOTED))
				sim:processReactions(self.unit)
			end
		end
	end
end

-- Don't forget to listen for the trigger!
local old_onSpawned = senses.onSpawned
senses.onSpawned = function(self, sim, unit)
	old_onSpawned(self, sim, unit)
	sim:addTrigger(simdefs.TRG_SAFE_LOOTED, self)
end

-- Got to remove the trigger as well. It's kinda awkward when dead things try to keep doing stuff...
local old_onDespawned = senses.onDespawned
senses.onDespawned = function(self)
	self.unit:getSim():removeTrigger(simdefs.TRG_SAFE_LOOTED, self)
	old_onDespawned(self)
end

return {
	RESPONSE_VALUES = RESPONSE_VALUES,
	setModInfo = setModInfo
}
