--------------------------------------------------------------------
-- Master file for surrender. In charge of making sure everything
-- is set up properly for this feature. modinit is responsible for
-- passing relevant campaign generation options.
--
-- Feature description:
-- When an agents is under overwatch, they can 'surrender'. This
-- K.O.s the agent for 1 turn.

local abilitydefs = include("sim/abilitydefs")
local modifiers = include("sim/modifiers")
local simquery = include("sim/simquery")
local simunit = include("sim/simunit")
local util = include("modules/util")
local myUtil = nil

local EASE_OF_DRAGGING_FRIENDLY = 1 -- Vanilla game assumes that guards will never drag anything. It also makes dragging your own agents easier than dragging guards.
-- Unfortunately, this ease of dragging your own agents also causes guards to be able to drag your agents easier than their own.
-- So we store this value so that we can cause guards to drag agents at the same speed your agents drag guards at.

--------------------------------------------------------------------
-- Tooltip to inform player of the severity of an agents wounds
local tooltips = {
	onAgentTooltip = function(tooltip, unit)
		if unit:getTraits().wounded or 0 > 0 then
			tooltip:addAbility(STRINGS.NEW_CORP_TACTICS.TOOLTIPS.WOUNDED, util.sformat(STRINGS.NEW_CORP_TACTICS.TOOLTIPS.WOUNDED_DESC, unit:getTraits().wounded), "gui/icons/arrow_small.png")
		end
	end
}

--------------------------------------------------------------------
-- Mod campaign option information, set by modinit
local FAIR_DRAG = false
local SURRENDER = false
local KO_DRAG_COST = 0
local WOUNDED = 0
local WOUNDED_MAX = 0

local function setModInfo(fairDrag, surrender, koDragCost, wounded, woundedMax)
	FAIR_DRAG = fairDrag or false
	SURRENDER = surrender or false
	KO_DRAG_COST = koDragCost or 0
	WOUNDED = wounded or 0
	WOUNDED_MAX = woundedMax or 0
end

local function init(scriptPath)
	myUtil = include(scriptPath.."/myUtil")
end

local function load(scriptPath, modApi)
	if scriptPath and modApi then
		modApi:addTooltipDef(tooltips)
		modApi:addAbilityDef("surrender", scriptPath .."/features/surrender/surrenderAbility")
		local surrender = abilitydefs.lookupAbility("surrender")
		if surrender then
			surrender.WOUNDED = WOUNDED
			surrender.WOUNDED_MAX = WOUNDED_MAX
		else
			log:write("NEW CORPORATE TACTICS - Lookup surrender FAILED !")
		end
	end
end

local function lateLoad()
	if SURRENDER == true then
		myUtil.addAbilityToAllAgents("surrender")
	else
		myUtil.removeAbilityFromAllAgents("surrender")
	end
end

local function lateUnload()
	myUtil.removeAbilityFromAllAgents("surrender")
end

--------------------------------------------------------------------
-- Extend simunit
local old_tickKO = simunit.tickKO
simunit.tickKO = function(self, sim)
	if self:getTraits().woundMe then
		if self:isKO() and not self:isDead() and not self:getTraits().paralyzed then
			local cell = sim:getCell(self:getLocation())
			local willTick = true
			-- Don't tick KO if there is another agent standing in this cell that is not us, and is not KO.
			for i, cellUnit in ipairs(cell.units) do
				if simquery.isAgent(cellUnit) and not cellUnit:isKO() then
					willTick = false
					break
				end
			end
			if self:getTraits().koTimer <= 1 and willTick == true then -- Unit is about to wake up
				-- Remove the woundMe trait on wake up
				self:getTraits().woundMe = nil
			end
		end
	end
	old_tickKO(self, sim)
end

--------------------------------------------------------------------
-- Extend moveBody

-- Note: What I would really like to do would be to modify the moveUnit() function in "scripts\sim\engine" and change the moveCost calculation.
-- However, there is no way to change that calculation by extending the function. The only option I know would be to completely overwrite the
-- function, which I do not want to do.
-- Instead, I am extending moveBody to change a variable that is used in the moveCost calculation. It's hackish and ugly, but I don't know of a
-- better way.
local moveBody = abilitydefs.lookupAbility("moveBody")
local old_moveBody_executeAbility = moveBody.executeAbility
moveBody.executeAbility = function(self, sim, unit, userUnit, target)
	old_moveBody_executeAbility(self, sim, unit, userUnit, target)
	if KO_DRAG_COST > 0 or FAIR_DRAG == true then
		if unit:getTraits().isGuard == true then
			local body = unit:getTraits().movingBody
			if body then
				-- We just picked up a body and are now dragging them
				local guardDragMod = 0

				if FAIR_DRAG == true then
					if body:getTraits().isGuard then
						guardDragMod = guardDragMod + EASE_OF_DRAGGING_FRIENDLY -- Bonus for dragging friendly
					else
						guardDragMod = guardDragMod - EASE_OF_DRAGGING_FRIENDLY -- Remove bonus that the game gives for draging a player's agent regardless of whether or not you are the player
					end
				end

				if body:isKO() == true and body:isDead() ~= true and not body:getTraits().isGuard then
					guardDragMod = guardDragMod - KO_DRAG_COST -- dragCostMod REDUCES the effort required to drag. Therefore we need to do subtraction to INCREASE the effort required to drag.
				end

				if guardDragMod ~= 0 then
					-- Note to self: When trying to figure out how far a guard will move when dragging, don't forget that guards sprint when alerted.
					unit:getModifiers():add("dragCostMod", "nct_guardDragMod", modifiers.ADD, guardDragMod)
				end
			elseif unit:getModifiers():has("dragCostMod", "nct_guardDragMod") then
				-- We just dropped a body
				unit:getModifiers():remove("nct_guardDragMod", "dragCostMod")
			end
		end
	end
end

return {
	setModInfo = setModInfo,
	init = init,
	load = load,
	lateLoad = lateLoad,
	lateUnload = lateUnload
}
