local line_of_sight = include( "sim/line_of_sight" )
local simdefs = include( "sim/simdefs" )
local simquery = include( "sim/simquery" )

local function clockwise(x1, y1, x2, y2)
	return -x1*y2 + y1*x2 > 0
end

local function isPointInArc(x, y, x0, y0, arc0X, arc0Y, arc1X, arc1Y, distance)
	local vx = x - x0
	local vy = y - y0
	return clockwise(arc1X,arc1Y,vx,vy) and not clockwise(arc0X,arc0Y,vx,vy) and (vx * vx + vy * vy) <= distance * distance
end

local function shouldSightblock( sim, cell, dir )
	local dx, dy = simquery.getDeltaFromDirection( dir )
	local tocell = sim:getCell( cell.x + dx, cell.y + dy )
	return cell.exits[ dir ] ~= nil and (tocell == nil or not tocell.sightblock)
end

local function addWall( segments, cell, dir )
	local dx, dy = simquery.getDeltaFromDirection( dir )
	local nx, ny = dy, -dx
	table.insert( segments, cell.x + dx/2 - nx/2 )
	table.insert( segments, cell.y + dy/2 - ny/2 )
	table.insert( segments, cell.x + dx/2 + nx/2 )
	table.insert( segments, cell.y + dy/2 + ny/2 )
end

function line_of_sight:init( sim )
	local W, H = sim:getBoardSize()
	local staticSegs, dynamicSegs = {}, {}
	for y = 1, H do
		for x = 1, W do
			local cell = sim:getCell( x, y )
			if cell and cell.sightblock then
				if shouldSightblock( sim, cell, simdefs.DIR_N ) then
					addWall( staticSegs, cell, simdefs.DIR_N )
				end
				if shouldSightblock( sim, cell, simdefs.DIR_E ) then
					addWall( staticSegs, cell, simdefs.DIR_E )
				end
				if shouldSightblock( sim, cell, simdefs.DIR_S ) then
					addWall( staticSegs, cell, simdefs.DIR_S )
				end
				if shouldSightblock( sim, cell, simdefs.DIR_W ) then
					addWall( staticSegs, cell, simdefs.DIR_W )
				end
			end
			for i, dir in ipairs( simdefs.DIR_SIDES ) do
				if cell and cell.exits[ dir ] == nil then
					addWall( dynamicSegs, cell, dir )
				elseif cell and cell.exits[ dir ] and cell.exits[ dir ].door and cell.exits[ dir ].closed then
					addWall( dynamicSegs, cell, dir )
				end
			end
		end
	end
	local losmap = KLEISightMap.new()
	losmap:init( W, H )
	losmap:insertSegments( staticSegs, dynamicSegs )

	self.sim = sim
	self.losmap = losmap
end

function line_of_sight:hasSight( unit, x, y )
	if unit:isValid() then
		if unit:getTraits().ignoresWalls then
			local halfArc = simquery.getLOSArc( unit ) / 2
			local distance = unit:getTraits().LOSrange
			local x0, y0 = unit:getLocation()
			local facing = unit:getFacingRad()
			local arc0X = math.cos(facing - halfArc)
			local arc0Y = math.sin(facing - halfArc)
			local arc1X = math.cos(facing + halfArc)
			local arc1Y = math.sin(facing + halfArc)
		
			return isPointInArc(x, y, x0, y0, arc0X, arc0Y, arc1X, arc1Y, distance)
		else
			return self.losmap:hasSight( unit:getID(), x, y )
		end
	else
		return false
	end
end

function line_of_sight:hasPeripheral( unit, x, y )
	if unit:getTraits().LOSperipheralArc and unit:isValid() then
		return self.losmap:hasSight( unit:getID() + simdefs.SEERID_PERIPHERAL, x, y )
    	else
        	return false
	end
end