local simquery = include("sim/simquery")
local simdefs = include("sim/simdefs")
local inventory = include("sim/inventory")
local util = include("modules/util")
local abilityutil = include("sim/abilities/abilityutil")

local abilities = {
	APEX_PreviewItems = {
		name = STRINGS.ABILITIES.SCAN_DEVICE,
		profile_icon = "gui/icons/action_icons/Action_icon_Small/icon-item_hijack_small.png",
	    proxy = true,

		createToolTip = function( self, sim, abilityOwner, abilityUser, targetID )
			local targetUnit = sim:getUnit( targetID )
			return abilityutil.formatToolTip(STRINGS.ABILITIES.SCAN, string.format( STRINGS.ABILITIES.TOOLTIPS.SCAN, targetUnit:getName() ))
		end,
			
		getName = function( self, sim, unit )
			return STRINGS.ABILITIES.SCAN_DEVICE
		end,

	    getProfileIcon = function( self, sim, abilityOwner )
	        return abilityOwner:getUnitData().profile_icon or self.profile_icon
	    end,

	    onSpawnAbility = function(self, sim, abilityOwner)
			self.abilityOwner = abilityOwner
		end,

		acquireTargets = function( self, targets, game, sim, unit )
			-- guard targets
			local userUnit = unit:getUnitOwner()
			while not simquery.isAgent( userUnit ) do
				userUnit = userUnit:getUnitOwner()
				if not userUnit then
					break
				end
			end
			assert( userUnit, tostring(unit and unit:getName())..", "..tostring(unit:getLocation()) )
			local cells = sim._los:calculateUnitLOS(sim:getCell(userUnit:getLocation()), userUnit)
			local units = {}
			for i, cell in pairs(cells) do
				for _, cellUnit in ipairs( cell.units ) do
					if #cellUnit:getChildren() ~= 0 and (not simquery.isAgent(cellUnit) or cellUnit:hasTrait("isGuard")) then
						table.insert( units, cellUnit )
					elseif cellUnit:getTraits().credits and cellUnit:getTraits().credits ~= 0 then
						table.insert( units, cellUnit )
					end
				end
			end

			return targets.multiUnitTarget( game, units, self, unit, userUnit )
		end,

		targetingQualifier = function( self, target )
			-- ITEM targets
			-- returns true if the (potential) targets should be displayed
			-- e.g. Loba can't see corporate intel in large safes the way Ghuff can
			-- Loba should also *only* see items equal to anarchy level
			local userUnit = self.abilityOwner
			while not simquery.isAgent( userUnit ) do
				--log:write(tostring(userUnit:getName()))
				userUnit = userUnit:getUnitOwner()
				if not userUnit then
					break
				end
			end

			--log:write(tostring(userUnit:getName()))

			if not userUnit then
				return false
			end

			local inventory = include("sim/inventory")

			local result, reason = inventory.canCarry( userUnit, target )
			if reason and reason == STRINGS.UI.REASON.CANT_REACH then -- being out of reach is a non-issue
				result = true
			end

			return result
		end,
			
		canUseAbility = function( self, sim, unit, target )
			-- Must have a user owner.
			local userUnit = unit:getUnitOwner()
			while not simquery.isAgent( userUnit ) do
				userUnit = userUnit:getUnitOwner()
				if not userUnit then
					break
				end
			end

			if not userUnit then
				return false
			end

			return true
		end,

		executeAbility = function( self, sim, unit, userUnit, target )
			-- we don't actually need anything to happen, the ability is passive
			-- but it has targeting, so we're pretending to do something with the targets
			-- the real reason we're targeting is to slap tooltips around
		end,
	},

	APEX_ThrowBracelet = {
		name = STRINGS.APEX.ITEMS.BRACELET.THROW,

		getName = function( self, sim, unit, userUnit )
			return self.name
		end,
	
		createToolTip = function( self,sim,unit,targetCell)
			return abilityutil.formatToolTip( STRINGS.APEX.ITEMS.BRACELET.THROW,  STRINGS.APEX.ITEMS.BRACELET.THROW_DESC, 1 )
		end,
	
		profile_icon = "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png",
		usesAction = true,

		acquireTargets = function( self, targets, game, sim, grenadeUnit, unit)
			if not self:canUseAbility( sim, grenadeUnit, unit ) then
				return nil
			end
			return targets.throwTarget( game, grenadeUnit:getTraits().range or 0, sim, unit, unit:getTraits().maxThrow, grenadeUnit:getTraits().targeting_ignoreLOS)
		end, 


		canUseAbility = function( self, sim, grenadeUnit, unit, targetCell )
            if unit:getTraits().movingBody then
                return false, STRINGS.UI.REASON.DROP_BODY_TO_USE
            end

            local ownerUnit = grenadeUnit:getUnitOwner()
			if grenadeUnit:getTraits().pwrCost and (ownerUnit:getPlayerOwner():isPC() and ownerUnit:getPlayerOwner():getCpus() < grenadeUnit:getTraits().pwrCost) then
				return false, STRINGS.UI.REASON.NOT_ENOUGH_PWR
			end

			if grenadeUnit:getTraits().cooldown and grenadeUnit:getTraits().cooldown > 0 then
				return false, util.sformat(STRINGS.UI.REASON.COOLDOWN,grenadeUnit:getTraits().cooldown)
			end

			if targetCell then
				local targetX,targetY = unpack(targetCell)
				local unitX, unitY = unit:getLocation()
	    		local raycastX, raycastY = sim:getLOS():raycast(unitX, unitY, targetX, targetY)
				if raycastX ~= targetX or raycastY ~= targetY then
					return false
				end
			end

			return true
		end,

		executeAbility = function( self, sim, grenadeUnit, userUnit, targetCell )
			local sim = grenadeUnit:getSim()
			local x0,y0 = userUnit:getLocation()
			local x1,y1 = unpack(targetCell)
			userUnit:getTraits().throwing = true
		
			local facing = simquery.getDirectionFromDelta(x1-x0, y1-y0)
			simquery.suggestAgentFacing(userUnit, facing)

			if userUnit:isValid() and not userUnit:getTraits().interrupted then
				sim:dispatchEvent( simdefs.EV_UNIT_THROW, { unit = userUnit, x1=x1, y1=y1, facing=facing } )
	
				inventory.dropItem( sim, userUnit, grenadeUnit )

				if sim:isVersion("0.17.15") and userUnit:getTraits().data_hacking then 
					userUnit:stopHacking(sim)
		        end

				if grenadeUnit.throw then
					grenadeUnit:throw(userUnit, sim:getCell(x1, y1) )
				end

				sim:dispatchEvent( simdefs.EV_UNIT_STOP_THROW, { unitID = userUnit:getID(), x1=x1, y1=y1, facing=facing } )
				--local fx_params = {color ={{symbol="inner_line",r=0,g=1,b=1,a=0.75},{symbol="wall_digital",r=0,g=1,b=1,a=0.75},{symbol="boxy_tail",r=0,g=1,b=1,a=0.75},{symbol="boxy",r=0,g=1,b=1,a=0.75}} }		
				sim:dispatchEvent( simdefs.EV_UNIT_ADD_FX, { unit = userUnit, kanim = "fx/emp_effect", symbol = "character", anim="idle", above=true} )
				sim:warpUnit( userUnit, sim:getCell(x1, y1) )

				if grenadeUnit:getTraits().pwrCost then
					if sim:getPC() == userUnit:getPlayerOwner() then
						sim:getPC():addCPUs( -grenadeUnit:getTraits().pwrCost, sim, x1,y1)
					end
				end	

				sim:processReactions( userUnit )
			end
			userUnit:getTraits().throwing = nil
			if userUnit:isValid() and not userUnit:getTraits().interrupted then
				simquery.suggestAgentFacing(userUnit, facing)
			end
		end
	}
}

return abilities