-- implement a general-case adaption of Bresenham's line algorithm

local function plotLineLow(x0, y0, x1, y1, retCells)
    local dx = x1 - x0
    local dy = y1 - y0
    local yi = 1
    if dy < 0 then
        yi = -1
        dy = -dy
    end
    local D = 2 * dy - dx
    local y = y0

    for x = x0, x1 do
        table.insert(retCells, x)
        table.insert(retCells, y)
        if D > 0 then
            y = y + yi
            D = D - 2 * dx
        end
        D = D + 2 * dy
    end
end

local function plotLineHigh(x0, y0, x1, y1, retCells)
    local dx = x1 - x0
    local dy = y1 - y0
    local xi = 1
    if dx < 0 then
        xi = -1
        dx = -dx
    end
    local D = 2 * dx - dy
    local x = x0

    for y = y0, y1 do
        table.insert(retCells, x)
        table.insert(retCells, y)
        if D > 0 then
            x = x + xi
            D = D - 2 * dy
        end
        D = D + 2 * dx
    end
end

local function reverseXYarray(t)
    local x
    local y
    for i = 1, #t * 0.5, 2 do
        x = t[i]
        y = t[i + 1]
        t[i] = t[#t - i]
        t[i + 1] = t[#t - i + 1]
        t[#t - i] = x
        t[#t - i + 1] = y
    end
end

local function getCellsBetween(x0, y0, x1, y1)
    local retCells = {}

    local dx = x1 - x0
    local dy = y1 - y0
    if math.abs(dy) < math.abs(dx) then
        if x0 > x1 then
            plotLineLow(x1, y1, x0, y0, retCells)
            reverseXYarray(retCells)
        else
            plotLineLow(x0, y0, x1, y1, retCells)
        end
    else
        if y0 > y1 then
            plotLineHigh(x1, y1, x0, y0, retCells)
            reverseXYarray(retCells)
        else
            plotLineHigh(x0, y0, x1, y1, retCells)
        end
    end

    --remove the first and last cells
    table.remove(retCells, #retCells)
    table.remove(retCells, #retCells)
    table.remove(retCells, 1)
    table.remove(retCells, 1)

    return retCells
end

-- implement a raycast utility

local function isFloat(x)
    return x ~= math.floor(x)
end

local function roundNearestWhole(x)
    return math.floor(x + .5)
end

local function getCloserSideOfWall(wall_point, target_point)
    if isFloat(wall_point) ~= isFloat(wall_point*2) then -- ends in .5 => is wall
        return wall_point > target_point and math.floor(wall_point) or math.ceil(wall_point)
    end
end

local function getFartherSideOfWall(wall_point, target_point)
    local closer_point = getCloserSideOfWall(wall_point, target_point)
    if closer_point then
        return wall_point > target_point and closer_point + 1 or closer_point - 1
    end
end

local function getLastCellOfRaycast(x1, y1, xr, yr)
    local xrLim = mathutil.getFartherSideOfWall(xr, x1)
    local yrLim = mathutil.getFartherSideOfWall(yr, y1)
    local x2, y2 = x1, y1
    
    if xrLim then
        x2 = xrLim
        y2 = roundNearestWhole(yr)
    elseif yrLim then
        y2 = yrLim
        x2 = roundNearestWhole(xr)
    end

    return x2, y2
end

-- TODO trigonomagic for extending a direction for a set distance,
    -- or conversely getting the angle from two points
    -- (This isn't difficult per se, but it'd be nice to encapsulate the mathemagic
    -- away from the gameplay logic. Don't make me remember what atan2t means.)

return {
    getCellsBetween = getCellsBetween,
    getCloserSideOfWall = getCloserSideOfWall,
    getFartherSideOfWall = getFartherSideOfWall,
    getLastCellOfRaycast = getLastCellOfRaycast,
}
