local util = include( "modules/util" )
local array = include( "modules/array" )
local simdefs = include( "sim/simdefs" )
local serverdefs = include("modules/serverdefs")

local modApi = include( "mod-api" )

-------- GENERATION OPTIONS -------- 

function modApi:addGenerationOption( option, name, tip, data )
	assert( type(option) == "string" )
	assert( type(name) == "string" )
	assert( type(tip) == "string" )
	log:write( "mod_api[%s]:addGenerationOption( %s, %s )", self.mod_id, option, name )

	self.mod_manager:addGenerationOption(self.mod_id, option, name, tip, data )
end

-- Adds a function to be called when launching a campaign from the generation options screen (replaces the start-of-campaign cutscene and loading of the team selection screen)
--
-- fn: function
-- 		The handler function, should accept parameters of the form fn( stateGenerationOptions, difficulty, difficultyOptions, ... )

function modApi:addPostGenerationOptionsFunction( fn, ... )
	assert( type(fn) == "function" )
	log:write( "mod_api[%s]:addPostGenerationOptionsFunction", self.mod_id )
	if self.mod_manager.postGenerationOptionsFunction then
		log:write( "WARNING: Overwriting existing postGenerationOptionsFunction!" )
	end
	
	self.mod_manager.postGenerationOptionsFunction = { fn, { ... } }
end

-------- MISSION SCRIPTS -------- 

-- Appends the mission script for a specific mission type, using the same scripts format as modApi:addEscapeScripts
function modApi:addSituationScripts(missionType, scripts)
	assert(type(missionType) == "string")
	assert(type(scripts) == "table")

	self.mod_manager.modMissionScriptsTyped[missionType] = self.mod_manager.modMissionScriptsTyped[missionType] or {}
    table.insert( self.mod_manager.modMissionScriptsTyped[missionType], scripts )
end

-- Convenience function for adding a script hook funtion to the level script for the given mission type, with an optional hook name
function modApi:addScriptHook(missionType, hookName, hookFn)
	if type(missionType) == "table" then
		hookFn = missionType[3]
		hookName = missionType[2]
		missionType = missionType[1]
	end
	
	if type(hookName) == "function" then
		hookFn = hookName
		hookName = "Mod-hook"
	end
	
	assert(type(hookFn) == "function")
	assert(type(hookName) == "string")
	assert(type(missionType) == "string")

	self.mod_manager.modMissionScriptsTyped[missionType] = self.mod_manager.modMissionScriptsTyped[missionType] or {}
    table.insert(
		self.mod_manager.modMissionScriptsTyped[missionType],
		{
			addHooks = {
				[hookName] = hookFn
			}
		}
	)
end

-------- UI EVENTS -------- 

-- Adds a listener for a simevent
--
-- eventType: string or number
--		The type of event, such as simdefs.EV_UNIT_WARPED
--
-- fn: function
-- 		The handler function, should accept parameters of the form fn( event, eventType, eventData, boardRig, hud, thread )
--
-- captures: boolean (optional)
--		Whether this behaviour replaces vanilla behaviour (rather than append it)
--		Defaults to true for non-numerical event types and false for numerical event types, since all vanilla events are integers
 
function modApi:addVizEvHandler( eventType, fn, captures )
	assert( eventType )
	assert( type(fn) == "function" )
	assert( (captures == true) or (captures == false) or (captures == nil) )
	
	if captures == nil then
		-- Assume that numbered events are vanilla and should not be captured
		captures = type( eventType ) ~= "number"
	end
	
	table.insert( mod_manager.vizEventHandlers, { eventType, fn, captures } )
end

-------- SCREEN MODDING --------

function modApi:addNewUIScreen( filename, file )
	local screenModding = self.mod_manager.screenModding

	assert(type(filename) == "string")
	assert(screenModding.newFilesRaw[filename] == nil)
	assert(screenModding.newFilePaths[filename] == nil)
	
	if type(file) == "string" then
		screenModding.newFilePaths[filename] = file
	elseif type(file) == "table" then
		screenModding.newFilesRaw[filename] = file
	else
		assert(false)
	end
end

function modApi:insertUIElement( filename, path, element, index )
	assert(type(filename) == "string")
	assert(type(path) == "table")
	assert(index == nil or type(index) == "number")
	
	local inserts = self.mod_manager.screenModding.inserts
	inserts[filename] = inserts[filename] or {}
	table.insert(inserts[filename], {path, element, index})
end

function modApi:insertUIElementsAtPath( filename, path, elements, index )
	for i, element in ipairs( elements ) do
		self:insertUIElement( filename, path, element, index )
		
		if index and index > 0 then
			index = index + 1
		end
	end
end

function modApi:insertUIElements( elements )
	for i, element in ipairs( elements ) do
		self:insertUIElement( element[1], element[2], element[3], element[4] )
	end
end

function modApi:modifyUIElement( filename, path, modifications )
	assert(type(filename) == "string")
	assert(type(path) == "table")
	
	local changes = self.mod_manager.screenModding.changes
	changes[filename] = changes[filename] or {}
	table.insert(changes[filename], {path, modifications})
end

function modApi:modifyUIElements( elements )
	for i, element in ipairs( elements ) do
		self:modifyUIElement( element[1], element[2], element[3] )
	end
end

function modApi:registerWidgetFactory( name, widgetClass )
	assert(type(name) == "string")
	assert(type(widgetClass.init) == "function")
	mod_manager.widgetFactories[name] = widgetClass
end

-------- CROSS BANTER -------- 

function modApi:addBanter( banter )
	local cross = include('client/cross_scripts')
	local agents = {}
	for i, id in pairs(banter.agents) do
		agents[id] = {}
		for j, agentdef in pairs(include("sim/unitdefs/agentdefs")) do
			if agentdef.agentID == id then
				table.insert(agents[id],agentdef)
			end
		end
	end
	
	local dialogue = {}
	
	for i, dialog in pairs(banter.dialogue) do
		for _, agent in pairs(agents[dialog[1]]) do
			if not agent.speech.CROSS_DEFAULT then
				agent.speech.CROSS_DEFAULT = agent.speech.CROSS or {}
				agent.speech.CROSS = util.tcopy(agent.speech.CROSS_DEFAULT)
			end
			local strngs = agent.speech.CROSS
			strngs[#strngs+1]=dialog[2]
			dialogue[i] = {dialog[1],#strngs}
		end
	end

	table.insert(cross,{
		id = #cross + 1,
		agents = banter.agents,
		dialogue = dialogue,
	})
end

-------- ITEM DEFS --------

local function createStoreItemMap( mod_manager, list )
	local simstore = include( "sim/units/store" )
	if list == simstore.STORE_ITEM.itemList then
		log:write("Mapping store items")
	elseif list == simstore.STORE_ITEM.weaponList then
		log:write("Mapping store weapons")
	else
		log:write("Mapping store augments")
	end
	
	mod_manager.storeItemMap[list] = {}
	
	for i,listitem in ipairs(list)do
		mod_manager.storeItemMap[list][listitem.id] = i
	end
end

local function findItemInList( mod_manager, item, list )
	if not mod_manager.storeItemMap[list] then
		createStoreItemMap( mod_manager, list )
	end
	
	local index = mod_manager.storeItemMap[list][item.id]
	if index then
		if list[index] and list[index].id == item.id then
			return index
		else
			createStoreItemMap( mod_manager, list )
			return mod_manager.storeItemMap[list][item.id]
		end
	end
end

local function addItemToStore( mod_manager, item, list )
	local index = findItemInList( mod_manager, item, list )
	if index then			
		list[index] = item
	else
		table.insert(list, item)
		mod_manager.storeItemMap[list][item.id] = #list
	end
end

local function isReplacement( mod_manager, item, list )
	if not mod_manager.storeItemMap[ list ] then
		createStoreItemMap( mod_manager, list )
	end
	
	return mod_manager.storeItemMap[ list ][ item.id ]
end

function modApi:addItemDef( name, itemDef )
	assert( type(name) == "string" )
	assert( type(itemDef) == "table" )

	local itemdefs = include( "sim/unitdefs/itemdefs" )
	itemdefs[ name ] = itemDef
	itemdefs[ name ].id = name

	local simstore = include( "sim/units/store" )
	
	local itemListItem = itemDef.ITEM_LIST or isReplacement( self.mod_manager, itemDef, simstore.STORE_ITEM.itemList )
	local weaponListItem = itemDef.WEAPON_LIST or isReplacement( self.mod_manager, itemDef, simstore.STORE_ITEM.weaponList )
	local augmentListItem = itemDef.AUGMENT_LIST or isReplacement( self.mod_manager, itemDef, simstore.STORE_ITEM.augmentList )

	if itemListItem then
		addItemToStore( self.mod_manager, itemDef, simstore.STORE_ITEM.itemList )
	end

	if weaponListItem then	  
		addItemToStore( self.mod_manager, itemDef, simstore.STORE_ITEM.weaponList )
	end
	
	if augmentListItem then
		addItemToStore( self.mod_manager, itemDef, simstore.STORE_ITEM.augmentList )
	end   

	if self._dlc_footer then
		itemdefs[ name ].traits.dlcFooter = self._dlc_footer
	end
end

-------- AGENT DEFS -------- 

-- Add agent to the list of selectable agent loadouts
-- mainAgentId
--		mainAgentId : string
--			the template name of the agent this loadout is added to e.g. "stealth_1"
--      loadoutId : string
--          the template name of the new loadout e.g. "stealth_1_but_cooler"

function modApi:addAgentLoadout( mainAgentId, loadoutId )
	assert( type(mainAgentId) == "string" )
	assert( type(loadoutId) == "string" )
	
	local serverdefs = include( "modules/serverdefs" )
	
    if array.find( serverdefs.SELECTABLE_AGENTS, mainAgentId ) == nil then
        table.insert( serverdefs.SELECTABLE_AGENTS, mainAgentId )
    end
	if not serverdefs.LOADOUTS[ mainAgentId ] then
		serverdefs.LOADOUTS[ mainAgentId ] = {}
	end
	
    table.insert( serverdefs.LOADOUTS[ mainAgentId ], loadoutId )
end

-- Add agents to the list of rescueable agents
-- unitDef: table
--      id : int
--          agentID from agentDefs
--      template : string
--          agents templateName eg: "sharpshooter_1"
--      upgrades : table
--          table of strings that are the item IDs    

function modApi:addRescueAgent( unitDef, template, upgrades )
    local serverdefs = include( "modules/serverdefs" )
	local t1 = type( unitDef )
	local t2 = type( template )
	
    if t1 == "table" then
		-- ( serverdefs.createAgent(...) )
		assert( unitDef.id, "Agent rescueable does not have an id" )
		assert( unitDef.template, "Agent rescueable does not have a template" )
		unitDef.upgrades = unitDef.upgrades or {}
		table.insert( serverdefs.TEMPLATE_AGENCY.unitDefsPotential, unitDef )
	elseif t2 == "table" or not template then
		-- ( template, upgrades )
		assert( t1 == "number" or t1 == "string", "addRescueAgent( template, upgrades ) parameter #1 is "..t1 )
		
		local agentdefs = include("sim/unitdefs/agentdefs")
		local agentDef = agentdefs[unitDef]
		
		assert( agentDef, "No agent ("..tostring(unitDef)..") found (did you remember to load the agent before adding them as a rescueable?)" )
		
		table.insert( serverdefs.TEMPLATE_AGENCY.unitDefsPotential, {
			id = agentDef.agentID,
			template = unitDef,
			upgrades = template or {}
		} )
	else
		-- ( agentID, template, upgrades )
		assert( t1 == "number" or t1 == "string", "addRescueAgent( agentID, template, upgrades ) parameter #1 is "..t1 )
		assert( t2 == "number" or t2 == "string", "addRescueAgent( agentID, template, upgrades ) parameter #2 is "..t2 )
		table.insert( serverdefs.TEMPLATE_AGENCY.unitDefsPotential, {
			id = unitDef,
			template = template,
			upgrades = upgrades or {}
		} )
	end
end

-------- MAP SCREEN --------

function modApi:addWorldPrefabt( path, world, prefabs )				
	self.mod_manager.modWorldPrefabs[world] = self.mod_manager.modWorldPrefabs[world] or {PREFABT0 = {}}
	for i, prefab in pairs(prefabs.PREFABT0) do
		table.insert(self.mod_manager.modWorldPrefabs[world].PREFABT0,prefab)
	end
end

function modApi:addCorpMap( corp, territory, normal_color, highlight_color, unhighlight_color, firstMission )
	assert(type(corp) == "string")
	assert(type(territory) == "table")
	assert(type(normal_color) == "table")
	assert(type(highlight_color) == "table")
	assert(type(unhighlight_color) == "table")
	assert(type(firstMission) == "boolean")
	simdefs.map_colours_normal[corp] = normal_color
	simdefs.map_colours_highlight[corp] = highlight_color
	simdefs.map_colours_unhighlight[corp] = unhighlight_color
	simdefs.map_territories[corp] = territory.name
	
	self:insertUIElement( "map_screen.lua", { "widgets", 4, "children", 1, "children", 1, "children" }, territory )

	table.insert(serverdefs.CORP_NAMES,corp)
	table.insert(simdefs.DEFAULT_MISSION_TAGS,corp)
	if firstMission then
		table.insert(simdefs.DEFAULT_MISSION_TAGS_FIRST,corp)
	end
end

function modApi:addAdvanceTimeCheck( check )
	assert(type(check) == "table")
	assert(type(check.fn) == "function")
	assert(type(check.stage) == "number")
	table.insert(self.mod_manager.advanceTimeEvents,check)
end