local util = include( "modules/util" )
local filesystem = include("modules/filesystem")
local array = include( "modules/array" )
local simdefs = include( "sim/simdefs" )
local serverdefs = include("modules/serverdefs")

local mod_manager = {}

-------- SEQUENTIAL MOD LOADER -------- 

local function orderMod( mod_manager, mod, order, traversed, sorted )
	if not traversed[mod.id] then
		traversed[mod.id] = true
		local modData = mod_manager:findMod( mod.id )
		if modData and modData.api and modData.installed and not modData.locale then
			if modData.api.requirements then
				for _, requiredMod in ipairs( sorted ) do
					local name = requiredMod.name and util.toupper(requiredMod.name)
					local id = requiredMod.id and util.toupper(requiredMod.id)
					
					for i, requirement in ipairs( modData.api.requirements ) do
						local upRequirement = requirement and util.toupper(requirement)
						
						if upRequirement == name or upRequirement == id then
							orderMod( mod_manager, requiredMod, order, traversed, sorted )
							break
						end
					end
				end
			end
			table.insert( order, modData )
		end
	end
end

local function orderMods( mod_manager, mods )
	local order = {}
	local traversed = {}
	local sorted = {}
	
	for id, mod in pairs(mods) do
		local skip = false
		mod.id = mod.id or id
		for i, other in ipairs(sorted) do
			if mod.id < other.id then
				skip = true
				table.insert(sorted,i,mod)
				break
			end
		end
		if not skip then
			table.insert(sorted,mod)
		end
	end
	
	for _, mod in ipairs( sorted ) do
		orderMod( mod_manager, mod, order, traversed, sorted )
	end
	return order
end

-------- SCREEN MODDING -------- 

mod_manager.widgetFactories = {}

function mod_manager:resetScreenModding()
	self.screenModding = {
		changes = {},
		newFilesRaw = {},
		newFilePaths = {},
		inserts = {}
	}
end

mod_manager:resetScreenModding()

-------- GENERATION OPTIONS -------- 

function mod_manager:addGenerationOption( mod_id, option, name, tip, data )
    local modData = self:findMod(mod_id)
	local data = data or {}
	if data.enabled == nil then
		data.enabled = true
	end
	if data.values then
		data.value = data.value or data.values[1]
	end
    if not self.generationOptions[mod_id] then
        self.generationOptions[mod_id] = { name = self:getModName(mod_id), id = mod_id, enabled = true, options = {}, icon=self:getModIcon(mod_id) }
    end

    local new_opt = util.extend(data){ option = option, name = name, tip = tip }
    table.insert(self.generationOptions[mod_id].options, new_opt)
end

function mod_manager:getModContentDefaults(difficulty)
    --log:write("mod_manager:getModContentDefaults")

    local options = util.tcopy(self.generationOptions)

    -- convert from array (for order) to keyed (for save game)
    for mod_id, mod_info in pairs(options) do
        local keyed_options = {}
        for i, opt_info in ipairs(mod_info.options) do
            keyed_options[opt_info.option] = { enabled = opt_info.enabled, value = opt_info.value }
			if difficulty and opt_info.difficulties then
				for i, difficultyOpt in pairs(opt_info.difficulties) do
					if difficultyOpt[1] == difficulty then
						keyed_options[opt_info.option].enabled = difficultyOpt[2]
						keyed_options[opt_info.option].value = difficultyOpt[2]
						break
					end
				end
			end
        end
        mod_info.options = keyed_options
    end

    --log:write("   %s", util.stringize(options))

    return options
end

-------- CROSS BANTER -------- 

local cross_default

function mod_manager:resetCrossBanter()
	cross_default = cross_default or util.tcopy(include('client/cross_scripts'))
	for i, agentdef in pairs(include("sim/unitdefs/agentdefs")) do
		agentdef.speech.CROSS_DEFAULT = agentdef.speech.CROSS_DEFAULT or agentdef.speech.CROSS or {}
		agentdef.speech.CROSS = util.tcopy(agentdef.speech.CROSS_DEFAULT)
	end
	util.tclear(include('client/cross_scripts'))
	util.tmerge(include('client/cross_scripts'),cross_default)
end

mod_manager:resetCrossBanter()

-------- MOD LOADING -------- 

function mod_manager:mountContentMod( id )
	if not self.mountedAllMods then
		self.mountedAllMods = true
		self.storeItemMap = {}
		for i, modData in ipairs(self.mods) do
			if not modData.locale then
				if not modData then
					log:write( "Could not mount missing content mod: '%s'", tostring(modData.id) )
				elseif not MOAIFileSystem.checkFileExists( filesystem.pathJoin( modData.folder, "scripts.zip" )) then
					log:write( "Could not mount content mod without scripts.zip: '%s'", tostring(modData.id) )
				else
					-- Mount the content archive.
					local scriptsArchive = string.format( "%s/scripts.zip", modData.folder, modData.scripts )
					local scriptsAlias = scriptsArchive:match( "/([-_%w]+)/scripts[.]zip$" )

					log:write( "Mounting content mod [%s] scripts at: [%s]", tostring(scriptsArchive), tostring(scriptsAlias)  )
					MOAIFileSystem.mountVirtualDirectory( scriptsAlias, scriptsArchive )
					
					modData.name = modData.name or "ID:"..modData.id

					local initFile = string.format( "%s/modinit.lua", scriptsAlias )
					log:write( "\tExecuting '%s':", initFile )
					local ok, mod = pcall( dofile, initFile )
					if ok then
						modData.modfn = mod
                     
						local modapi = include( "mod-api" )
						modData.api = modapi( self, modData.id, modData.folder, scriptsAlias )
						modData.installed = true
						
						if modData.name == "Rescueable Archive Agents" then
							modData.api:addGenerationOption( "multiRescue", "SEPARATE ARCHIVE RESCUABLES", "Turning this <c:73d216>on</c> makes it possible for detention centers to have different versions of agents you already have.\nThat is, you can rescue the archive version of one of your on-file agents or vice versa.", { noUpdate = true, enabled = false } )
							mod.load = include(self._raaHack)
						end
					else
						modData.installed = false
						log:write( "mod.init ERROR: %s", tostring(mod) )
					end
				end
			end
		end
		
		local order = orderMods( self, self.mods )
		
		for i, mod in ipairs(order) do
			if not mod.locale and mod.installed and mod.modfn.earlyInit then
				log:write( "Early initiating mod %s, with id: %s", mod.name, mod.id )
				local ok, res = pcall(function()
					mod.modfn.earlyInit( mod.api )
				end)
				if ok then
					log:write( "\tEARLY MOD-INIT OK")
					mod.installed = true
				else
					log:write( "\tEARLY MOD-INIT FAILED: %s", tostring(res))
					mod.installed = false
				end
			end
		end
		
		local order = orderMods( self, self.mods )
		
		for i, mod in ipairs(order) do
			if not mod.locale and mod.installed and mod.modfn.init then
				log:write( "Initiating mod %s, with id: %s", mod.name, mod.id )
				local ok, res = pcall(function()
					mod.modfn.init( mod.api )
				end)
				if ok then
					log:write( "\tMOD-INIT OK")
					mod.installed = true
				else
					log:write( "\tMOD-INIT FAILED: %s", tostring(res))
					mod.installed = false
				end
			end
		end
		
		local order = orderMods( self, self.mods )
		
		for i, mod in ipairs(order) do
			if not mod.locale and mod.installed and mod.modfn.lateInit then
				log:write( "Late initiating mod %s, with id: %s", mod.name, mod.id )
				local ok, res = pcall(function()
					mod.modfn.lateInit( mod.api )
				end)
				if ok then
					log:write( "\tLATE MOD-INIT OK")
					mod.installed = true
				else
					log:write( "\tLATE MOD-INIT FAILED: %s", tostring(res))
					mod.installed = false
				end
			end
		end
		
		for i, mod in pairs(self.mods) do
			if not self.generationOptions[mod.id] then
				self.generationOptions[mod.id] = { name = self:getModName(mod.id), enabled = true, options = {}, icon=self:getModIcon(mod.id) }
			end
			if not mod.installed then
				self.generationOptions[mod.id].enabled = false
			end
		end
	end
end

function mod_manager:resetAgentSubLoadouts()
	log:write("Deep resetting agent loadouts for safety")
	for agentID, loadout in pairs(serverdefs.LOADOUTS) do
		serverdefs.LOADOUTS[agentID] = util.tcopy( serverdefs.LOADOUTS[agentID] )
	end
end

function mod_manager:resetContent()
    self:defaultResetContent()
	log:write( "Reset Sim Constructor" )
	
	self:resetAgentSubLoadouts()
	self.resetSimDefs()
	self:resetScreenModding()
	
	serverdefs.CORP_NAMES = { "ftm", "ko", "plastech", "sankaku" }
	serverdefs.MISSION_TAGS_INITIAL_EXEC = { "executive_terminals", "ftm", "ko", "plastech", "not_close" }
	
	if self.modWorldPrefabsDefault == nil then
		self.modWorldPrefabsDefault = {}
		for world, prefabt in pairs(serverdefs.worldPrefabts) do
			self.modWorldPrefabsDefault[world] = {PREFABT0 = util.tdupe(prefabt.PREFABT0)}
		end
	end
	
	self.modWorldPrefabs = {}
	self.advanceTimeEvents = {}
	self.postGenerationOptionsFunction = nil
	
	for world, prefabt in pairs(self.modWorldPrefabsDefault) do
		self.modWorldPrefabs[world] = {PREFABT0 = util.tdupe(prefabt.PREFABT0)}
	end
	
	self.modPrefabs = {}
	self.modMissionScripts = {}
	self.modMissionScriptsTyped = {}
	self.vizEventHandlers = {}
end

function mod_manager:loadModContent( mod_options, params )
	local order = orderMods( self, self:getModContentDefaults() )
	mod_options = mod_options or {}
	
	-- Early load
	for _, modData in ipairs(order) do
		local mod = mod_options[modData.id]
		if mod and mod.enabled then
			if modData.modfn.earlyLoad then
				log:write( "Early loading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.earlyLoad( modData.api, mod.options, params, mod_options )
			end
		else
			if modData.modfn.earlyUnload then
				log:write( "Early unloading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.earlyUnload( modData.api, mod_options, params )
			end
		end
	end
	-- Standard load
	for _, modData in ipairs(order) do
		local mod = mod_options[modData.id]
		if mod and mod.enabled then
			if modData.modfn.load then
				log:write( "Loading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.load( modData.api, mod.options, params, mod_options )
			end
		else
			if modData.modfn.unload then
				log:write( "Unloading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.unload( modData.api, mod_options, params )
			end
		end
	end
	-- Late load
	for _, modData in ipairs(order) do
		local mod = mod_options[modData.id]
		if mod and mod.enabled then
			if modData.modfn.lateLoad then
				log:write( "Late loading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.lateLoad( modData.api, mod.options, params, mod_options )
			end
		else
			if modData.modfn.lateUnload then
				log:write( "Late unloading mod %s, with id: %s", modData.name, modData.id )
				modData.modfn.lateUnload( modData.api, mod_options, params )
			end
		end
	end
	
	local old_upgrade_templates = include( "sim/upgradedefs" ).upgrade_templates
	package.loaded["sim/upgradedefs"] = nil
	local new_upgrade_templates = include( "sim/upgradedefs" ).upgrade_templates
	for k,v in pairs(old_upgrade_templates) do
		if not new_upgrade_templates[k] then
			new_upgrade_templates[k] = v
		end
	end
	
	self.storeItemMap = {}
	for id, template in pairs(include("sim/unitdefs/itemdefs")) do
		template.id = id
	end
	log:write("SEQUENTIAL MOD LOADER: SUCCESS")
end

return mod_manager