local modifiers = include("sim/modifiers")

local BASE = "." -- Base value of trait
local ADD = '+' -- Adds to a numeric trait
local MULT = '*' -- Multiplies a numeric trait 
local CLA = '~' -- Clamps a numeric trait 
local SET = '=' -- Sets a trait
local FN = 'f' -- Custom function

--Operation Priority
--Note that addition goes first, as I imagine addition is used for minor buffs while the others are for more powerful effects

local OPS = {
	[ADD] = 4,
	[MULT] = 3,
	[CLA] = 2,
	[SET] = 1,
	[FN] = 0,
}

modifiers.OPS = OPS
modifiers.CLA = CLA
modifiers.FN = FN

function modifiers:eval( trait )
	local val = nil
	for i, modifier in ipairs( self.modifiers[ trait ] ) do
		if modifier.op == BASE or modifier.op == SET then
			val = modifier.val
		elseif modifier.op == FN then
			val = modifier.val( val, modifier--[[, false]] )
		elseif type(val) == "number" then
			local prev = val
			if modifier.op == ADD then
				val = val + modifier.val
			elseif modifier.op == MULT then
				val = val * modifier.val
			end
			
			if modifier.min and (val < prev or modifier.op == CLA) and val < modifier.min then
				val = modifier.min
			end
			if modifier.max and (val > prev or modifier.op == CLA) and val > modifier.max then
				val = modifier.max
			end
		end
	end
	return val
end

function modifiers:reeval( trait )
	self.traits[ trait ] = self:eval( trait )
end

function modifiers:has( trait, name )
	if trait then
		local modifiers = self.modifiers[ trait ]
		if modifiers then
			for i = #modifiers, 1, -1 do
				if modifiers[i] == name or modifiers[i].name == name then
					return modifiers[i], i--Now returns the modifier instance
				end
			end
		end
	else
		for trait, modifiers in pairs(self.modifiers) do
			for i = #modifiers, 1, -1 do
				if modifiers[i] == name or modifiers[i].name == name then
					return modifiers[i], i--Now returns the modifier instance
				end
			end
		end
	end

    return false
end

function modifiers:hasAny( trait )
	local modifiers = self.modifiers[ trait ]
	if modifiers then
		return #modifiers > 0
	end
end

function modifiers:change( trait, name, op, val, min, max, priority, opPriority )
	local modifier, i = self:has( trait, name )
	
	if modifier then
		if op then
			modifier.op = op
		end
		if val then
			modifier.val = val
		end
		if min then
			modifier.min = min
		end
		if max then
			modifier.max = max
		end
		if priority then
			modifier.pr = priority
		end
		if opPriority then
			modifier.opPr = opPriority
		end
		
		if priority or opPriority then
			table.remove( self.modifiers[modifier.trait], i )
			self:insert( modifier.trait, modifier )
		else
			self:reeval( modifier.trait )
		end
		return modifier
	end
end

--Greater than (>) operator for modifiers
function modifiers:compare( mod1, mod2 )
	if mod2.op == BASE then
		return false
	elseif mod1.pr > mod2.pr then
		return true
	elseif mod1.pr < mod2.pr then
		return false
	elseif mod1.opPr > mod2.opPr then
		return true
	end
	
	return false
end

function modifiers:add( trait, name, op, val, min, max, priority, opPriority )
	--Removed duplicate modifier assertion
	
	local modifier = { name = name, op = op, val = val, trait = trait, min = min, max = max, pr = priority or 0, opPr = opPriority or OPS[op] }
	if self.modifiers[ trait ] == nil then
		self.modifiers[ trait ] = { { op = BASE, val = self.traits[ trait ], trait = trait } }
	end
	
	return self:insert( trait, modifier )
end

function modifiers:insert( trait, modifier )
	local idx = #self.modifiers[ trait ] + 1
	--Ordered modifiers
	for i, compareTo in ipairs( self.modifiers[ trait ] ) do
		if self:compare( modifier, compareTo ) then
			idx = i
			break
		end
	end
	
	table.insert( self.modifiers[ trait ], idx, modifier )
	self:reeval( trait )
	return modifier
end

--You can now specify trait to narrow the search, and also insert the actual modifier instead of its name
function modifiers:remove( name, trait )
	assert( name )
	local count = 0
	if trait then
		local modifiers = self.modifiers[trait]
		if modifiers then
			for i = #modifiers, 1, -1 do
				if modifiers[i] == name or modifiers[i].name == name then
					table.remove( modifiers, i )
					count = count + 1
					break--Added this break, assuming that when there are duplicates we want to remove one at a time
				end
			end
			
			if count > 0 then
				self:reeval( trait )
				if #modifiers == 1 then
					self.modifiers[ trait ] = nil
				end
			end
		end
	else
		for trait, modifiers in pairs( self.modifiers ) do
			for i = #modifiers, 1, -1 do
				if modifiers[i] == name or modifiers[i].name == name then
					table.remove( modifiers, i )
					count = count + 1
					break--Added this break, assuming that when there are duplicates we want to remove one at a time
				end
			end
			
			if count > 0 then
				self:reeval( trait )
				if #modifiers == 1 then
					self.modifiers[ trait ] = nil
				end
			end
		end
	end
	
	return count > 0
end
