local context = include( "sim/procgen_context" )
local simdefs = include( "sim/simdefs" )
local unitdefs = include( "sim/unitdefs" )
local util = include( "modules/util" )

function context:scaleCredits( amount, params )
    local scalar = params.difficultyOptions.credit_scalar or 0.25
	local maxscalar = params.difficultyOptions.credit_ceil or 4
	
	local s1 = 1 + (math.min(maxscalar,params.difficulty) - 1) * scalar
    local s2 = params.difficultyOptions.creditMultiplier

    return math.floor( amount * s1 * s2 )
end

function context:hasCPUs( unit )
	local templateUnitData = unitdefs.lookupTemplate( unit.template )
	assert( templateUnitData, unit.template )
	return templateUnitData.traits.maxcpus ~= nil
end

function context:hasICE( unit )
	local templateUnitData = unitdefs.lookupTemplate( unit.template )
	assert( templateUnitData, unit.template )
	if not templateUnitData.traits.mainframe_console and not templateUnitData.traits.mainframe_ice_set_in_def  then
		return templateUnitData.traits.mainframe_iceMax ~= nil
	else 
		return false 
	end 
end

function context:fillCPUs( unit )
	local templateUnitData = unitdefs.lookupTemplate( unit.template )
	local cpus = math.ceil( unit.unitData.traits.lootValue )
	unit.unitData.traits.cpus = cpus
	--log:write( "PWR - <%d, %d> %d/%d", unit.x, unit.y, cpus, unit.unitData.traits.cpus )
end

function context:hasLoot( unit )
	local templateUnitData = unitdefs.lookupTemplate( unit.template )
	assert( templateUnitData, unit.template )
	return templateUnitData.lootTable ~= nil
end

function context:shouldChooseLoot( lootValue, valueLeft )
	if valueLeft <= 0 then
		-- Always bail in this case.
		return false
	end

	local rnd = self.rnd:next()
	-- Chance to pick this item is based on its value and the valueLeft.
	local p = math.min( 1, (valueLeft / lootValue) * (valueLeft / lootValue) )
	return rnd <= p
end

function context:fillLoot( unit )
	if unit.unitData.children == nil then
		unit.unitData.children = {}
	end
	local templateUnitData = unitdefs.lookupTemplate( unit.template )
	local lootTable = util.weighted_list( simdefs.LOOT_TABLES[ templateUnitData.lootTable ])
	local lootValue = 0
	local lootCount = 0
	local unitLootValue = math.ceil(unit.unitData.traits.lootValue / 10) * 10

	while lootTable:getCount() > 0 and lootCount < simdefs.MAX_ITEMS do
		local j = self.rnd:nextInt( 1, lootTable:getTotalWeight() )
		local lootItem = lootTable:removeChoice( j )
		if lootItem == nil then
			break
		end
		if self:shouldChooseLoot( type(lootItem) == "number" and lootItem or type(lootItem) == "string" and unitdefs.lookupTemplate( lootItem ).value or lootItem.value, unitLootValue - lootValue ) then
			if type(lootItem) == "number" then
				unit.unitData.traits.credits = (unit.unitData.traits.credits or 0) + lootItem
				lootValue = lootValue + lootItem
			elseif type(lootItem) == "string" then
				table.insert( unit.unitData.children, unitdefs.lookupTemplate( lootItem ) )
				lootValue = lootValue + unitdefs.lookupTemplate( lootItem ).value
			else
				table.insert( unit.unitData.children, lootItem )
				lootValue = lootValue + lootItem.value
				--log:write( "%s <%d, %d> - %s (%d total val, %d target)", unit.template, unit.x, unit.y, lootItem.name, lootValue, unitLootValue )
			end
			lootCount = lootCount + 1
		end
	end
	unit.unitData.traits.credits = (unit.unitData.traits.credits or 0) + math.max( 0, unitLootValue - lootValue )
	--log:write( "%s <%d, %d> - %d credits", unit.template, unit.x, unit.y, unit.unitData.traits.credits )
end

function context:assignLoot()
	-- Find total loot value.
	local totalLootWeight = 0
	local lootUnits = {}
	for i, unit in ipairs( self.units ) do
		if self:hasLoot( unit ) then
			local room = self:roomContaining( unit.x, unit.y )
			unit.lootWeight = room.lootWeight + (unit.unitData and unit.unitData.lootWeight or 0)
			totalLootWeight = totalLootWeight + unit.lootWeight
			table.insert( lootUnits, unit )
		end
	end

	-- Sort loot from great value to least.
	table.sort( lootUnits, function( lhs, rhs ) return lhs.lootWeight > rhs.lootWeight end )
	local TIER2_IDX = math.ceil( #lootUnits * simdefs.TIER2_IDX )
	local TIER2_VALUE = self:scaleCredits( simdefs.TIER2_VALUE, self.params )
	local TIER1_VALUE = self:scaleCredits( simdefs.TIER1_VALUE, self.params )
	
	-- Generate the actual lootage.
	for i, unit in ipairs( lootUnits ) do



		if unit.unitData == nil then
			unit.unitData = {}
		end
		if unit.unitData.traits == nil then
			unit.unitData.traits = {}
		end
		if i <= TIER2_IDX then
			if unit.template:find("_safe") ~= nil then
                unit.template = unit.template .. "_tier2"
			end
			unit.unitData.traits.lootValue = TIER2_VALUE
		else
			unit.unitData.traits.lootValue = TIER1_VALUE
		end

		self:fillLoot( unit )

		-- One intel per mission.. in the most difficult safe.		
		if i==1 then
			if not self.params.world == "omni" then
				self.numIntel = self.numIntel + 1
				table.insert( unit.unitData.children, unitdefs.prop_templates.item_corpIntel )
			end 
		end
	end
end

function context:assignICE( difficulty )
	local iceLow = simdefs.ICE_LOWS[self.params.difficultyOptions.ice_lows or "NORMAL"][difficulty]
	local iceHigh = simdefs.ICE_HIGHS[self.params.difficultyOptions.ice_highs or "NORMAL"][difficulty]
	local magneticChance = simdefs.MAGNETIC_REINFORCEMENTS_CHANCE[self.params.difficultyOptions.magnetic_reinforcement_chance or "NORMAL"][difficulty]

	local boostedIce = math.floor( simdefs.NUM_BOOSTED_ICE[self.params.difficultyOptions.firewallWeight or "NORMAL"][difficulty] * ((not self.params.difficultyOptions.firewallWeight) and self.params.difficultyOptions.firewallMultiplier or 0.5) )

	local iceUnits = {}
	for i, unit in ipairs( self.units ) do
		if self:hasICE( unit ) then
			table.insert( iceUnits, unit )
		end
	end

	for i, unit in ipairs( iceUnits ) do

		if unit.unitData == nil then
			unit.unitData = {}
		end
		if unit.unitData.traits == nil then
			unit.unitData.traits = {}
		end

    	local templateUnitData = unitdefs.lookupTemplate( unit.template )
        if templateUnitData.traits.mainframe_ice <= iceLow then
		    if boostedIce > 0 and self.rnd:nextInt(1, 3) == 1 then 
			    boostedIce = boostedIce - 1 
			    unit.unitData.traits.mainframe_iceMax = iceHigh 
		    else
			    unit.unitData.traits.mainframe_iceMax = iceLow 
		    end

            if templateUnitData.traits.mainframe_iceBonus then
                unit.unitData.traits.mainframe_iceMax = unit.unitData.traits.mainframe_iceMax + templateUnitData.traits.mainframe_iceBonus
            end

            if self.rnd:next() < magneticChance then
            	unit.unitData.traits.magnetic_reinforcement = true
            end

            unit.unitData.traits.mainframe_ice = unit.unitData.traits.mainframe_iceMax 
        end
	end
end

function context:assignCPUs( totalValue )
	-- Find total loot value.
	local totalLootWeight = 0
	local lootUnits = {}
	for i, unit in ipairs( self.units ) do
		if self:hasCPUs( unit ) then
			local room = self:roomContaining( unit.x, unit.y )
			unit.lootWeight = room.lootWeight
			totalLootWeight = totalLootWeight + unit.lootWeight
			table.insert( lootUnits, unit )
		end
	end
	
	-- Generate the actual lootage.
	for i, unit in ipairs( lootUnits ) do
		if unit.unitData == nil then
			unit.unitData = {}
		end
		if unit.unitData.traits == nil then
			unit.unitData.traits = {}
		end
		unit.lootWeight = (unit.lootWeight / totalLootWeight)
		unit.unitData.traits.lootValue = (unit.unitData.traits.lootValue or 0) + unit.lootWeight * totalValue

		self:fillCPUs( unit )
	end
end

function context:finalize()
	self.board.walltop_color = { 0.2, 0.2, 0.2, 1 }

	-- Assign da phat loots.
	self:assignLoot()
	self:assignCPUs( self.params.difficultyOptions.powerPerLevel )
	self:assignICE( self.difficulty )
	
	if (MOAIEnvironment.OS == "PS4" or MOAIEnvironment.OS == "XBO" or MOAIEnvironment.OS == "NX") and MOAIEnvironment.Preload == 1 then
		-- Analyze all for preload
		preloader.clear()
		preloader.analyzeUnits(self.units)
		preloader.analyzeDecos(self.decos)
		preloader.analyzeLights(self.lights)
		preloader.analyzeRooms(self.rooms)
		preloader.analyzeCandidates(self.candidates)
		preloader.analyzePrograms(self.ice_programs)
		preloader.analyzeGuards(self.params)
	end

	return self:invokeScriptGen( "finalizeProcgen" ) or 1 -- 1 is a passing metric
end