local context = include( "sim/procgen_context" )
local prefabs = include( "sim/prefabs" )
local roomgen = include( "sim/roomgen" )
local simdefs = include( "sim/simdefs" )
local unitdefs = include( "sim/unitdefs" )
local util = include( "modules/util" )
local rand = include( "modules/rand" )

local oldinit = context.init
local oldfinalize = context.finalize

local function defaultFitnessFn( cxt, prefab, x, y )
    local tileCount = cxt:calculatePrefabLinkage( prefab, x, y )
    if tileCount == 0 then
        return 0 -- Doesn't link up
    else
        -- This is a *selection weighting*, so boost fitness based on tileCount exponentially.
 	    return 1 + math.pow( tileCount, 3 )
    end
end

function context:init( params, pass )
	local random_consoles = false
	local random_safes = false
	self.genOptsRand = rand.createGenerator( (params.seed + pass * 101 ) % 2^32 )
	if params.difficultyOptions.consolesPerLevel == "RANDOM" then
		random_consoles = true
		params.difficultyOptions.consolesPerLevel = self.genOptsRand:nextInt(math.min(params.difficultyOptions.min_consolesPerLevel,params.difficultyOptions.max_consolesPerLevel),math.max(params.difficultyOptions.min_consolesPerLevel,params.difficultyOptions.max_consolesPerLevel))
	end	
	if params.difficultyOptions.safesPerLevel == "RANDOM" then
		random_safes = true
		params.difficultyOptions.safesPerLevel = self.genOptsRand:nextInt(math.min(params.difficultyOptions.min_safesPerLevel,params.difficultyOptions.max_safesPerLevel),math.max(params.difficultyOptions.min_safesPerLevel,params.difficultyOptions.max_safesPerLevel))
	end
	--params.seed = rnd._seed
	oldinit( self, params, pass )
	if random_consoles then
		params.difficultyOptions.consolesPerLevel = "RANDOM"
	end
	if random_safes then
		params.difficultyOptions.safesPerLevel = "RANDOM"
	end
end

function context:finalize()
	local options = self.params.difficultyOptions
	if type(options.powerPerLevelOverride) == "number" then
		options.powerPerLevel = options.powerPerLevelOverride
	elseif options.powerPerLevelOverride == "RANDOM" then
		options.powerPerLevel = self.genOptsRand:nextInt(math.min(options.min_powerPerLevelOverride,options.max_powerPerLevelOverride),math.max(options.min_powerPerLevelOverride,options.max_powerPerLevelOverride))
	elseif options.powerPerConsole == "RANDOM" then
		options.powerPerLevel = self.NUM_CONSOLES*self.genOptsRand:nextInt(math.min(options.min_powerPerConsole,options.max_powerPerConsole),math.max(options.min_powerPerConsole,options.max_powerPerConsole))
	elseif type(options.powerPerConsole) == "number" then
		options.powerPerLevel = self.NUM_CONSOLES*options.powerPerConsole
	else
		options.powerPerLevel = self.NUM_CONSOLES*2
	end
	local old_ICE_LOW = simdefs.ICE_LOWS[self.difficulty]
	local old_ICE_HIGH = simdefs.ICE_HIGHS[self.difficulty]
	local firewalls = options.firewalls
	local firewalldelta = options.firewalldelta
	local firewallscalar = options.firewallscalar
	local firewalldeltascalar = options.firewalldeltascalar
	if firewalls == "RANDOM" then
		firewalls = self.genOptsRand:next()*math.abs(options.min_firewalls - options.max_firewalls) + math.min(options.min_firewalls,options.max_firewalls)
	end
	if firewalldelta == "RANDOM" then
		firewalldelta = self.genOptsRand:next()*math.abs(options.min_firewalldelta - options.max_firewalldelta) + math.min(options.min_firewalldelta,options.max_firewalldelta)
	end
	if firewallscalar == "RANDOM" then
		firewallscalar = self.genOptsRand:next()*math.abs(options.min_firewallscalar - options.max_firewallscalar) + math.min(options.min_firewallscalar,options.max_firewallscalar)
	end
	if firewalldeltascalar == "RANDOM" then
		firewalldeltascalar = self.genOptsRand:next()*math.abs(options.min_firewalldeltascalar - options.max_firewalldeltascalar) + math.min(options.min_firewalldeltascalar,options.max_firewalldeltascalar)
	end
	local i = self.difficulty
	simdefs.ICE_LOWS[i] = (firewalls  or 1) + 0.4 + (firewallscalar or 1) * (0.353*(i-1) - ((i-1)*(i-1)*0.007))
	simdefs.ICE_HIGHS[i] = math.max(math.floor(simdefs.ICE_LOWS[i] + (firewalldelta or 1) -0.302 + (firewalldeltascalar or 1) * ((i-1) * 0.5302 - (i-1)*(i-1)*0.04073 + (i-1)*(i-1)*(i-1)*0.00122)),1)
	simdefs.ICE_LOWS[i] = math.max(math.floor(simdefs.ICE_LOWS[i]),1)
	local ret = oldfinalize(self)
	simdefs.ICE_LOWS[self.difficulty] = old_ICE_LOW
	simdefs.ICE_HIGHS[self.difficulty] = old_ICE_HIGH
	return ret
end

function context:generateRooms()
    local tagSet = {}
	local options = self.params.difficultyOptions
	
	local roomCount = options.roomCount
	if roomCount == "RANDOM" then
		options.roomCount = self.genOptsRand:nextInt(math.min(options.min_roomCount,options.max_roomCount),math.max(options.min_roomCount,options.max_roomCount))
	end
	local scaling = 0
	if options.roomCount_scaling == "RANDOM" then
		scaling = self.genOptsRand:next()*math.abs(options.min_roomCount_scaling - options.max_roomCount_scaling) + math.min(options.min_roomCount_scaling,options.max_roomCount_scaling)
	elseif type(options.roomCount_scaling) == "number" then
		scaling = options.roomCount_scaling
	end
	options.roomCount = math.floor(scaling * (self.difficulty - 1) + options.roomCount + 0.5)
    self:invokeScriptGen( "pregeneratePrefabs", tagSet )
	self.params.difficultyOptions.roomCount = roomCount

    for pass, tags in ipairs( tagSet ) do
	    while #tags > 0 do
		    local tag = table.remove( tags, self.rnd:nextInt( 1, #tags ))
            local fitnessFn = defaultFitnessFn
            if type(tag) == "table" then
                tag, fitnessFn = tag[1], tag[2]
                assert( type(tag) == "string" and type(fitnessFn) == "function" )
            end

		    if prefabs.generatePrefabs( self, self.candidates, tag, 1, fitnessFn, tags.fitnessSelect ) == 0 then
                error( "BAD_LEVEL: cannot fit "..tag )
		    end
	    end
    end

	return roomgen.createRooms( self, self.candidates )
end