local util = include( "client_util" )
local array = include("modules/array")
local skilldefs = include( "sim/skilldefs" )
local rand = include( "modules/rand" )
local simdefs = include( "sim/simdefs" )

local serverdefs = include( "modules/serverdefs" )

serverdefs.SITUATION_MODIFIERS = {}
serverdefs.SITREP_DAEMONS = {}
serverdefs.INCOGNITA_PENALTIES = {}

function serverdefs.refreshSituations( campaign )
	if campaign.difficultyOptions.SITREP then
		local gen = rand.createGenerator( campaign.seed )
		for _, situation in ipairs(campaign.situations) do
			if not situation.sitrep_applied then
				----------------------
				--add sitrep daemons--
				----------------------
				
				--roll first how many sitreps the situation should get
				--add 1 sitrep every day, starting with 1 on day 1
				local sitrep_count = math.floor((campaign.currentDay - campaign.difficultyOptions.SITREP_day) / campaign.difficultyOptions.SITREP_increase) + 1
				
				--special case for day 1
				if not campaign.agency.missions_completed then
					sitrep_count = sitrep_count - 1
				end
				--50% chance to have 1 less sitrep
				if gen:nextInt(1, 2) == 2 then
					sitrep_count = sitrep_count - 1
				end
				
				--cap count between 0 and 5
				sitrep_count = math.min( 5, sitrep_count )
				sitrep_count = math.max( 0, sitrep_count )
				
				if sitrep_count > 0 then
					--find possible candidates for this situation
					local candidates = {}
					for name, sitrep in pairs( serverdefs.SITREP_DAEMONS ) do
						if sitrep and sitrep.random_weight and sitrep.random_weight > 0 and sitrep.canBeSelected and sitrep.canBeSelected( situation, campaign ) then
							candidates[name] = sitrep.random_weight
						end
					end
					--pick sitreps
					local w_list = util.weighted_list(candidates)
					
					for i=1, sitrep_count do
						if w_list:getTotalWeight() <= 0 then
							break
						end
						local wt = gen:nextInt(1, w_list:getTotalWeight())
						local selected_sitrep = w_list:removeChoice( wt )
						if not situation.sitreps then
							situation.sitreps = {}
						end
						table.insert( situation.sitreps, selected_sitrep )
					end
					
					--find possible candidates for modifiers; add a modifier in 50% of the cases
					if gen:nextInt(1, 2) == 2 then
						candidates = {}
						for name, mod in pairs( serverdefs.SITUATION_MODIFIERS ) do
							if mod and mod.random_weight and mod.random_weight > 0 and mod.canBeSelected and mod.canBeSelected( situation, campaign ) then
								candidates[name] = mod.random_weight
							end
						end
						--pick mod
						w_list = util.weighted_list(candidates)
						if w_list:getTotalWeight() > 0 then
							local wt = gen:nextInt(1, w_list:getTotalWeight())
							local selected_mod = w_list:removeChoice( wt )
							local mod = serverdefs.SITUATION_MODIFIERS[selected_mod]
							if mod.onSelect then
								mod.onSelect( situation )
							end
						end
					end
					
					--add an incognita penalty from day 5 onwards
					if campaign.difficultyOptions.SITREP_incognita_day and campaign.difficultyOptions.SITREP_incognita_day >= 0 and campaign.currentDay >= campaign.difficultyOptions.SITREP_incognita_day then
						candidates = {}
						for name, penalty in pairs( serverdefs.INCOGNITA_PENALTIES ) do
							if penalty and penalty.random_weight and penalty.random_weight > 0 then
								candidates[name] = penalty.random_weight
							end
						end
						--pick penalty
						w_list = util.weighted_list(candidates)
						if w_list:getTotalWeight() > 0 then
							local wt = gen:nextInt(1, w_list:getTotalWeight())
							local selected_penalty = w_list:removeChoice( wt )
							table.insert( situation.sitreps, selected_penalty )
						end
					end
				end

				situation.sitrep_applied = true
			end
		end
		campaign.seed = gen._seed
	end
end

local baseCreateCampaignSituations = serverdefs.createCampaignSituations
function serverdefs.createCampaignSituations( campaign, count, tags, difficulty )
	baseCreateCampaignSituations( campaign, count, tags, difficulty )

	serverdefs.refreshSituations( campaign )
end

local baseNewDay = serverdefs.newDay
function serverdefs.newDay(campaign, hours)
	if campaign.difficultyOptions.SITREP then
		local gen = rand.createGenerator( campaign.seed )
		for i, situation in ipairs( campaign.situations ) do
			if (not situation.sitreps or #situation.sitreps <= 5) and (campaign.currentDay - campaign.difficultyOptions.SITREP_day) % campaign.difficultyOptions.SITREP_increase == 0 and campaign.currentDay >= campaign.difficultyOptions.SITREP_day then
				--add 1 sitrep to all situations
				local candidates = {}
				for name, sitrep in pairs( serverdefs.SITREP_DAEMONS ) do
					if sitrep and not array.find( situation.sitreps or {}, name) and sitrep.random_weight and sitrep.random_weight > 0 and sitrep.canBeSelected and sitrep.canBeSelected( situation, campaign ) then
						candidates[name] = sitrep.random_weight
					end
				end
				
				--for beautification
				local index = #(situation.sitreps or {}) + 1
				if situation.sitreps then
					if serverdefs.INCOGNITA_PENALTIES[ situation.sitreps[#situation.sitreps] ] then
						index = #situation.sitreps
					end
				end
				
				local w_list = util.weighted_list(candidates)
				if w_list:getTotalWeight() > 0 then
					local wt = gen:nextInt(1, w_list:getTotalWeight())
					local selected_sitrep = w_list:removeChoice( wt )
					if not situation.sitreps then
						situation.sitreps = {}
					end
					
					table.insert( situation.sitreps, index, selected_sitrep )
				end
			end
			
			--add an incognita penalty on day 5
			if campaign.difficultyOptions.SITREP_incognita_day and campaign.difficultyOptions.SITREP_incognita_day >= 0 and campaign.currentDay == campaign.difficultyOptions.SITREP_incognita_day then
				--check if there isnt one already
				local no_penalty = false
				for _, sitrep in ipairs( situation.sitreps ) do
					if serverdefs.INCOGNITA_PENALTIES[sitrep] then
						no_penalty = true
						break
					end
				end
				
				if not no_penalty then
					local candidates = {}
					for name, penalty in pairs( serverdefs.INCOGNITA_PENALTIES ) do
						if penalty and penalty.random_weight and penalty.random_weight > 0 then
							candidates[name] = penalty.random_weight
						end
					end
					--pick penalty
					w_list = util.weighted_list(candidates)
					if w_list:getTotalWeight() > 0 then
						local wt = gen:nextInt(1, w_list:getTotalWeight())
						local selected_penalty = w_list:removeChoice( wt )
						table.insert( situation.sitreps, selected_penalty )
					end
				end
			end
		end
		campaign.seed = gen._seed
	end
	
	baseNewDay(campaign, hours)
end


local baseCreateNewCampaign = serverdefs.createNewCampaign
function serverdefs.createNewCampaign(agency, campaignDifficulty, difficultyOptions)
	local campaign = baseCreateNewCampaign(agency, campaignDifficulty, difficultyOptions)
	
	if difficultyOptions.SITREP_incognita_upgrade then
		if (difficultyOptions.SITREP_slot or not campaign.difficultyOptions.programCount) then
			campaign.difficultyOptions.programCount = 2 --we overwrite GenOptPlus here
		end
		if (difficultyOptions.SITREP_slot or not campaign.difficultyOptions.maxPwr) then
			campaign.difficultyOptions.maxPwr = 10
		end
	end

	return campaign
end

local baseCreateAgency = serverdefs.createAgency
function serverdefs.createAgency( agentIDs, programIDs )
	local agency = baseCreateAgency( agentIDs, programIDs )

	agency.sitrep_incognita =
	{
		skills = {},
	}
	for id, skill in pairs(skilldefs.sitrep_incognita_skills) do
		agency.sitrep_incognita.skills[ id ] = {}
		agency.sitrep_incognita.skills[ id ].level = skill.starting_level or 1
	end

	agency.programLoadout = util.tcopy(agency.abilities)

	return agency
end

function serverdefs.getProgramSlots( campaign )
	return simdefs.MAX_PROGRAMS + (campaign.agency.extraPrograms or 0) + (campaign.difficultyOptions.programCount or 5) - 6 + (campaign.agency.sitrep_incognita.skills.slots.level or 1) + (campaign.agency.W93_aiTerminals or 0) --anything we can do here to not hardcode other mods?...-> probably should go into SC
end
