local util = include( "modules/util" )
local array = include( "modules/array" )
local mathutil = include( "modules/mathutil" )
local simdefs = include( "sim/simdefs" )
local simquery = include( "sim/simquery" )
local simfactory = include( "sim/simfactory" )
local unitdefs = include( "sim/unitdefs" )
local simunit = include( "sim/simunit" )
local mainframe_abilities = include( "sim/abilities/mainframe_abilities" )
local simstore = include( "sim/units/store" )

-----------------------------------------------------

function simstore.createProgramUnitData( abilityName, sim )
	local abilityDef = mainframe_abilities[ abilityName ]

	assert( abilityDef )
	assert( sim )

	if abilityDef.abilityOverride then
		if sim:isVersion(abilityDef.abilityOverride[2]) then		
			abilityDef =  mainframe_abilities[abilityDef.abilityOverride[1]]
		end
	end

	return
	{
		type = "simunit", 
		name = abilityDef.name,
		program = true,
		traits = 
		{
			mainframe_program = abilityName, 
		},
		onTooltip = function( tooltip, unit )
			tooltip:addLine( abilityDef.name, util.sformat( STRINGS.PROPS.STORE_PROGRAM_TOOLTIP, abilityDef:getCpuCost() ))
			if abilityDef.maxCooldown and abilityDef.maxCooldown > 0  then
				tooltip:addLine( util.sformat( STRINGS.PROGRAMS.COOLDOWN, abilityDef.maxCooldown )  )
			end					
			tooltip:addAbility( abilityDef.shortdesc, abilityDef.desc, "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png" )
			end,

		profile_icon_100 = abilityDef.icon_100,
		value = abilityDef.value + abilityDef.value * (sim:getPC():getTraits().shopPenalty or 0)
	}
end

function simstore.cleanItemList( sim, itemsList, needsFloorWeight )
	if sim:isVersion("0.17.18") then
		for i=#itemsList,1,-1 do
			if itemsList[i].markedForDeletion then
				table.remove(itemsList,i)
			end
		end
	end
	
	if sim:getParams().difficultyOptions.W93_corp_specific and not ( sim:getParams().world == "omni" or sim:getParams().world == "omni2" or sim:getParams().world == "neptune" ) then
		for i=#itemsList,1,-1 do
			local item = itemsList[i]
			if item.ITEM_LIST_CORP and item.ITEM_LIST_CORP ~= sim:getParams().world then
				table.remove(itemsList,i)
			end
		end
	end
	
	local campaignHours = sim:getParams().campaignHours
	if campaignHours then
		for i=#itemsList,1,-1 do
			local item = itemsList[i]
			if item.soldAfter and campaignHours <= item.soldAfter then 
				table.remove( itemsList, i )
			elseif item.notSoldAfter and campaignHours > item.notSoldAfter then
				table.remove( itemsList, i )
			elseif needsFloorWeight and not item.floorWeight then
				table.remove( itemsList, i )
			end
		end
	end
end

function simstore.cleanProgramList( sim, programList )
	if sim:isVersion("0.17.18") then
		for i=#programList,  1, -2 do
			if mainframe_abilities[programList[i]].markedForDeletion then
				programList:removeByItem(programList[i])
			end
		end
	end

	if sim:getParams().difficultyOptions.W93_corp_specific and not ( sim:getParams().world == "omni" or sim:getParams().world == "omni2" or sim:getParams().world == "neptune" ) then
		for i=#programList,  1, -2 do
			if mainframe_abilities[programList[i]].PROGRAM_LIST_CORP and mainframe_abilities[programList[i]].PROGRAM_LIST_CORP ~= sim:getParams().world then
				programList:removeByItem(programList[i])
			end
		end
	end

	local campaignHours = sim:getParams().campaignHours
	if campaignHours then
		for i=#programList,  1, -2 do
			if mainframe_abilities[programList[i]].soldAfter and mainframe_abilities[programList[i]].soldAfter >= campaignHours then
				programList:removeByItem(programList[i])
			elseif mainframe_abilities[programList[i]].notSoldAfter and mainframe_abilities[programList[i]].notSoldAfter < campaignHours then
				programList:removeByItem(programList[i])
			elseif mainframe_abilities[programList[i]].PROGRAM_LIST_CORP and mainframe_abilities[programList[i]].PROGRAM_LIST_CORP ~= sim:getParams().world then
				programList:removeByItem(programList[i])
			end
		end
	end

	return programList
end

function simstore.chooseItem( sim, currentItemLimitCounts, itemList, noDupes )
	local rand = nil
	local item = nil
	local limit = nil
	
	while #itemList > 0 and not item do
		if itemList.getTotalWeight then
			rand = sim:nextRand(1, itemList:getTotalWeight())
			item = itemList:getChoice( rand )
		else
			rand = sim:nextRand(1, #itemList)
			item = itemList[ rand ]
		end
		
		limit = item.traits.teamLimitCategory
		
		if limit and currentItemLimitCounts[limit] and currentItemLimitCounts[limit] <= 0 then
			if itemList.getTotalWeight then
				itemList:removeChoice( rand )
			else
				table.remove( itemList, rand )
			end
			
			item = nil
		end
	end
		
	if limit and currentItemLimitCounts[limit] then
		currentItemLimitCounts[limit] = currentItemLimitCounts[limit] - 1
	end
	
	if item and noDupes then
		if itemList.getTotalWeight then
			itemList:removeChoice( rand )
		else
			table.remove( itemList, rand )
		end
	end

	return item
end

function simstore.createStoreItems( store, storeUnit, sim )
	local soldItems = {}
	local soldWeapons = {}
	local soldAugments = {}

	local itemsList = util.tdupe( store.itemList )
	local weaponsList = util.tdupe( store.weaponList )
	local augmentsList = util.tdupe( store.augmentList )
	
	simstore.cleanItemList( sim, itemsList, true )
	simstore.cleanItemList( sim, weaponsList )
	simstore.cleanItemList( sim, augmentsList )
	
	-- For items with a max limit across the entire agency
	local currentItemLimitCounts = {}
	
	if rawget(simdefs, "AGENCY_ITEM_LIMITS") then
		if sim.storeItemLimit then
			currentItemLimitCounts = sim.storeItemLimit
		else
			currentItemLimitCounts = util.tdupe( simdefs.AGENCY_ITEM_LIMITS )
			sim.storeItemLimit = currentItemLimitCounts
		
			-- Do this merge to count both items in an agent's inventory and in storage
			local allAgencyUpgrades = util.tmerge( { sim:getParams().agency }, sim:getParams().agency.unitDefs )
			
			-- Count agency's current items
			for i, unitDef in pairs( allAgencyUpgrades ) do
				for j, upgrade in ipairs( unitDef.upgrades or {} ) do
					local upgradeName = type( upgrade ) == "string" and upgrade or upgrade.upgradeName
					local itemdef = unitdefs.tool_templates[ upgradeName ]
					
					if itemdef and itemdef.traits.teamLimitCategory and currentItemLimitCounts[itemdef.traits.teamLimitCategory] then
						currentItemLimitCounts[itemdef.traits.teamLimitCategory] = currentItemLimitCounts[itemdef.traits.teamLimitCategory] - 1
					end
				end
			end
		end
	end

	-- Initial set of items to sell
	if not storeUnit:getTraits().noMandatoryItems then
		if rawget(simdefs, "AGENCY_ITEM_LIMITS") then
			for i, item in pairs(store.mandatoryList) do
				local picked = simstore.chooseItem( sim, currentItemLimitCounts, { item } )
				if picked then
					table.insert( soldItems, picked )
				end
			end
		else
			soldItems = util.tdupe( store.mandatoryList )
		end
	end

	-- Items to sell
	local itemsLeft = math.max( store.storeType[storeUnit:getTraits().storeType].itemAmount - #soldItems, 0 )
	if not sim:getParams().difficultyOptions.W93_NoShopItemWeighting then
		-- Weighted item selection
		local weightedItems = util.weighted_list()

		for i=1, #itemsList do 
			if itemsList[i].floorWeight then 
				local floorWeight = itemsList[i].floorWeight
				local floor = sim:getParams().difficulty
				local weight = ( 4 - math.abs( math.min(floor, 4) - floorWeight ) ) ^ 5
				weightedItems:addChoice( itemsList[i], weight )
			end
		end
		
		for i=1, itemsLeft do
			if #itemsList == 0 then
				break
			end
			table.insert( soldItems, simstore.chooseItem( sim, currentItemLimitCounts, weightedItems, storeUnit:getTraits().noDupes ) )
		end
	else
		-- Unweighted selection
		for i=1, itemsLeft do
			if #itemsList == 0 then
				break
			end
			table.insert( soldItems, simstore.chooseItem( sim, currentItemLimitCounts, itemsList, storeUnit:getTraits().noDupes ) )
		end
	end

	-- Weapons to sell
	local weaponsLeft = store.storeType[storeUnit:getTraits().storeType].weaponAmount
	for i = 1, weaponsLeft do
		if #weaponsList == 0 then
			break
		end
		table.insert( soldWeapons, simstore.chooseItem( sim, currentItemLimitCounts, weaponsList, storeUnit:getTraits().noDupes ) )
	end
	
	-- Augments to sell
	local augmentsLeft = store.storeType[storeUnit:getTraits().storeType].augmentAmount
	for i = 1, augmentsLeft do
		if #augmentsList == 0 then
			break
		end
		table.insert( soldAugments, simstore.chooseItem( sim, currentItemLimitCounts, augmentsList, storeUnit:getTraits().noDupes ) )
	end

	-- Programs to sell
	if store.progList then
		local progList = nil
		local progList1 = nil 
		local progList2 = nil
		if sim:getParams().campaignDifficulty == simdefs.NORMAL_DIFFICULTY then
			if store.storeType[storeUnit:getTraits().storeType].noBreakers then 
				progList1 = util.tcopy( store.beginner_breakers )
				progList2 = util.tcopy( store.beginner_nobreakers )
				if sim:isVersion("0.17.5") then
					progList1:addList( store.beginner_breakers_17_5 )
					progList2:addList( store.beginner_nobreakers_17_5 )					
				end
			else
				progList = util.tcopy( store.progList )
				if sim:isVersion("0.17.5") then
					progList:addList( store.progList_17_5 )		
				end				
			end 
		else 
			if store.storeType[storeUnit:getTraits().storeType].noBreakers then 
				progList = util.tcopy( store.noBreakerProgList )
				if sim:isVersion("0.17.5") then
					progList:addList( store.noBreakerProgList_17_5 )		
				end					
			else
				progList = util.tcopy( store.progList )
				if sim:isVersion("0.17.5") then
					progList:addList( store.progList_17_5 )		
				end					
			end 
		end 
		local programsLeft = store.storeType[storeUnit:getTraits().storeType].progAmount
		if progList1 and progList2 then 
			progList1 = simstore.cleanProgramList( sim, progList1 )
			progList2 = simstore.cleanProgramList( sim, progList2 )
			local alternate = true

			while programsLeft > 0 do
				if alternate and progList1:getCount() > 0 then 
					local w = sim:nextRand(1, progList1:getTotalWeight() )
					local programName = progList1:removeChoice( w )
					local programUnit = simstore.createProgramUnitData( programName, sim )
					
					table.insert( soldItems, programUnit )
					programsLeft = programsLeft - 1
					alternate = not alternate
				elseif not alternate and progList2:getCount() > 0 then 
					local w = sim:nextRand(1, progList2:getTotalWeight() )
					local programName = progList2:removeChoice( w )
					local programUnit = simstore.createProgramUnitData( programName, sim )
					
					table.insert( soldItems, programUnit )
					programsLeft = programsLeft - 1
					alternate = not alternate
				else
					programsLeft = 0
				end
			end
		else 
			progList = simstore.cleanProgramList( sim, progList )

			while programsLeft > 0 and progList:getCount() > 0 do				
				local w = sim:nextRand(1, progList:getTotalWeight() )
				local programName = progList:removeChoice( w )
				local programUnit = simstore.createProgramUnitData( programName, sim )
				
				table.insert( soldItems, programUnit )
				programsLeft = programsLeft - 1
			end
		end
	end

	for i, item in ipairs( soldItems ) do
		soldItems[i] = simfactory.createUnit( item, sim )
	end

	for i, item in ipairs( soldWeapons ) do
		soldWeapons[i] = simfactory.createUnit( item, sim )
	end

	for i, item in ipairs( soldAugments ) do
		soldAugments[i] = simfactory.createUnit( item, sim )
	end

	return soldItems, soldWeapons, soldAugments
end

--------------------------------------------------------------------------------------
-- Instantiation of a store and its stock.

local store = { ClassType = "store" }

function store:onWarp( sim, oldcell, cell )
	if self.items == nil then
		self.items, self.weapons, self.augments = simstore.createStoreItems( simstore.STORE_ITEM, self, sim )
		self:clearBuyback()
	end
end

function store:clearBuyback()
	self.buyback = {}
	self.buyback.items = {}
	self.buyback.weapons = {}
	self.buyback.augments = {}
end

function store:isEmpty()
	return #self.items == 0 and #self.weapons == 0 and #self.augments == 0
end

function store:addItem( newItem )
	if newItem:getTraits().slot == "gun" or newItem:getTraits().slot == "melee" then
		table.insert( self.weapons, newItem )
	elseif newItem:getTraits().augment then
		table.insert( self.augments, newItem )
	else
		table.insert( self.items, newItem )
	end
end

-----------------------------------------------------
-- Interface functions

local function createStore( unitData, sim )
	return simunit.createUnit( unitData, sim, store )
end

simfactory.register( createStore )