local array = include( "modules/array" )
local util = include( "modules/util" )
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local inventory = include("sim/inventory")
local speechdefs = include("sim/speechdefs")
local abilityutil = include( "sim/abilities/abilityutil" )
local serverdefs = include( "modules/serverdefs" )
local mission_util = include( "sim/missions/mission_util" )

local function tooltip_activate( self, screen )
	abilityutil.hotkey_tooltip.activate( self, screen )

	if self.cell and self._range > 0 then
		local coords = simquery.rasterCircle( self._game.simCore, self.cell.x, self.cell.y, self._range )
		self._hiliteID = self._game.boardRig:hiliteCells( coords, {0.2,0.2,0.2,0.2} )
		
		self._hiliteUnits = {}
		for i = 1, #coords, 2 do
			local cell = self._game.simCore:getCell( coords[i], coords[i+1] )
			if cell then
				for j, cellUnit in ipairs( cell.units ) do
					if self.ability:isDaemonDevice( self._game.simCore, nil, nil, cellUnit ) then
						table.insert( self._hiliteUnits, cellUnit )
						self._game.boardRig:getUnitRig( cellUnit:getID() ):getProp():setRenderFilter( cdefs.RENDER_FILTERS["focus_target"] )
					end
				end
			end
		end
	elseif self.targetUnit then
		self._hiliteUnits = {self.targetUnit}
        self._game.boardRig:getUnitRig( self.targetUnit:getID() ):getProp():setRenderFilter( cdefs.RENDER_FILTERS["focus_target"] )
    end
end

local function tooltip_deactivate( self )
	abilityutil.hotkey_tooltip.deactivate( self )
	
	if self.cell then
		self._game.boardRig:unhiliteCells( self._hiliteID )
		self._hiliteID = nil
	end
	
	if self._hiliteUnits then
		for i, unit in ipairs(self._hiliteUnits) do
			self._game.boardRig:getUnitRig( unit:getID() ):refreshRenderFilter()
		end
		
		self._hiliteUnits = nil
	end

    if self.targetUnit then
        self._game.boardRig:getUnitRig( self.targetUnit:getID() ):refreshRenderFilter()
    end
end
		
local spellDef = util.extend( abilityutil.SPELL_UNITTARGET ) {
	strings = STRINGS.WIZARD.SPELLS.COUNTERSPELL,

	profile_icon = "gui/icons/spells/spellbreak.png",
	spellLevel = 1,
	iconColor = util.color( 120/255, 169/255, 231/255 ),
	
	getDescParams = function( self, sim, unit )
		return self:getEMPduration( sim )
	end,
	
	getEMPduration = function( self, sim )
		return 3 * self:getLevel()
	end,
	
	isDaemonDevice = function( self, sim, unit, userUnit, targetUnit)
		if (targetUnit:getTraits().mainframe_ice or 0) <= 0 then
			return false
		end
		
		return sim:getHideDaemons() or targetUnit:getTraits().mainframe_program
	end,
	
	isTarget = function( self, sim, unit, userUnit, targetUnit)
		if not self:isDaemonDevice( sim, unit, userUnit, targetUnit) then
			return false
		end
		
		local x0,y0 = targetUnit:getLocation()
		local x1,y1 = userUnit:getLocation()
		return sim:canUnitSeeUnit( userUnit, targetUnit ) or (sim:canPlayerSeeUnit( sim:getPC(), targetUnit ) and mathutil.dist2d( x0, y0, x1, y1 ) <= 12)
	end,
	
	canUseAbility = function( self, sim, abilityOwner, userUnit, targetUnitID )
		local canCast, canCastReason = self:canCastSpell( sim, abilityOwner, userUnit )
		if not canCast then
			return false, canCastReason
		end
	
		if targetUnitID then
			local targetUnit = sim:getUnit( targetUnitID )
			
			if not sim:canUnitSeeUnit( userUnit, targetUnit ) then
				return false, STRINGS.UI.COMBAT_PANEL_NO_LOS
			end
		end
	
		return true
	end,
	
	onTooltip = function( self, hud, sim, abilityOwner, abilityUser, targetUnitID )
		local addDesc
		if self:getArea() > 0 then
			addDesc = { util.sformat( STRINGS.WIZARD.SPELLS.COUNTERSPELL.radius, self:getArea() ) }
		end
		
		local tooltip, section = abilityutil.SPELL_TEMPLATE.onTooltip( self, hud, sim, abilityOwner, abilityUser, targetUnitID, addDesc )
		
		if targetUnitID then
			local targetUnit = sim:getUnit( targetUnitID )
			local cell = sim:getCell( targetUnit:getLocation() )
			
			tooltip._game = hud._game
			tooltip._range = self:getArea()
			tooltip.cell = cell
			tooltip.targetUnit = targetUnit
			tooltip.activate = tooltip_activate
			tooltip.deactivate = tooltip_deactivate
			tooltip.ability = self
		end
		
		return tooltip, section
	end,
	
	getArea = function( self )
		return 2 * (self:getLevel() - 1)
	end,

	executeAbility = function( self, sim, abilityOwner, unit, targetUnitID )
		local targetUnit = sim:getUnit( targetUnitID )
		if self:castSpell( sim, abilityOwner, unit, targetUnit ) then
			if self:getArea() > 0 then
				local x0, y0 = targetUnit:getLocation()
				local cells = simquery.rasterCircle( sim, x0, y0, self:getArea() )
				for i = 1, #cells, 2 do
					local cell = sim:getCell( cells[i], cells[i+1] )
					if cell then
						for j, cellUnit in ipairs( cell.units ) do
							if self:isDaemonDevice( sim, abilityOwner, unit, cellUnit ) then
								cellUnit:getTraits().mainframe_program = nil
								sim:dispatchEvent( simdefs.EV_KILL_DAEMON, {unit = cellUnit})
								
								if cellUnit:getTraits().daemonHost then
									sim:getUnit(cellUnit:getTraits().daemonHost):killUnit(sim)
									cellUnit:getTraits().daemonHost =nil
								end
							end
						end
					end
				end
				
				sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_DAEMON_REVEAL.path )
			else
				if targetUnit:getTraits().mainframe_program == nil then
					sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_HUD_INCIDENT_NEGATIVE.path )
					mission_util.showDialog( sim, STRINGS.UI.DIALOGS.NO_DAEMON_TITLE, STRINGS.UI.DIALOGS.NO_DAEMON_BODY )
				else
					targetUnit:getTraits().mainframe_program = nil
					sim:dispatchEvent( simdefs.EV_KILL_DAEMON, {unit = targetUnit})
					
					if targetUnit:getTraits().daemonHost then
						sim:getUnit(targetUnit:getTraits().daemonHost):killUnit(sim)
						targetUnit:getTraits().daemonHost =nil
					end

					sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_DAEMON_REVEAL.path )
				end
			end
		end
		
		self:endThrow( sim, abilityOwner, unit )
	end,
}

return spellDef