local array = include( "modules/array" )
local util = include( "modules/util" )
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local inventory = include("sim/inventory")
local speechdefs = include("sim/speechdefs")
local abilityutil = include( "sim/abilities/abilityutil" )
local serverdefs = include( "modules/serverdefs" )

local function tooltip_activate( self, screen )
	abilityutil.hotkey_tooltip.activate( self, screen )

	if self.cell then
		local coords = simquery.rasterCircle( self._game.simCore, self.cell.x, self.cell.y, self._range )
		self._hiliteID = self._game.boardRig:hiliteCells( coords, {0.2,0.2,0.2,0.2} )
	end

	if self.targetUnit then
        self._game.boardRig:getUnitRig( self.targetUnit:getID() ):getProp():setRenderFilter( cdefs.RENDER_FILTERS["focus_target"] )
    end
end

local function tooltip_deactivate( self )
	abilityutil.hotkey_tooltip.deactivate( self )
	
	if self.cell then
		self._game.boardRig:unhiliteCells( self._hiliteID )
		self._hiliteID = nil
	end

    if self.targetUnit then
        self._game.boardRig:getUnitRig( self.targetUnit:getID() ):refreshRenderFilter()
    end
end
		
local spellDef = util.extend( abilityutil.SPELL_TEMPLATE ) {
	strings = STRINGS.WIZARD.SPELLS.KNOCK,

	profile_icon = "gui/icons/spells/knock.png",
	spellLevel = 1,
	showTargets = true,
	iconColor = util.color( 208/255, 134/255, 64/255 ),
	
	getDescParams = function( self, sim, unit )
		return self:getEMPduration( sim )
	end,
	
	getEMPduration = function( self, sim )
		return 3 * self:getLevel()
	end,
	
	acquireTargets = function( self, targets, game, sim, unit, userUnit )
		-- Unit targeting
		local units = {}
		for _, targetUnit in pairs(sim:getAllUnits()) do
			if self:isTarget( sim, unit, userUnit, targetUnit ) then
				table.insert( units, targetUnit )
			end
		end

		local unitTargeting = targets.unitTarget( game, units, self, unit, userUnit )
		
		-- Exit targeting
		if not self.allDoors then
			self.allDoors = {}
			sim:forEachCell(
				function( c )
					for i, exit in pairs( c.exits ) do
						if exit.door then
							exit.dir = i
							table.insert( self.allDoors, exit )
						end
					end
				end
			)
		end
		
		local exits = {}
		local x0, y0 = userUnit:getLocation()
		
		for i, door in ipairs(self.allDoors) do
			if door.closed and door.locked then
				local toCell = door.cell
				local rdir = simquery.getReverseDirection( door.dir )
				local reverseExit = toCell.exits[ rdir ]
				local fromCell = reverseExit.cell
				
				local canSee = sim:canUnitSee( userUnit, fromCell.x, fromCell.y )
				local canSeeReverse = sim:canUnitSee( userUnit, toCell.x, toCell.y )
				
				local dist0 = mathutil.distSqr2d( x0, y0, fromCell.x, fromCell.y )
				local dist1 = mathutil.distSqr2d( x0, y0, toCell.x, toCell.y )
				
				if canSee and not ( canSeeReverse and dist1 < dist0 ) then
					table.insert( exits, { x = fromCell.x, y = fromCell.y, dir = door.dir } )
				end
			end
		end

		local exitTargeting = targets.exitTarget( game, exits, self, unit, userUnit )
		
		-- Composite targeting
		local targeting = targets.multiTarget( game, sim, { unitTargeting, exitTargeting }, true )
		
		return targeting
	end,
	
	isTarget = function( self, sim, unit, userUnit, targetUnit)
		if targetUnit:getPlayerOwner() == userUnit:getPlayerOwner() then
			return false
		end
		if not targetUnit:getTraits().safeUnit then
			return false
		end
		if not sim:canUnitSeeUnit( userUnit, targetUnit ) then
			return false
		end
		
		return true
	end,
	
	
	canUseAbility = function( self, sim, abilityOwner, unit, targetUnitID, y, dir )
		local canCast, canCastReason = self:canCastSpell( sim, abilityOwner, unit )
		if not canCast then
			return false, canCastReason
		end
	
		if y then
			local cell = sim:getCell( targetUnitID, y )
			local exit = cell.exits[dir]
			if exit.keybits == simdefs.DOOR_KEYS.SECURITY then
				return true
			elseif exit.keybits == simdefs.DOOR_KEYS.SPECIAL_EXIT then
				if self:getLevel() >= 2 then
					return true
				else
					return false, STRINGS.WIZARD.UI.CANT_USE_SPELLLEVEL2
				end
			elseif exit.keybits == simdefs.DOOR_KEYS.VAULT then
				if self:getLevel() >= 3 then
					return true
				else
					return false, STRINGS.WIZARD.UI.CANT_USE_SPELLLEVEL3
				end
			else
				return false
			end
			
		elseif targetUnitID then
			local targetUnit = sim:getUnit( targetUnitID )
			
			if targetUnit:getTraits().MM_emp_safe and self:getLevel() == 1 then
				return false, STRINGS.WIZARD.UI.CANT_USE_SPELLLEVEL2
			end
		end
	
		return true
	end,
	
	onTooltip = function( self, hud, sim, abilityOwner, abilityUser, targetUnitID, y, dir )
		local tooltip, section = abilityutil.SPELL_TEMPLATE.onTooltip( self, hud, sim, abilityOwner, abilityUser, targetUnitID, nil, y, dir )
		
		local cell
		local targetUnit
		
		if y then
			cell = sim:getCell( targetUnitID, y )
		elseif targetUnitID then
			targetUnit = sim:getUnit( targetUnitID )
			cell = sim:getCell( targetUnit:getLocation() )
		end
		
		if cell or targetUnit then
			tooltip._game = hud._game
			tooltip._range = self:getSoundRange()
			tooltip.cell = cell
			tooltip.targetUnit = targetUnit
			tooltip.activate = tooltip_activate
			tooltip.deactivate = tooltip_deactivate
		end
		
		return tooltip, section
	end,
	
	getSoundRange = function( self )
		return 8
	end,

	executeAbility = function( self, sim, abilityOwner, unit, targetUnitID, y, dir )
		local soundPath = "SpySociety/HitResponse/deathfall_agent_hardwood"
	
		if y then
			local cell = sim:getCell( targetUnitID, y )
			if self:castSpell( sim, abilityOwner, unit, cell ) then
				local keybits
				
				if cell.exits[dir].keybits == simdefs.DOOR_KEYS.SPECIAL_EXIT or cell.exits[dir].keybits == simdefs.DOOR_KEYS.VAULT then
					keybits = cell.exits[dir].keybits
					cell.exits[dir].keybits = 1
				end
				sim:modifyExit( cell, dir, simdefs.EXITOP_UNLOCK )
				if keybits then
					cell.exits[dir].keybits = keybits
				end
				sim:emitSound( { path = soundPath, range = self:getSoundRange(), ignoreSight = true }, cell.x, cell.y )
				sim:emitSound( simdefs.SOUND_DOOR_UNLOCK, cell.x, cell.y )
			end
		else
			local targetUnit = sim:getUnit( targetUnitID )
			if self:castSpell( sim, abilityOwner, unit, targetUnit ) then
				sim:emitSound( { path = soundPath, range = self:getSoundRange(), ignoreSight = true }, targetUnit:getLocation() )
				if targetUnit:getTraits().security_box_locked then
					targetUnit:getTraits().security_box_locked = nil
					targetUnit:takeControl( sim:getPC() )
				else
					for i = 1, self:getLevel() do
						targetUnit:processEMP( self:getEMPduration() )
						if targetUnit:getTraits().mainframe_booting then
							targetUnit:getTraits().MM_emp_safe = false
							break
						end
					end
				end
			end
		end
		
		self:endThrow( sim, abilityOwner, unit )
	end,
}

return spellDef