local array = include( "modules/array" )
local util = include( "client_util" )
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local inventory = include("sim/inventory")
local speechdefs = include("sim/speechdefs")
local abilityutil = include( "sim/abilities/abilityutil" )
local serverdefs = include( "modules/serverdefs" )
local unitdefs = include( "sim/unitdefs" )
local simfactory = include( "sim/simfactory" )
local mainframe = include( "sim/mainframe" )
local astar = include( "modules/astar" )
local astar_handlers = include( "sim/astar_handlers" )

local brokenDoor = {
	door = true,
	closed = true,
	locked = nil,
	openSound = simdefs.SOUND_DOOR_OPEN,
	closeSound = simdefs.SOUND_DOOR_CLOSE,
	breakSound = simdefs.SOUND_DOOR_BREAK,
	keybits =  simdefs.DOOR_KEYS.OFFICE,
	wallUVs = {
		unlocked = cdefs.WALL_DOOR_UNLOCKED,
		locked = cdefs.WALL_DOOR_LOCKED,
		broken = cdefs.WALL_DOOR_BROKEN
	},
}

local tileDir = {
	[0] = { 1, 0 },
	[2] = { 0, 1 },
	[4] = { -1, 0},
	[6] = { 0, -1},
}

local handler = util.tcopy(astar_handlers.handler)

function handler:getAdjacentNodes( cur_node, goal_cell, closed )
	local result = {}	
	local cell = cur_node.location
	local sim = self._sim

	for i, exit in pairs( cell.exits ) do
		if not exit.keybits or exit.keybits < simdefs.DOOR_KEYS.ELEVATOR then
			local target_cell = exit.cell
			local target_lid = self:getNodeLid( target_cell.x, target_cell.y )
			if not closed[ target_lid ] then
				local n = self:_handleNode( target_cell, cur_node, goal_cell )
				if n then
					table.insert(result, n)
				end
			end
		end
	end

	return result
end

local spellDef = util.extend( abilityutil.SPELL_CELLTARGET ) {
	strings = STRINGS.WIZARD.SPELLS.SHATTER,

	profile_icon = "gui/icons/spells/shatter.png",		
	spellLevel = 2,
	soundRange = 20,
	areaPassesThroughWalls = true,
	
	iconColor = util.color( 232/255, 96/255, 69/255 ),
	
	isAreaTargetUnit = function( self, sim, userUnit, cellUnit )
		return cellUnit:hasTrait("isAgent") or cellUnit:hasTrait("isGuard")
	end,
	
	getEffectRadius = function( self, sim, abilityOwner, userUnit, targetCell )
		return self:getLevel() - 2
	end,
	
	executeAbility = function( self, sim, abilityOwner, userUnit, targetCell )
		local x0,y0 = userUnit:getLocation()
		local x1,y1 = unpack(targetCell)
		
		if self:castSpell( sim, abilityOwner, userUnit, targetCell ) then
			local fromCell = sim:getCell( x1, y1 )
			local cells = simquery.rasterCircle( sim, x1, y1, self:getEffectRadius() )
			local hitUnits = {}
			
			local trueCost = astar.AStar:new( handler:new( sim ) )
			
			for j = 1, #cells, 2 do
				local cell = sim:getCell( cells[j], cells[j + 1] )
				
				if cell then
					for k, cellUnit in ipairs( cell.units ) do
						if cellUnit:hasTrait("isAgent") or cellUnit:hasTrait("isGuard") then
							table.insert(hitUnits, cellUnit)
						end
					end

					for i = 0, 6, 2 do
						local dx0, dy0 = cell.x - x1, cell.y - y1
						local dx1, dy1 = tileDir[i][1], tileDir[i][2]
						if (dx0 * dx1 + dy0 * dy1) >= 0 then
							if simquery.checkIsDoor( sim, cell, i ) and cell.exits[i].keybits < simdefs.DOOR_KEYS.ELEVATOR then
								if cell.exits[i].locked then
									sim:modifyExit( cell, i, simdefs.EXITOP_UNLOCK )
								end
								sim:modifyExit(cell, i, simdefs.EXITOP_BREAK_DOOR )
							elseif simquery.checkIsWall( sim, cell, i ) then
								local targetCell = sim:getCell( cell.x + dx1, cell.y + dy1 )

								if targetCell and not targetCell.isSolid then
									if trueCost:findNodeCost( fromCell, targetCell, targetCell ) and trueCost:findNodeCost( fromCell, cell, cell ) then
										local rdir = simquery.getReverseDirection(i)
										cell.sides = util.tdupe(cell.sides)
										targetCell.sides = util.tdupe(targetCell.sides)

										cell.exits = util.tdupe(cell.exits)
										targetCell.exits = util.tdupe(targetCell.exits)

										cell.sides[i] = util.tcopy(brokenDoor)
										targetCell.sides[rdir] = util.tcopy(brokenDoor)
										cell.exits[i] = util.tcopy(brokenDoor)
										targetCell.exits[rdir] = util.tcopy(brokenDoor)
										targetCell.exits[rdir].cell = cell
										cell.exits[i].cell = targetCell
										sim:dispatchEvent( "EV_WALL_DESTROYED", { cell = cell, rCell = targetCell, dir = i, rdir = rdir } )
										sim:modifyExit( cell, i, simdefs.EXITOP_BREAK_DOOR )
									end
								end
							end
						end
					end
				end
			end
			
			sim:dispatchEvent( simdefs.EV_FLASH_VIZ, {x = x1, y = y1, units = nil, range = self.soundRange } )

			sim:startTrackerQueue(true)
			sim:startDaemonQueue()
			
			for i, hitUnit in ipairs( hitUnits ) do
				sim:damageUnit( hitUnit, 0, hitUnit:processKOresist( self:getLevel() ) )
			end
			
			for i, enemyUnit in ipairs( sim:getNPC():getUnits() ) do
				local x2, y2 = enemyUnit:getLocation()
				if x2 and enemyUnit:getBrain() and not enemyUnit:isKO() and mathutil.distSqr2d( x1, y1, x2, y2 ) <= 64 and not enemyUnit:getTraits().noInterestDistraction then
					enemyUnit:setAlerted(true)
					enemyUnit:getBrain():getSenses():addInterest( x1, y1, simdefs.SENSE_HEARING, simdefs.REASON_NOISE, userUnit )
				end
			end
			
			sim:emitSound( { range = self.soundRange }, x1, y1, userUnit )
			
			sim:startTrackerQueue(false)				
			sim:processDaemonQueue()
		end

		self:endThrow( sim, abilityOwner, userUnit )
	end,
}

return spellDef