local cdefs = include( "client_defs" )
local util = include( "client_util" )
local mui = include("mui/mui")
local mui_defs = include("mui/mui_defs")
local mui_util = include("mui/mui_util")
local cheatmenu = include( "fe/cheatmenu" )
local simdefs = include("sim/simdefs")


local statedebug = include("states/state-debug")

-- A local "environment" for managing debug state
local debugenv =
{
	__index = function( t, k )
		local v = rawget(_G, k)
		if v then
			return v
		end
	end,

	getCurrentGame = function( self )
        local states = statemgr.getStates()
        for i = 1, #states do
            if states[i].simCore and states[i].levelData then
                return states[i] -- If it acts like a duck and quacks like a duck...
            end
        end
	end,

	updateEnv = function( self )
		self.game = self:getCurrentGame()
		self.boardRig = self.game and self.game.boardRig
		self.sim = self.game and self.game.simCore
		self.simquery = self.sim and self.sim:getQuery()
		self.simdefs = self.sim and self.sim:getDefs()
		self.currentPlayer = self.sim and self.sim:getCurrentPlayer()
		self.localPlayer = self.game and self.game:getLocalPlayer()
        if self.game then
            self.cx, self.cy = self.game:wndToSubCell( inputmgr.getMouseXY() )
            self.cx, self.cy = math.floor(self.cx), math.floor(self.cy)
        end
		self.statedebug = statedebug

		setmetatable( self, self )
	end,
}

function statedebug:toggleVisible()
	if self.screen:isActive() then
		for cellID, fx in pairs( self.soundMarkers ) do
			fx:destroy()
		end
		self.soundMarkers = {}
		mui.deactivateScreen( self.screen )
        self.cheatMenu = nil

	else
		self.startTime = MOAISim.getDeviceTime ()
		self.screen:setPriority( math.huge )
		mui.activateScreen( self.screen )

        local cheatmenu = include( "fe/cheatmenu" )
        self.cheatMenu = cheatmenu.menu( self.screen, debugenv, self )

		if self.replayPanel then
            self.replayPanel:updatePanel()
        end
	end
end

function statedebug:onInputEvent( event )
	if event.eventType == mui_defs.EVENT_KeyDown or event.eventType == mui_defs.EVENT_KeyRepeat then
		if (event.key == mui_defs.K_BACKQUOTE or event.key == mui_defs.K_PERIOD) and event.controlDown then
			self:toggleVisible()
		else
			for i, dbgBinding in ipairs( self.keybindings ) do
                if mui_util.isBinding( event, dbgBinding.binding ) then
                    assert( type(dbgBinding.fn) == "function" )
					debugenv:updateEnv()
                    local bindingTxt = mui_util.getBindingName( dbgBinding.binding )
                    cheatmenu.executeBinding( bindingTxt, debugenv, dbgBinding.fn )
					return true
				end
			end
		end
	end

    if self.cheatMenu then
        if self.cheatMenu:onInputEvent( event ) then
            return true
        end
    end

	if (statedebug.ALOS_Debug or statedebug.ELOS_Debug) and event.eventType == mui_defs.EVENT_MouseMove then
		local game = debugenv:getCurrentGame()
		local shadow_map = game and game.shadow_map
		if shadow_map then
			local x,y = game:wndToWorld( event.wx, event.wy )
			if x and y then
				--print( 'mouse', event.wx, event.wy, x,y )
				if statedebug.ALOS_Debug then
					shadow_map:setALOSDebugPos( x, y )
				end
				if statedebug.ELOS_Debug then
					shadow_map:setELOSDebugPos( x, y )
				end
			end
		end
	end

	return false
end

function statedebug:addKeyBinding( keys, fn )
    local newBinding = mui_util.makeBinding( keys )
	for i, dbgBinding in ipairs( self.keybindings ) do
        --if mui_util.isBinding( dbgBinding.binding, newBinding ) then--This is wrong, it always returns false
		if #dbgBinding.binding == #newBinding then
			local sameBinding = true
			for j, key in pairs(dbgBinding.binding) do
				if newBinding[j] ~= key then
					sameBinding = false
					break
				end
			end
			if sameBinding then
				log:write("Replacing duplicate key binding: %s", mui_util.getBindingName( newBinding ))
				dbgBinding.fn = fn
				return
			end
        end
	end

    --log:write( "Debug binding: %s", mui_util.getBindingName( newBinding ))
	table.insert( self.keybindings, { binding = newBinding, fn = fn } )
end

cdefs.DBG_MODIFIERS = cdefs.DBG_MAX
cdefs.DBG_PATROLS = cdefs.DBG_MAX + 1
cdefs.DBG_MAX = cdefs.DBG_MAX + 2

function statedebug:updatePatrols(selectedUnit)
	if not debugenv.sim then
		return
	end
	
	if not self.patrolMarkers then
		self.patrolMarkers = {}
	end
	if not self.poolMarkers then
		self.poolMarkers = {}
	end

	if self.mode == cdefs.DBG_PATROLS and debugenv.game.simHistoryIdx == self.simHistoryIndex then
		if self.selectedUnit ~= selectedUnit or not self.isPatrolMode then
			self.isPatrolMode = true
			
			for k,v in pairs(self.patrolMarkers) do
				v:destroy()
			end
			self.patrolMarkers = {}
			
			if self.hilitePatrolID then
				debugenv.game.boardRig:unhiliteCells( self.hilitePatrolID )
				self.hilitePatrolID = nil
			end
			
			if selectedUnit and selectedUnit:getTraits().patrolPath then
				local cells = {}
				for i, patrolPoint in ipairs(selectedUnit:getTraits().patrolPath) do
					local wx, wy = debugenv.game:cellToWorld(patrolPoint.x, patrolPoint.y)
					table.insert(self.patrolMarkers, debugenv.game.fxmgr:addLabel( "p"..tostring(i), wx, wy, nil, {r=0.5,g=1,b=0,a=1} ))
				end
				self.hilitePatrolID = debugenv.game.boardRig:hiliteCells( selectedUnit:getTraits().patrolPath, {0,0,1,0.5} )
			end
			
			if self.hilitePoolID then
				debugenv.game.boardRig:unhiliteCells( self.hilitePoolID )
				self.hilitePoolID = nil
			end
			
			for k,v in pairs(self.poolMarkers) do
				v:destroy()
			end
			self.poolMarkers = {}
			
			if selectedUnit and selectedUnit:getTraits().patrolPool then
				local bestScores = {}
				local cells = {}
				
				for i, cell in pairs(selectedUnit:getTraits().patrolPool) do
					if not bestScores[cell[1].id] then
						table.insert(cells,cell[1])
						bestScores[cell[1].id] = cell[2]
						local wx, wy = debugenv.game:cellToWorld(cell[1].x, cell[1].y)
						self.poolMarkers[cell[1].id] = debugenv.game.fxmgr:addLabel( "c"..tostring(math.floor(cell[2] * 10) * 0.1), wx + 8, wy, nil, {r=0,g=0.25,b=1,a=0.75} )
					elseif bestScores[cell[1].id] > cell[2] then
						self.poolMarkers[cell[1].id]:setString("c"..tostring(cell[2]))
					end
				end
				self.hilitePoolID = debugenv.game.boardRig:hiliteCells( cells, {1,0,0,0.2} )
			end
		end
	elseif self.isPatrolMode then
		self.isPatrolMode = false
		for k,v in pairs(self.patrolMarkers) do
			v:destroy()
		end
		self.patrolMarkers = {}
		for k,v in pairs(self.poolMarkers) do
			v:destroy()
		end
		self.poolMarkers = {}
		if self.hilitePoolID then
			debugenv.game.boardRig:unhiliteCells( self.hilitePoolID )
			self.hilitePoolID = nil
		end
		if self.hilitePatrolID then
			debugenv.game.boardRig:unhiliteCells( self.hilitePatrolID )
			self.hilitePatrolID = nil
		end
	end
end

local oldGetDebugText = statedebug.getDebugText

function statedebug:getDebugText(unit, debugMode)
	if debugMode == cdefs.DBG_MODIFIERS then
		if not unit or not unit.getUnitData then
			return
		end

		local unitText = {
			string.format("UNIT: %s (%s)[%d]\n", unit:getUnitData().name, tostring(unit:getUnitData().class or "guard"), unit:getID() ),
		}
		
		local traits = unit:getModifiers().traits
		for trait, modifiers in pairs(unit:getModifiers().modifiers) do
			table.insert(unitText,string.format("\n----------------------------------\nTRAIT: %s %s => %s\n", trait, tostring(modifiers[1].val), tostring(traits[trait] )))
			local val = nil
			for i, modifier in ipairs(modifiers) do
				----Must match the updated modifiers.lua from Sim Constructor!!----
				local BASE = "." -- Base value of trait
				local ADD = '+' -- Adds to a numeric trait
				local MULT = '*' -- Multiplies a numeric trait 
				local CLA = '~' -- Clamps a numeric trait 
				local SET = '=' -- Sets a trait
				local FN = 'f' -- Custom function
				
				if modifier.op == BASE or modifier.op == SET then
					val = modifier.val
				elseif modifier.op == FN then
					val = modifier.val( val, modifier, true )
				elseif type(val) == "number" then
					local prev = val
					if modifier.op == ADD then
						val = val + modifier.val
					elseif modifier.op == MULT then
						val = val * modifier.val
					end
					
					if modifier.min and (val < prev or modifier.op == CLA) and val < modifier.min then
						val = modifier.min
					end
					if modifier.max and (val > prev or modifier.op == CLA) and val > modifier.max then
						val = modifier.max
					end
				end
				-------------------------------------------------------------------
			
				table.insert(unitText,string.format("    MODIFIER: %s => %s (%s %s) [%s %s %s %s]\n", tostring(modifier.name), tostring(val), modifier.op, tostring(modifier.val), tostring(modifier.min), tostring(modifier.max), tostring(modifier.pr), tostring(modifier.opPr) ))
			end
		end
		
		return table.concat(unitText)
	elseif debugMode == cdefs.DBG_PATROLS then
		if not unit or not unit.getUnitData then
			return
		end

		local unitText = {
			string.format("UNIT: %s (%s)[%d]\n\n", unit:getUnitData().name, tostring(unit:getUnitData().class or "guard"), unit:getID() ),
		}
		
		local l = 0
		
		if unit:getTraits().patrolPath then
			for i, point in ipairs(unit:getTraits().patrolPath) do
				table.insert(unitText,string.format("Patrol Point (%s, %s) [%s]\n", point.x, point.y, tostring(math.floor((point.l or 0) * 10) * 0.1) ))
				l = l + (point.l or 0)
			end
		
			table.insert(unitText,string.format("\nPatrol Length: %s / %s", tostring(math.floor(l * 10) * 0.1), tostring(math.floor(unit:getTraits().mpMax * 10 * #unit:getTraits().patrolPath) * 0.1) ))
		end
		
		return table.concat(unitText)
	else
		return oldGetDebugText(self, unit, debugMode)
	end
end

local oldOnLoad = statedebug.onLoad

statedebug.onLoad = function ( self )
	oldOnLoad( self )
	
	self.screen.binder.modeCmb:addItem( "DBG_MODIFIERS" )
	self.screen.binder.modeCmb:addItem( "DBG_PATROLS" )
	
	for i, binding in ipairs( simdefs.BINDINGS ) do
		self:addKeyBinding( binding[1], binding[2] )
	end
end

local oldOnUpdate = statedebug.onUpdate

statedebug.onUpdate = function ( self )
	debugenv:updateEnv()
	
	self:updatePatrols( debugenv.game and debugenv.game.hud and debugenv.game.hud:getSelectedUnit() )
	oldOnUpdate( self )
end