local stateGenerationOptions = include( "states/state-generation-options" )

local util = include("client_util")
local array = include( "modules/array" )
local mui = include( "mui/mui" )
local serverdefs = include( "modules/serverdefs" )
local modalDialog = include( "states/state-modal-dialog" )
local rig_util = include( "gameplay/rig_util" )
local cdefs = include("client_defs")
local scroll_text = include("hud/scroll_text")
local unitdefs = include( "sim/unitdefs" )
local metadefs = include( "sim/metadefs" )
local simdefs = include( "sim/simdefs" )
local movieScreen = include('client/fe/moviescreen')
local stateTeamPreview = include( "states/state-team-preview" )
local mui_button = include( "mui/widgets/mui_button" )
local mui_defs = include( "mui/mui_defs")

local function FindIndexOf(val, array)
	for k,v in pairs(array) do 
		if v == val then return k end
	end
end

stateGenerationOptions.findOptions = function(self)
	if not self.CUSTOM_OPTIONS then
		self.CUSTOM_OPTIONS = {}
		for index, name in ipairs(simdefs.CUSTOM_OPTIONS_ORDER) do
			self.CUSTOM_OPTIONS[index] = simdefs.CUSTOM_OPTIONS[name]
		end
	end
	return self.CUSTOM_OPTIONS
end

local DLC_SETTING = 
{
	check = true,
	apply = function( self, options, value )
		options[self.dlc_id].options[self.dlc_option].enabled = value
	end,
	retrieve = function( self, options )
		return options[self.dlc_id].options[self.dlc_option].enabled
	end
}

local function updateRewinds( self )
	local item = self.allItems[self.rewindIndex]
	local widget = item.widget.binder.widget

	self._panel:findWidget( "numRewinds" ):selectIndex( widget:getIndex() )
end

local function updateRetries( self )
	local item = self.allItems[self.retriesIndex]
	local widget = item.widget.binder.widget

	self._panel:findWidget( "levelRetriesBtn" ):setChecked( widget:isChecked() )
end

local function onClickLevelRetries( self, checkBox )
	local item = self.allItems[self.retriesIndex]
	local widget = item.widget.binder.widget
	
	widget:setChecked( checkBox:isChecked() )
end

local function onChangedRewindOption( self )
	local item = self.allItems[self.rewindIndex]
	local widget = item.widget.binder.widget
	
	widget:selectIndex( self._panel:findWidget( "numRewinds" ):getIndex() )
end

local function checkForMasks(self)
	local list = self._screen:findWidget("list")
	local items = {}
	local newItems = {}
	local masks = {}
	
	-- Create a table of currently set masks
	for i, item in ipairs(self.allItems) do
		local setting = item.user_data
		if setting.masks then
			local widget = item.widget.binder.widget
			for i, mask in pairs(setting.masks) do
				local value
				if setting.check then
					value = widget:getValue()
				else
					value = setting.values[widget:getIndex()]
				end
			
				if mask.raw then
					masks[mask.mask] = value
				else
					if mask.requirement == value then
						masks[mask.mask] = true
					end
				end
			end
		end
	end
	
	self.masks = masks
	
	-- Create a new list of shown items based on the masks
	for i, item in ipairs(self.allItems) do
		local setting = item.user_data
		local ok = true
		-- Hide items under a toggled header
		if item.widget._header and item.widget._header.hide then
			ok = false
		end
		
		if setting.requirements then
			for i, requirement in pairs(setting.requirements) do
				if requirement.passFunction then
					if not requirement.passFunction( masks ) then
						ok = false
					end
				elseif masks[requirement.mask] ~= requirement.requirement and not (requirement.requirement == false and masks[requirement.mask] == nil) then
					ok = false
				end
			end
		end
		
		if ok then
			items[#items+1] = item
			newItems[item] = true
		end
	end
	
	-- Deattach items that need to be hidden
	for i, item in ipairs(list:getItems()) do
		if not newItems[item] then
			if item.isAttached then
				item.isAttached = nil
				list:removeChild( item.widget )
			end
		end
	end
	
	-- Refresh the listbox
	list._items = items
	list:scrollItems(list._scrollIndex)
	if list:getScreen() then
		list:getScreen():refreshPriority()
	end
	
	if list._scrollbar then
		local count = list:getMaxVisibleItems()
		if #list._items > count then
			list._scrollbar.binder.scrollbar:setRange( 0, #list._items - count )
			list._scrollbar:setVisible( true )
		else
			list._scrollbar:setVisible( false )
		end
	end
end

local function checkLinkedValues( self, setting, options )
	--[[self.updatingLinkedValues = true

	local value
	if setting.check then
		value = widget:getValue()
	else
		value = setting.values[ widget:getIndex() ]
	end
	
	for i, item in ipairs(self.allItems) do
		local linkedSetting = item.user_data
		if linkedSetting.linkedValues then
			for i, requirement in pairs( linkedSetting.linkedValues ) do
				if requirement.key == setting.linkKey then
					if requirement.matchValue == value then
						if linkedSetting.values then
							local idx = FindIndexOf( requirement.updateValue, linkedSetting.values )
							item:selectIndex( idx )
						else
							item:setValue( requirement.updateValue )
						end
					end
				end
			end
		end
	end
	
	self.updatingLinkedValues = nil]]
end

local function onChangedOption( self, setting, widget )
	-- Iron man setting is special: it links to the iron man checkbox, and does not imply custom difficulty.
	if setting.ironManCheck then
		local options = {}
		setting:apply( options, widget:getIndex()  )

		updateRewinds(self)
	elseif not self.updatingDifficulty and not setting.noUpdate then
		-- If current difficulty is not CUSTOM, and an option was change to the non-default value...
		-- automatically switch to CUSTOM DIFFICULTY.
		self:selectDifficulty( simdefs.CUSTOM_DIFFICULTY )
	end
	
	if setting.linkKey then
		checkLinkedValues( self, setting, options )
	end

	updateRetries(self)
	if setting.masks then
		checkForMasks( self )
	end
end

local function onClickHideOptions( self, widget )
	widget.hide = not widget.hide
	checkForMasks(self)
end

local function onClickDiff( self, difficulty)
	self:selectDifficulty( difficulty )
	if self._screen.setFocusedWidget and MOAIEnvironment.Use_Gamepad == 1 then
		self._screen:setFocusedWidget( self._panel.binder.startBtn )
	end
end

local function addGenerationOption(self, setting)
	local list =  self._screen:findWidget("list")
	local widget
	if setting.check then
		widget = list:addItem( setting, "CheckOption" )
		widget.binder.widget:setText( setting.name )
		widget.binder.widget.onClick = util.makeDelegate( nil, onChangedOption, self, setting, widget.binder.widget )

		if self._currentHeader then
			self._currentHeader:setVisible(true)
			widget._header = self._currentHeader
		end
	elseif setting.values then
		widget = list:addItem( setting, "ComboOption" )
		for i, item in ipairs(setting.values) do

			widget.binder.widget:addItem( setting.strings and setting.strings[i] or item )
		end
		widget.binder.widget.onTextChanged = util.makeDelegate( nil, onChangedOption, self, setting, widget.binder.widget )
		widget.binder.dropTxt:setText( setting.name )
		
		if self._currentHeader then
			self._currentHeader:setVisible(true)
			widget._header = self._currentHeader
		end
	elseif setting.section then
		widget = list:addItem( setting, "SectionHeader" )
		widget.binder.label:setText( setting.section )
		widget.binder.icon:setVisible(false)		
		if setting.icon then
			widget.binder.icon:setVisible(true) 
			widget.binder.icon:setImage(setting.icon)
		end
		if widget.binder:tryBind("hideBtn") then
			self._currentHeader = widget.binder.hideBtn
			self._currentHeader:setVisible(false)
			self._currentHeader.onClick = util.makeDelegate( nil, onClickHideOptions, self, self._currentHeader )
		end
	end

	setting.list_index = list:getItemCount()
	assert(setting.list_index > 0)
	if setting.name == STRINGS.UI.DIFF_OPTION_REWINDS then
		self.rewindIndex = setting.list_index
	end
	if setting.name == STRINGS.UI.DIFF_OPTION_LEVEL_RETRIES then
		self.retriesIndex = setting.list_index
	end

	widget:setTooltip(setting.tip)
	
	if setting.nestedOptions then
		for i, nestedsetting in ipairs(setting.nestedOptions) do
			local opt_setting = nestedsetting
			opt_setting.dlc_id = nestedsetting.dlc_id
			opt_setting.dlc_option = nestedsetting.dlc_option and (opt_setting.option or opt_setting.dlc_option or opt_setting.name)
			addGenerationOption(self,opt_setting,true)
		end
	end
	
	return widget
end

local function onClickCancelBtn(self)
	MOAIFmodDesigner.playSound( cdefs.SOUND_HUD_GAME_WOOSHOUT )
	statemgr.deactivate( self )
	local stateMainMenu = include( "states/state-main-menu" )
	statemgr.activate( stateMainMenu() )
end

local function onClickShowOptionsBtn(self) 
	self:showDifficultyOptions( self._screen.binder.showOptionsBtn:isChecked() )
end

local function onClickSelectDLCBtn(self)
	local restoreOptions = self:retrieveOptions()

	local dlcDialog = include( "fe/select-dlc-dialog" )
	local dlc_res = dlcDialog:show(self._dlcSelections)
	if dlc_res then
		self:refreshGenerationOptions()
		self:restoreGenerationOptionSelections( restoreOptions )
	end
end

local function onToggleUniqueDLC(self)
	local settings = savefiles.getSettings( "settings" )

	if not settings.data.dlcWarnShown then
		if STRINGS.UI.DLC_WARNING then
			local modalDialog = include( "states/state-modal-dialog" )
			local result = modalDialog.showYesNo( STRINGS.UI.DLC_WARNING, STRINGS.UI.DLC_ENABLE, nil, STRINGS.UI.DLC_ENABLE, nil, true )
			if result == modalDialog.OK then
				settings.data.dlcWarnShown = true
			end
		else
			settings.data.dlcWarnShown = true
		end
	end

	if settings.data.dlcWarnShown then
		self._dlcSelections[self.uniqueModID].enabled = not self._dlcSelections[self.uniqueModID].enabled
		
		local restoreOptions = self:retrieveOptions()
		self:refreshGenerationOptions()
		self:restoreGenerationOptionSelections( restoreOptions )
		
		if self._dlcSelections[self.uniqueModID].enabled then
			self._panel.binder.dlcBtn.binder.btn:setText(STRINGS.UI.HUD_ENABLED)
		else
			self._panel.binder.dlcBtn.binder.btn:setText(STRINGS.UI.HUD_DISABLED)
		end
	end
end

local function launchGame( self )
	local settings = savefiles.getSettings( "settings" )
	settings.data.lastdiff = self._diff
	settings:save()

	MOAIFmodDesigner.playSound( cdefs.SOUND_HUD_GAME_WOOSHOUT )

	local options = self:retrieveOptions()

	statemgr.deactivate( self )
	MOAIFmodDesigner.stopSound("theme")

	if config.CAMPAIGN_HOURS ~= 72 then
		options.maxHours = config.CAMPAIGN_HOURS
	end

	mod_manager:resetContent()
	mod_manager:loadModContent( options.enabledDLC, options )
	
	local exitFn = mod_manager.postGenerationOptionsFunction
	if exitFn then
		exitFn[1] ( self, self._diff, options, unpack( exitFn[2] ) )
	elseif options.exitfunction then
		options.exitfunction( self, self._diff, options )
	else
		local endless = options.maxHours == math.huge
		if endless or not config.SHOW_MOVIE then
			statemgr.activate( stateTeamPreview( self._diff, options ) )
		else
			local SCRIPTS = include('client/story_scripts')
			
			movieScreen("data/movies/IntroCinematic.ogv", function() 
				statemgr.activate( stateTeamPreview(  self._diff, options ))
			end,  SCRIPTS.SUBTITLES.INTRO )
		end
	end
end

local function onClickStartBtn(self) 
	launchGame(self)
end

local function onItemClicked( self, idx, user_data )
	if MOAIEnvironment.Use_Gamepad == 1 then
		local item = self._screen:findWidget("list"):getItem(idx)

		if item.widget._name == "CheckOption" then
			item.widget.binder.widget:setChecked( not item.widget.binder.widget:isChecked() )
		elseif item.widget._name == "ComboOption" then
			local ev = {}
			ev.eventType = mui_defs.EVENT_ButtonClick
			item.widget.binder.widget:handleEvent( ev )
		elseif item.widget._name == "SectionHeader" then
			if item.widget.binder.tryBind("hideBtn") then
				onClickHideOptions( self, item.widget.binder.hideBtn )
			end
		end
	end
end

local function onItemFocused( self, old_idx, new_idx )
	local list = self._screen:findWidget("list")
	local item = list:getItem( new_idx )
	
	if item then
		if item.widget._name == "CheckOption" or item.widget._name == "ComboOption" then
			self._screen:setFocus(item.widget.binder.widget)
			list:selectIndex(new_idx)
		elseif item.widget._name == "SectionHeader" then
			if item.widget.binder:tryBind("hideBtn") and item.widget.binder.hideBtn:isVisible() then
				self._screen:setFocus(item.widget.binder.hideBtn)
				list:selectIndex(new_idx)
			else
				if old_idx > new_idx then
					onItemFocused( self, old_idx, new_idx - 1 )
				else
					onItemFocused( self, old_idx, new_idx + 1 )
				end
			end
		end
	elseif new_idx < 1 then
		local candidate = nil
		if self._screen._currentFocusedWidget.NavUp ~= -1 then
			candidate = self._screen:findWidgetByNavID( self._screen._currentFocusedWidget.NavUp )
		end
		if candidate ~= nil then
			self._screen:setFocusedWidget(candidate)
		end
	else
		list:selectIndex(old_idx)
	end
end

local function setTooltipPosition( self, widget, tooltip )
	local list = self._screen:findWidget("list")
	if list._selectedIndex and list._items[ list._selectedIndex ].widget then
		local x, y = list._items[ list._selectedIndex ].widget:getAbsolutePosition()
		local W, H = self._screen:getResolution()
		local w, h = list:getSize()
	
		if MOAIEnvironment.OS == "NX" then
			if MOAIEnvironment.isHandHeld == 1 then
				-- This must be the same value than in mui_scene.lua 
				x = x * 0.8
				y = y * 0.8
			end
		end
	
		tooltip:setPosition(x*W+w/2, H-y*H)
	end
end

local function extend( t1, t2 )
	-- Deep-merge all of t1 into t2
	for k,v in pairs( t1 ) do
		if type(v) == "table" then
			t2[k] = extend( v, t2[k] or {} )
		elseif t2[k] == nil then
			t2[k] = v
		end
	end
		
	return t2
end

stateGenerationOptions.onLoad = function( self, difficulty, difficultyOptions )
	self._navDificulty = {}

	if MOAIEnvironment.OS == "PS4" then
		self._screen = mui.createScreen( "generation-options_ps4.lua" )
	elseif MOAIEnvironment.OS == "XBO" then
		self._screen = mui.createScreen( "generation-options_ps4.lua" )
	elseif MOAIEnvironment.OS == "NX" then
		self._screen = mui.createScreen( "generation-options_ps4.lua" )
	else
		self._screen = mui.createScreen( "generation-options.lua" )
	end
	
	mui.activateScreen( self._screen )

	local CUSTOM_OPTIONS = self:findOptions()
	
	-- failsafe
	FMODMixer:popMix("nomusic")
	
	local CHARACTER_IMAGES = simdefs.STATE_GENERATION_OPTIONS.CHARACTER_IMAGES

	local imgs = {}
	for i = 1, #CHARACTER_IMAGES do
		if CHARACTER_IMAGES[i].unlock == nil then
			table.insert(imgs, CHARACTER_IMAGES[i].png)
		elseif metadefs.isRewardUnlocked(CHARACTER_IMAGES[i].unlock) then
			table.insert(imgs, CHARACTER_IMAGES[i].png)
		end
	end

	local idx = math.random(1,#imgs)
	
	self._screen.binder.agent:setImage( "gui/menu pages/main/"..imgs[idx]..".png")
	self._screen.binder.agent:createTransition("activate_left")
	
	self._scroll_text = scroll_text.panel( self._screen.binder.bg )

	self._panel = self._screen.binder.pnl

	local installedMods = mod_manager:getInstalledMods()
	
	if next(installedMods) then
		self._panel.binder.dlcBtn:setVisible(true)
		self._panel.binder.dlcTitle:setVisible(true)
		
		if self._panel.binder:tryBind("dlcTitlePs4") then
			if MOAIEnvironment.OS == "PS4" or MOAIEnvironment.OS == "XBO" or MOAIEnvironment.OS == "NX" then
				self._panel.binder.dlcTitle:setVisible(false)
				self._panel.binder.dlcTitlePs4:setVisible(true)
			else
				self._panel.binder.dlcTitle:setVisible(true)
				self._panel.binder.dlcTitlePs4:setVisible(false)
			end
		end
		
		local enabledDLC = (difficultyOptions and difficultyOptions.enabledDLC) or {}
		local defaultDlc = mod_manager:getModContentDefaults(difficulty or simdefs.STATE_GENERATION_OPTIONS.DEFAULT_DIFF)
		self._dlcSelections = extend( defaultDlc, enabledDLC )
		
		if MOAIEnvironment.OS == "PS4" or MOAIEnvironment.OS == "XBO" or MOAIEnvironment.OS == "NX" then
			-- Disable DLC until a warning is shown and accepted
			local settings = savefiles.getSettings( "settings" )
			if not difficultyOptions and not settings.data.dlcWarnShown then
				for _, modID in ipairs(installedMods) do
					self._dlcSelections[modID].enabled = false
				end
			end
		end
		
		-- If there's only one DLC, convert the button into an on/off button
		self.uniqueModID = ""
		if MOAIEnvironment.OS == "NX" then
			if #installedMods == 1 then
				self.uniqueModID = installedMods[1]
				self._panel.binder.dlcTitlePs4:setText(string.upper(mod_manager:getModName(self.uniqueModID)) .. " DLC")
			end
		end
	end

	self:refreshGenerationOptions()

	-- Setup difficulty buttons
	self._difficultyButtons = {}
	for i, widget in self._panel.binder:forEach( "menuBtn" ) do 
		if not simdefs.STATE_GENERATION_OPTIONS.DIFFICULTY_MENU[i] then
			widget:setVisible(false)
		else
			widget.binder.btn:setText( simdefs.STATE_GENERATION_OPTIONS.DIFFICULTY_MENU[i].name )
			widget.binder.btn.onClick = util.makeDelegate( nil, onClickDiff, self, simdefs.STATE_GENERATION_OPTIONS.DIFFICULTY_MENU[i].diff )
			self._difficultyButtons[ simdefs.STATE_GENERATION_OPTIONS.DIFFICULTY_MENU[i].diff ] = widget
			if util.addToNavigationList then
				util.addToNavigationList(self._navDificulty, 0, widget.binder.btn, true, true)
			end
		end
	end 

	for i, v in ipairs( CUSTOM_OPTIONS[1].strings) do
		self._panel:findWidget( "numRewinds" ):addItem( v )
	end

	updateRewinds( self )
	updateRetries( self )
	self._panel:findWidget("numRewinds").onTextChanged = util.makeDelegate( nil, onChangedRewindOption, self )
	self._panel:findWidget("numRewindsTxt"):setTooltip(STRINGS.UI.DIFF_OPTION_REWINDS_TIP)
	self._panel:findWidget("numRewinds"):setTooltip(STRINGS.UI.DIFF_OPTION_REWINDS_TIP)

	self._panel:findWidget("levelRetriesBtn").onClick = util.makeDelegate( nil, onClickLevelRetries, self) 
	self._panel:findWidget("levelRetriesBtn"):setTooltip(STRINGS.UI.DIFF_OPTION_LEVEL_RETRIES_TIP)
	self._panel:findWidget("levelRetriesBtn"):setClickSound(cdefs.SOUND_HUD_MENU_CLICK)

	self._panel.binder.showOptionsBtn.onClick = util.makeDelegate( nil, onClickShowOptionsBtn, self) 
	self._panel.binder.showOptionsBtn:setTooltip(STRINGS.UI.DIFF_OPTION_SHOW_SETTINGS_TIP)
	self._panel.binder.showOptionsBtn:setClickSound(cdefs.SOUND_HUD_MENU_CLICK)

	self._panel.binder.startBtn.onClick = util.makeDelegate( nil, onClickStartBtn, self) 
	self._panel.binder.startBtn:setClickSound(cdefs.SOUND_HUD_MENU_CONFIRM)
	
	if self.uniqueModID ~= "" and STRINGS.UI.HUD_ENABLED then
		self._panel.binder.dlcBtn.binder.btn.onClick = util.makeDelegate( nil, onToggleUniqueDLC, self) 
		if self._dlcSelections[self.uniqueModID].enabled then
			self._panel.binder.dlcBtn.binder.btn:setText(STRINGS.UI.HUD_ENABLED)
		else
			self._panel.binder.dlcBtn.binder.btn:setText(STRINGS.UI.HUD_DISABLED)
		end
	else
		self._panel.binder.dlcBtn.binder.btn.onClick = util.makeDelegate( nil, onClickSelectDLCBtn, self) 
		self._panel.binder.dlcBtn.binder.btn:setText(STRINGS.UI.SELECT_DLC)
	end
	self._panel.binder.dlcBtn.binder.btn:setClickSound(cdefs.SOUND_HUD_MENU_CLICK)
	
	
	-- Navigation
	if util.addToNavigationList then
		util.addToNavigationList(self._navDificulty, 100, self._panel.binder.dlcBtn.binder.btn, true, true)
	end
	-- !Navigation

	self._panel.binder.cancelBtn.binder.btn.onClick = util.makeDelegate( nil, onClickCancelBtn, self) 
	self._panel.binder.cancelBtn.binder.btn:setHotkey( "pause" )
	if MOAIEnvironment.OS == "PS4" or MOAIEnvironment.OS == "XBO" or MOAIEnvironment.OS == "NX" then
		self._panel.binder.cancelBtn.binder.btn:setText(STRINGS.UI.EXIT_CONSOLE)
	else
		self._panel.binder.cancelBtn.binder.btn:setText(STRINGS.UI.EXIT)
	end
	self._panel.binder.cancelBtn.binder.btn:setClickSound(cdefs.SOUND_HUD_MENU_CANCEL)
	-- Navigation
	if util.addToNavigationList then
		util.addToNavigationList(self._navDificulty, 100, self._panel.binder.cancelBtn.binder.btn, true, true)
	end
	-- !Navigation

	self._panel.binder.pnl_options:setVisible(false)	

	if not MOAIFmodDesigner.isPlaying("theme") then
		MOAIFmodDesigner.playSound("SpySociety/Music/music_title","theme")
	end

	self._screen.binder.agent:createTransition("activate_left")
	
	self._panel.binder.title_txt:spoolText(STRINGS.UI.SCREEN_NAME_GENERATION_OPTIONS)

	-- Setup initial difficulty setting
	if MOAIEnvironment.OS == "NX" then
		-- Tactile input fails with direct custom options
		self:selectDifficulty( simdefs.STATE_GENERATION_OPTIONS.DEFAULT_DIFF )
	else
		self:selectDifficulty( difficulty or simdefs.STATE_GENERATION_OPTIONS.DEFAULT_DIFF ) 
	end
	self:restoreGenerationOptionSelections( difficultyOptions )

	assert(self._diff)
	if self._diff ~= simdefs.CUSTOM_DIFFICULTY and difficultyOptions then
		--rewinds is a special case first-class option if we are coming here from a retry scenario, so seek it out specifically and apply it
		self.updatingDifficulty = true
		
		
		local setting = CUSTOM_OPTIONS[1]
		local item = self.allItems[self.rewindIndex]
		local widget = item.widget.binder.widget
		local val = setting:retrieve( difficultyOptions )		
		if val then
			widget:setValue(val)
		end
		self.updatingDifficulty = nil
	end

	if self._panel.binder:tryBind("standardTitlePs4") then
		if MOAIEnvironment.OS == "PS4" or MOAIEnvironment.OS == "XBO" or MOAIEnvironment.OS == "NX" then
			self._panel.binder.standardTitle:setVisible(false)
			self._panel.binder.standardTitlePs4:setVisible(true)
			self._panel.binder.advancedTitle:setVisible(false)
			self._panel.binder.advancedTitlePs4:setVisible(true)
		else
			self._panel.binder.standardTitle:setVisible(true)
			self._panel.binder.standardTitlePs4:setVisible(false)
			self._panel.binder.advancedTitle:setVisible(true)
			self._panel.binder.advancedTitlePs4:setVisible(false)
		end
	end

	-- Navigation
	local numRewinds = self._panel:findWidget("numRewinds")
	local retriesBtn = self._panel:findWidget("levelRetriesBtn")
	local customlist = self._screen:findWidget("list")

	if numRewinds.resetNavigation then
		numRewinds:resetNavigation()
		retriesBtn:resetNavigation()
		self._panel.binder.showOptionsBtn:resetNavigation()
		self._panel.binder.startBtn:resetNavigation()
		customlist:resetNavigation()
	end

	numRewinds.NavID = 20
	retriesBtn.NavID = 21
	self._panel.binder.showOptionsBtn.NavID = 22
	self._panel.binder.startBtn.NavID = 23
	customlist.NavID = 25

	numRewinds.NavDown = retriesBtn.NavID
	retriesBtn.NavUp = numRewinds.NavID
	retriesBtn.NavDown = self._panel.binder.showOptionsBtn.NavID
	self._panel.binder.showOptionsBtn.NavUp = retriesBtn.NavID

	numRewinds.NavRight = self._panel.binder.startBtn.NavID
	retriesBtn.NavRight = self._panel.binder.startBtn.NavID
	retriesBtn.NavRight = self._panel.binder.startBtn.NavID
	self._panel.binder.showOptionsBtn.NavRight = self._panel.binder.startBtn.NavID

	self._panel.binder.startBtn.NavLeft = self._panel.binder.showOptionsBtn.NavID

	if self._panel.binder.pnl_options:isVisible() then
		local list = self._screen:findWidget("list")
		if list.selectIndex then
			self._panel.binder.showOptionsBtn.NavDown = list.NavID
			self._panel.binder.startBtn.NavDown = list.NavID
			list.NavUp = self._panel.binder.startBtn.NavID
			list:selectIndex(1)
		end
	end
		
	if self._screen.setFocusedWidget and MOAIEnvironment.Use_Gamepad == 1 then
		self._screen:setFocusedWidget( self._navDificulty[1] )
	end

	if self.onInputEvent then
		inputmgr.addListener( self, 1 )
	end

	customlist.onItemClicked = util.makeDelegate( nil, onItemClicked, self )
	customlist.onItemFocused = util.makeDelegate( nil, onItemFocused, self )
	customlist.setTooltipPosition = util.makeDelegate( nil, setTooltipPosition, self )
	-- !Navigation
end

stateGenerationOptions.refreshGenerationOptions = function(self)
	
	local list =  self._screen:findWidget("list")
	list:clearItems()

	self.allItems = {}
	
	self.rewindIndex = nil
	self.retriesIndex = nil
	local showGeneralHeader = false

	if self._dlcSelections then
		local modOrder = {}
		
		for id,info in pairs(self._dlcSelections) do
			if mod_manager.generationOptions[id] and mod_manager.generationOptions[id].enabled then
				table.insert(modOrder,id)
			else
				self._dlcSelections[id] = nil
			end
		end
		
		table.sort(modOrder,function(m1, m2) return mod_manager.generationOptions[m1].name < mod_manager.generationOptions[m2].name end )
	 
		for i,id in ipairs(modOrder) do
			local info = self._dlcSelections[id]
			if info.enabled then
				showGeneralHeader = true
				addGenerationOption(self, { section = util.sformat( STRINGS.UI.DLC_OPTIONS_HEADER, info.name ), icon=info.icon, dlc_id = id }) 

				local mod_opts = mod_manager:getModGenerationOptions(id)
				
				if mod_opts then
					for i, opt_info in ipairs(mod_opts) do
						local opt_setting = util.tcopy(DLC_SETTING)
						
						opt_setting = util.tconcat(opt_setting,opt_info)
						opt_setting.dlc_id = id
						opt_setting.dlc_option = opt_info.option
						if opt_info.values then
							opt_setting.check = false
							opt_setting.apply = opt_info.apply or function( self, options, index )
								options[self.dlc_id].options[self.dlc_option].value = self.values[index]
							end
							opt_setting.retrieve = opt_info.retrieve or function( self, options )
								if self.strings then
									return self.strings[FindIndexOf(options[self.dlc_id].options[self.dlc_option].value, self.values)]
								else
									return tostring(options[self.dlc_id].options[self.dlc_option].value)
								end
							end
						end

						addGenerationOption(self, opt_setting)
					end
				end
			end
		end

	end
	
	if showGeneralHeader then
		addGenerationOption(self, { section = STRINGS.UI.GENERAL_OPTIONS_HEADER, dlc_id = 0 })	
	end
	
	local CUSTOM_OPTIONS = self:findOptions()
	
	-- Setup custom generation options. 
	for i, setting in ipairs( CUSTOM_OPTIONS ) do
		addGenerationOption(self, setting)
	end 
	
	self.allItems = util.tdupe(list:getItems())
	checkForMasks(self)

	assert(self.rewindIndex)
	assert(self.retriesIndex)
end

stateGenerationOptions.restoreGenerationOptionSelections = function( self, difficultyOptions, dlc_options )
	local options = difficultyOptions or simdefs.DIFFICULTY_OPTIONS[ self._diff ] or simdefs.DIFFICULTY_OPTIONS[ simdefs.NORMAL_DIFFICULTY ]
	dlc_options = dlc_options or self._dlcSelections
	self:applyOptions( options )
	self:applyDLCOptions( dlc_options, options )
end

stateGenerationOptions.retrieveOptions = function(self)
	local options = nil

	if self._diff ~= simdefs.CUSTOM_DIFFICULTY then
		options = util.tcopy( simdefs.DIFFICULTY_OPTIONS[ self._diff ] )
	else
		options = util.tcopy( simdefs.DIFFICULTY_OPTIONS[ simdefs.NORMAL_DIFFICULTY ] )
	end
	
	local CUSTOM_OPTIONS = self:findOptions()

	for i, setting in ipairs( CUSTOM_OPTIONS ) do
		local item = self.allItems[setting.list_index]
		-- All option list box items have a child named 'widget' which holds the value.
		local widget = item.widget.binder.widget
		if setting.check then
			setting:apply( options, widget:getValue() )
		else
			setting:apply( options, widget:getIndex() )
		end
		if setting.nestedOptions then
			for i, setting in ipairs( setting.nestedOptions ) do
				local item = self.allItems[setting.list_index]
				-- All option list box items have a child named 'widget' which holds the value.
				if item then
					local widget = item.widget.binder.widget
					if setting.check then
						setting:apply( options, widget:getValue() )
					else
						setting:apply( options, widget:getIndex() )
					end
				end
			end
		end
	end

	options.enabledDLC = self:retrieveDLCOptions( options )
	return options
end

stateGenerationOptions.retrieveDLCOptions = function(self,base_options)

	self.updatingDifficulty = true

	local items = self.allItems
	
	if not base_options.hideOptions then
		base_options.hideOptions = {}
	end

	-- Assign these options to the UI.
	for i, item in ipairs( items ) do
		local setting = item.user_data
		if setting.dlc_option then
			local widget = item.widget.binder.widget
			if setting.check then
				setting:apply( self._dlcSelections, widget:getValue(), base_options )
			else
				setting:apply( self._dlcSelections, widget:getIndex(), base_options )
			end
		end
		if setting.section and item.widget.binder:tryBind("hideBtn") then
			base_options.hideOptions[setting.dlc_id] = item.widget.binder.hideBtn.hide
		end
	end

	self.updatingDifficulty = nil

	return self._dlcSelections
end

stateGenerationOptions.applyOptions = function(self, options )
	self.updatingDifficulty = true
	-- Assign these options to the UI.
	
	local CUSTOM_OPTIONS = self:findOptions()
	
	for i, setting in ipairs( CUSTOM_OPTIONS ) do
		local item = self.allItems[setting.list_index]
		local widget = item.widget.binder.widget
		widget:setValue( setting:retrieve( options ))
		if setting.nestedOptions then
			for i, setting in ipairs( setting.nestedOptions ) do
				local item = self.allItems[setting.list_index]
				if item then
					local widget = item.widget.binder.widget
					widget:setValue( setting:retrieve( options, self._diff ))
				end
			end
		end
	end
	self.updatingDifficulty = nil
end

stateGenerationOptions.applyDLCOptions = function(self, options, base_options )
	self.updatingDifficulty = true

	-- Assign these options to the UI.
	for i, item in ipairs( self.allItems ) do
		local setting = item.user_data
		if setting.dlc_option then
			local widget = item.widget.binder.widget
			widget:setValue( setting:retrieve( options, base_options, self._diff ) )
		end
		if setting.section and base_options.hideOptions and item.widget.binder:tryBind("hideBtn") then
			item.widget.binder.hideBtn.hide = base_options.hideOptions[setting.dlc_id]
		end
	end
	
	checkForMasks(self)

	self.updatingDifficulty = nil
end

stateGenerationOptions.showDifficultyOptions = function(self, show )
	if show and not self._panel.binder.pnl_options:isVisible() then
		MOAIFmodDesigner.playSound(cdefs.SOUND_HUD_MENU_POPUP)
		self._panel.binder.pnl_options:createTransition("activate_above")
		self._panel.binder.pnl_options:setVisible(true)	

		if MOAIEnvironment.Use_Gamepad == 1 then
			local list = self._screen:findWidget("list")
			self._panel.binder.showOptionsBtn.NavDown = list.NavID
			self._panel.binder.startBtn.NavDown = list.NavID
			list.NavUp = self._panel.binder.startBtn.NavID
			list:selectIndex(1)
		end
	elseif not show and self._panel.binder.pnl_options:isVisible() then
		MOAIFmodDesigner.playSound(cdefs.SOUND_HUD_MENU_POPDOWN)
		self._panel.binder.pnl_options:createTransition("deactivate_above",
			function( transition )
				self._panel.binder.pnl_options:setVisible( false )
			end,
			{ easeOut = true } )
			
		-- Disconnect navigation
		if MOAIEnvironment.Use_Gamepad == 1 then
			self._panel.binder.showOptionsBtn.NavDown = -1
			self._panel.binder.startBtn.NavDown = -1
		end
	end
	
	if MOAIEnvironment.OS == "NX" then
	  -- Needed for tactile input
	  local restoreOptions = self:retrieveOptions()
	  self:refreshGenerationOptions()
	  self:restoreGenerationOptionSelections( restoreOptions )
	end

	self._panel.binder.showOptionsBtn:setChecked( show )
end

stateGenerationOptions.selectDifficulty = function(self, diff )
	if diff == self._diff then
		return
	end

	local widget = self._difficultyButtons[ diff ]
	assert( widget, tostring(diff) )
	local difficultyMenu = nil
	for i, menu in ipairs( simdefs.STATE_GENERATION_OPTIONS.DIFFICULTY_MENU ) do
		if menu.diff == diff then
			difficultyMenu = menu
			break
		end
	end

	for i, widgetSub in self._panel.binder:forEach( "menuBtn" ) do
		widgetSub.binder.btn:setColorInactive( unpack(simdefs.STATE_GENERATION_OPTIONS.MENU_UNSELECTED) )
		widgetSub.binder.btn:setTextColorInactive( unpack(simdefs.STATE_GENERATION_OPTIONS.INACTIVE_TXT) )
		if MOAIEnvironment.Use_Gamepad == 1 then
			widgetSub.binder.btn:setStateInactive()
		else
			widgetSub.binder.btn:updateImageState()
		end
	end

	widget.binder.btn:setColorInactive( unpack(simdefs.STATE_GENERATION_OPTIONS.MENU_SELECTED) )
	widget.binder.btn:setTextColorInactive( unpack(simdefs.STATE_GENERATION_OPTIONS.ACTIVE_TXT) )
	widget.binder.btn:updateImageState()

	self._screen:findWidget( "diff_title" ):spoolText( difficultyMenu.short_desc )
	self._screen:findWidget( "diff_txt" ):setText( difficultyMenu.desc )

	self._diff = diff

	if diff == simdefs.CUSTOM_DIFFICULTY then
		self:showDifficultyOptions( true )
	else
		-- Reset the difficulty options to the custom options widgets.
		self:applyOptions( simdefs.DIFFICULTY_OPTIONS[ diff ] )
		self:applyDLCOptions( mod_manager:getModContentDefaults(self._diff), simdefs.DIFFICULTY_OPTIONS[ diff ] )
	end
end

stateGenerationOptions.onUnload = function(self)
	if self.onInputEvent then
		inputmgr.removeListener( self )
	end
	
	self._scroll_text:destroy()
	self._scroll_text = nil
	
	mui.deactivateScreen( self._screen )
	self._screen = nil
end

stateGenerationOptions._CHARACTER_IMAGES = simdefs.STATE_GENERATION_OPTIONS.CHARACTER_IMAGES

return(stateGenerationOptions)