local targeting = include("hud/targeting")
local util = include("client_util")
local mathutil = include("modules/mathutil")
local geo_util = include("geo_util")
local simquery = include("sim/simquery")
-- unfortunately, areaTargetBase is not exposed directly; luckily we can hack class structure
local areaTargetBase = targeting.simpleAreaTarget._base

local function isCellImpassable(cell, sim, unit)
    return not cell or cell.impass > 0 or simquery.isUnitCellFull(sim, unit)
end

local function drawRectIfSeen(game, sim, player, x, y)
    -- if sim:canPlayerSee(player, x, y) then
        local x0, y0 = game:cellToWorld(x + 0.4, y + 0.4)
        local x1, y1 = game:cellToWorld(x - 0.4, y - 0.4)
        MOAIDraw.fillRect(x0, y0, x1, y1)
    -- end
end

local areaTargetDelegatable = class(areaTargetBase)

function areaTargetDelegatable:init(game, range, sim, locTargetFn, locTargetData, ...)
    areaTargetBase.init(self, game, range, sim, ...)
    self.locTargetFn = locTargetFn
    self.locTargetData = locTargetData
end

function areaTargetDelegatable:isValidTargetLoc(x, y)
    local isvalid, reason = self.locTargetFn(self.sim, x, y, self.locTargetData)
    return x and y and isvalid, reason
end

function areaTargetDelegatable:generateTooltip(x, y)
    local isvalid, reason = self:isValidTargetLoc(x, y)
    if not isvalid then
        return "<c:ff0000>" .. (reason or STRINGS.UI.REASON.BAD_TARGET_VIZ) .. "</>"
    end
end

local lineTarget = class(areaTargetBase)

function lineTarget:init(game, sim, unit, ...)
    areaTargetBase.init(self, game, 0, sim, ...)
    -- self.range = 0
    self.unit = unit
    self.maxDist = 1000 -- nil is not permitted
    self.validTargetCellFn = nil
    self.showArc = false
    self.ignoreStaticImpass = true
    self.ignoreDynamicImpass = true

    self._warnClr = {0.5, 0.0, 0.0, 0.5}
end

function lineTarget:endTargeting()
    areaTargetBase.endTargeting(self)
    if self.prop then
        self._game.boardRig:getLayer("ceiling"):removeProp(self.prop)
        self.prop = nil
    end
end

function lineTarget:onInputEvent(event)
    local x, y = self.mx, self.my
    local result = areaTargetBase.onInputEvent(self, event)
    if result then
        return result
    end
    if self.showArc and (self.mx ~= x or self.my ~= y) then
        if self:isValidTargetLoc(self.mx, self.my) then
            if not self.prop then
                self.prop = MOAIProp.new()
            end
            local x0, y0 = self._game:cellToWorld(self.unit:getLocation())
            local x1, y1 = self._game:cellToWorld(self.mx, self.my)

            local msh = geo_util.generateArcMesh(self._game.boardRig, x0, y0, 0, x1, y1, 0, 30, 10)
            self.prop:setDeck(msh)
            self.prop:setLoc(x0, y0)
            self._game.boardRig:getLayer("ceiling"):insertProp(self.prop)
        else
            if self.prop then
                self._game.boardRig:getLayer("ceiling"):removeProp(self.prop)
                self.prop = nil
            end
        end
    end
end

function lineTarget:onDraw()
    local result, reason, blockedCells = self:isValidTargetLoc(self.mx, self.my)

    if self.mx and self.my and (result or blockedCells) then
        local player = self.unit:getPlayerOwner()
        local x0, y0 = self.unit:getLocation()
        local x1, y1 = self.mx, self.my

        MOAIGfxDevice.setPenColor(unpack(self._hiliteClr))

        local cells = self.sim:getQuery().rasterCircle(self.sim, x1, y1, self.range)
        for i = 1, #cells, 2 do
            drawRectIfSeen(self._game, self.sim, player, cells[i], cells[i + 1])
        end

        local cells = mathutil.getCellsBetween(x0, y0, x1, y1)
        for i = 1, #cells, 2 do
            local x2, y2 = cells[i], cells[i + 1]
            if blockedCells then
                local clr = blockedCells[simquery.toCellID(x2, y2)] and self._warnClr or self._hiliteClr
                MOAIGfxDevice.setPenColor(unpack(clr))
            end
            drawRectIfSeen(self._game, self.sim, player, x2, y2)
        end
    end

    return result
end

function lineTarget:isValidTargetLoc(x, y)
    if not x or not y then
        return false
    end

    if not self.sim:getQuery().is360viewClear(self.sim, self.unit, self.maxDist, x, y) then
        return false
    end

    local targetCell = self.sim:getCell(x, y)

    if not (self.ignoreStaticImpass and self.ignoreDynamicImpass) then
        local x0, y0 = self.unit:getLocation()
        local cells = mathutil.getCellsBetween(x0, y0, x, y)
        local blockedCells = {}
        for i = 1, #cells, 2 do
            local x2, y2 = cells[i], cells[i + 1]
            local cell = self.sim:getCell(x2, y2)
            if
                (not self.ignoreStaticImpass and isCellImpassable(cell, self.sim, self.unit)) or
                    (not self.ignoreDynamicImpass and simquery.checkDynamicImpass(self.sim, cell))
             then
                blockedCells[simquery.toCellID(x2, y2)] = true
            end
        end
        if next(blockedCells) then
            return false, STRINGS.HEATSIGNATURE.REASON.BAD_TARGET_IMPASSABLE, blockedCells
        end
    end

    if targetCell and self.sim:getQuery().cellHasTag(self.sim, targetCell, "interruptNonCentral") then
        return false
    end

    if self.validTargetCellFn then
        return self.validTargetCellFn(self.sim, self.unit, targetCell)
    end

    return true
end

function lineTarget:generateTooltip(x, y)
    local b, reason = self:isValidTargetLoc(x, y)
    if not b then
        reason = reason or STRINGS.UI.REASON.BAD_TARGET_RANGE
        return "<c:ff0000>" .. reason .. "</>"
    end
end

-- TODO tooltip with KO strength if applicable?
local lineKnockbackTarget = class(lineTarget)

-- function lineKnockbackTarget:init(...)
--     lineTarget.init(self, ...)
--     self.knockbackMod = 0  -- TODO this would be cool but implies extending the line somehow
--     self.requiresSpace = false
-- end

-- function lineKnockbackTarget:isValidTargetLoc(x, y)
--     local result, reason = lineTarget.isValidTargetLoc(self, x, y)

--     if result and self.requiresSpace then
--         local x0, y0 = self.unit:getLocation()
--         local cells = mathutil.getCellsBetween(x0, y0, x, y)
--         local x2, y2 = 2 * x - cells[#cells - 1], 2 * y - cells[#cells]
--         local cell = self.sim:getCell(x2, y2)
--         if isCellImpassable(cell, self.sim, self.unit) then
--             return false, STRINGS.UI.REASON.BLOCKED, {[simquery.toCellID(x2, y2)] = true}
--         end
--     end

--     return result, reason
-- end

function lineKnockbackTarget:onDraw()
    local result = lineTarget.onDraw(self)

    if result then
        local player = self.unit:getPlayerOwner()
        local x0, y0 = self.unit:getLocation()
        local x1, y1 = self.mx, self.my
        local x2 = 2 * x1 - x0
        local y2 = 2 * y1 - y0
        -- raycast is float, normalise to first blocked cell
        local xr, yr = self.sim:getLOS():raycast(x1, y1, x2, y2)
        local xrLim = mathutil.getCloserSideOfWall(xr, x2)
        local yrLim = mathutil.getCloserSideOfWall(yr, y2)

        MOAIGfxDevice.setPenColor(unpack(self._hiliteClr))

        local cells = mathutil.getCellsBetween(x1, y1, x2, y2)
        for i = 1, #cells, 2 do
            local x, y = cells[i], cells[i + 1]
            local cell = self.sim:getCell(x, y)
            if (xrLim and x == xrLim) or (yrLim and y == yrLim) or isCellImpassable(cell, self.sim, self.unit) then
                MOAIGfxDevice.setPenColor(unpack(self._warnClr))
            end
            drawRectIfSeen(self._game, self.sim, player, x, y)
        end
    end

    return result
end

return {
    areaTargetDelegatable = areaTargetDelegatable,
    lineTarget = lineTarget,
    lineKnockbackTarget = lineKnockbackTarget
}
