local array = include( "modules/array" )
local util = include( "client_util" )
local mui_defs = include( "mui/mui_defs" ) 
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local geo_util = include( "geo_util" )

local targeting = include( "hud/targeting" )

---------------------------------------------------------------------
-- Spell targeting.
local areaSpellTarget = class( targeting.throwTarget )

function areaSpellTarget:init( game, range, sim, userUnit, abilityData )
	targeting.throwTarget.init( self, game, range, sim, userUnit, nil, abilityData.areaPassesThroughWalls )
	self.ability = abilityData
	if abilityData.isAreaTargetUnit then
		self.unitTargetFn = function( cellUnit )
			local hasLos = abilityData.areaPassesThroughWalls or self.offsetX
			if not hasLos then
				local x,y = cellUnit:getLocation()
				local raycastX, raycastY = sim:getLOS():raycast(self.mx, self.mx, x,y)
				if raycastX == x and raycastY == y then
					hasLos = true
				end
			end
			
			if hasLos then
				return abilityData.isAreaTargetUnit( abilityData, sim, userUnit, cellUnit )
			end
		end
	end
end

function areaSpellTarget:isValidTargetLoc( x, y )
	local cell = self.sim:getCell( x, y )
	if not cell then
		return false, STRINGS.UI.REASON.BAD_TARGET_VIZ
	end
	if not self.sim:getPC():getLastKnownCell( self.sim, x, y ) then
		return false, STRINGS.UI.REASON.BAD_TARGET_VIZ
	end
	
	local ux, uy = self.unit:getLocation()
	if self.ability:getThrowRange() then
		local dist = mathutil.dist2d(ux, uy, x, y)
		if dist > self.ability:getThrowRange() then
			return false, STRINGS.UI.REASON.BAD_TARGET_RANGE
		end
	end
	
	if not self.ability.ignoreLOS then
		local raycastX, raycastY = self.sim:getLOS():raycast(ux, uy, x, y)
		if raycastX ~= x or raycastY ~= y then
			return false, STRINGS.WIZARD.UI.BAD_TARGET_LOS
		end
	end

	if cell.impass > 0 and self.ability.staticImpass then
		return false, STRINGS.WIZARD.UI.BAD_TARGET_IMPASS
	end

	if self.sim:getQuery().cellHasTag( self.sim, cell, "interruptNonCentral" ) then		
		return false, STRINGS.WIZARD.UI.BAD_TARGET_RANGE
	end
	
	if self.ability.dynamicImpass and self.sim:getQuery().checkDynamicImpass(self.sim, cell) then
		return false, STRINGS.WIZARD.UI.BAD_TARGET_IMPASS
	end

	return true
end

function areaSpellTarget:generateTooltip( x, y )
	local ok, reason = self:isValidTargetLoc( x, y )
    if reason then
        return "<c:ff0000>" .. reason .. "</>"
    end
end

---------------------------------------------------------------------
-- Square area targeting.
local squareSpellTarget = class( areaSpellTarget )

local function fillSquare( sim, cellx, celly, range, userUnit, player, ignoreLOS )
	local cells = {}
	local offsetX, offsetY = 0, 0
	if (range % 2) == 0 then
		offsetX, offsetY = 0.5, 0.5
	end
	local x0 = math.ceil(cellx + offsetX - 0.5 * range)
	local y0 = math.ceil(celly + offsetY - 0.5 * range)
	local x1 = math.floor(cellx + offsetX + 0.5 * range)
	local y1 = math.floor(celly + offsetY + 0.5 * range)
	
	local losX = cellx + offsetX
	local losY = celly + offsetY
	if userUnit then
		local ux, uy = userUnit:getLocation()
		losX = losX + 0.001 * (ux - (cellx + offsetX))
		losY = losY + 0.001 * (uy - (celly + offsetY))
	end

	for x = x0, x1 do 
		for y = y0, y1 do
			local hasLos = true
			
			if not ignoreLOS then
				local raycastX, raycastY = sim:getLOS():raycast(losX, losY, x, y)
				if raycastX ~= x or raycastY ~= y or (player and not sim:canPlayerSee( player, x, y)) then
					hasLos = false
				end
			end
			
			if hasLos then
				table.insert( cells, x )
				table.insert( cells, y )
			end
		end
	end
	
	return cells
end

function squareSpellTarget:init( game, range, sim, userUnit, abilityData )
	areaSpellTarget.init( self, game, range, sim, userUnit, abilityData )
	if (range % 2) == 0 then
		self.offsetX, self.offsetY = 0.5, 0.5
	else
		self.offsetX, self.offsetY = 0, 0
	end
end

function squareSpellTarget:setTargetCell( cellx, celly )
    self.mx, self.my = cellx, celly

	if self:isValidTargetLoc( cellx, celly ) then
		self.cells = fillSquare( self.sim, cellx, celly, self.range, self.unit, self.unit:getPlayerOwner(), self.ability.ignoreLOS )
    else
        self.cells = nil
    end

     if self.unitTargetFn then
        local count = #self.unitTargets
        if self.cells then
		    for i = 1, #self.cells, 2 do
                local cell = self.sim:getCell( self.cells[i], self.cells[i+1] )
                if cell then
                    for j, cellUnit in ipairs(cell.units) do
                        if self.unitTargetFn( cellUnit ) then
                            -- Hey: does this already exist in the list?
                            local idx = array.find( self.unitTargets, cellUnit:getID() )
                            if idx then
                                table.insert( self.unitTargets, table.remove( self.unitTargets, idx ))
                                count = count - 1
                            else
                                table.insert( self.unitTargets, cellUnit:getID() )
                                self._game.boardRig:getUnitRig( cellUnit:getID() ):getProp():setRenderFilter( cdefs.RENDER_FILTERS["focus_target"] )
                            end
                        end
                    end
                end
            end
		end

        while count > 0 do
            local unitRig = self._game.boardRig:getUnitRig( table.remove( self.unitTargets, 1 ))
            unitRig:refreshRenderFilter()
            count = count - 1
        end
    end
end

function squareSpellTarget:onInputEvent( event )
	if event.eventType == mui_defs.EVENT_MouseDown and event.button == mui_defs.MB_Left then
        self.ux, self.uy = event.wx, event.wy

	elseif event.eventType == mui_defs.EVENT_MouseUp and event.button == mui_defs.MB_Left then
        if mathutil.dist2d( self.ux or event.wx, self.uy or event.wy, event.wx, event.wy ) < 16 then
		    local x, y = self._game:wndToSubCell( event.wx, event.wy )	
		    local cellx, celly = math.floor(x - self.offsetX), math.floor(y - self.offsetY)
		    if self:isValidTargetLoc(cellx, celly) then
			    return {cellx, celly}
		    end
        end
	elseif event.eventType == mui_defs.EVENT_MouseMove then
		local x, y = self._game:wndToSubCell( event.wx, event.wy )
        x, y = math.floor(x - self.offsetX), math.floor(y - self.offsetY)
        if x ~= self.mx or y ~= self.my then
            self.tooltip = self:generateTooltip( x, y )
            self:setTargetCell( x, y )
			
			if self:isValidTargetLoc(self.mx, self.my) then
				if not self.prop then
					self.prop = MOAIProp.new()
				end
				local x0, y0 = self._game:cellToWorld(self.unit:getLocation() )
				local x1, y1 = self._game:cellToWorld(x + self.offsetX, y + self.offsetY)

				local msh = geo_util.generateArcMesh(self._game.boardRig, x0, y0, 0, x1, y1, 0, 30, 10)
				self.prop:setDeck(msh)
				self.prop:setLoc(x0, y0)
				self._game.boardRig:getLayer("ceiling"):insertProp(self.prop)
			else
				if self.prop then
					self._game.boardRig:getLayer("ceiling"):removeProp(self.prop)
					self.prop = nil
				end
			end
        end
    end
end

function squareSpellTarget:isValidTargetLoc( x, y )
	if not x then
		return false, STRINGS.UI.REASON.BAD_TARGET_VIZ
	end
	
	local cell = self.sim:getCell( x, y )
	local wx, wy = x + self.offsetX, y + self.offsetY
	
	local ux, uy = self.unit:getLocation()
	if self.ability:getThrowRange() then
		local dist = mathutil.dist2d(ux, uy, wx, wy)
		if dist > self.ability:getThrowRange() then
			return false, STRINGS.UI.REASON.BAD_TARGET_RANGE
		end
	end
	
	if not self.ability.ignoreLOS then
		local raycastX, raycastY = self.sim:getLOS():raycast(ux, uy, wx, wy)
		if raycastX ~= wx or raycastY ~= wy then
			return false, STRINGS.WIZARD.UI.BAD_TARGET_LOS
		end
	end

	if self.sim:getQuery().cellHasTag( self.sim, cell, "interruptNonCentral" ) then		
		return false, STRINGS.WIZARD.UI.BAD_TARGET_RANGE
	end

	return true
end

squareSpellTarget.onDraw = targeting.simpleAreaTarget.onDraw

---------------------------------------------------------------------
-- Multiple simultaneous targeting types.

local multiTarget = class()

function multiTarget:init( game, sim, components, delayStart )
	self._components = components
	self._delayStart = delayStart
	self._game = game
end

function multiTarget:startTargeting( ... )
	if not self._delayStart or self._game.hud._state == self._game.hud.STATE_ABILITY_TARGET then
		for i, component in ipairs( self._components ) do
			if component.startTargeting then
				component:startTargeting( ... )
			end
		end
	end
end

function multiTarget:endTargeting( ... )
	for i, component in ipairs( self._components ) do
		if component.endTargeting then
			component:endTargeting( ... )
		end
	end
end

function multiTarget:onInputEvent( ... )
	for i, component in ipairs( self._components ) do
		if component.onInputEvent then
			local target = component:onInputEvent( ... )
			if target then
				return target
			end
		end
	end
end

function multiTarget:getTooltip( ... )
	local tt = ""
	for i, component in ipairs( self._components ) do
		if component.getTooltip then
			local addTT = component:getTooltip( ... )
			if addTT then
				tt = tt..addTT
			end
		end
	end
	
	return tt
end

function multiTarget:getDefaultTarget()

end

function multiTarget:hasTargets()
	for i, component in ipairs( self._components ) do
		if component.hasTargets and component:hasTargets() then
			return true
		end
	end
	
	return false
end

function multiTarget:onDraw()
	for i, component in ipairs( self._components ) do
		if component.onDraw then
			component:onDraw()
		end
	end
end

targeting.multiTarget = multiTarget
targeting.areaSpellTarget = areaSpellTarget
targeting.squareSpellTarget = squareSpellTarget
targeting.fillSquare = fillSquare