local util = include( "client_util" )
local mathutil = include( "modules/mathutil" )
local cdefs = include( "client_defs" )
local mui_defs = include( "mui/mui_defs")
local array = include( "modules/array" )
local simquery = include( "sim/simquery" )
local targeting = include( "hud/targeting" )

local customAreaTarget = class()

function customAreaTarget:init( game, sim, params, player )
	self._game = game
	self.sim = sim 
	self._hiliteClr = { 0.5, 0.5, 0.5, 0.5 }
	self.mx = nil
	self.my = nil
	self.params = params
	self.unitTargets = {}
	self.player = player
end

function customAreaTarget:setUnitPredicate( unitTargetFn )
	self.unitTargetFn = unitTargetFn
end

function customAreaTarget:setHiliteColor( clr )
	self._hiliteClr = clr
end

function customAreaTarget:hasTargets()
	return true
end

function customAreaTarget:isValidTargetLoc(x, y)
	if self.player then
		return x and y and self.player:getLastKnownCell( self.sim, x, y ) ~= nil
	else
		return x and y
	end
end

function customAreaTarget:setTargetCell( cellx, celly )
	self.mx, self.my = cellx, celly

	if self:isValidTargetLoc( cellx, celly ) then
		if self.params[1] == "O" then
			self.cells = simquery.rasterCircle( self.sim, cellx, celly, self.params[3], self.params[2] )
		elseif self.params[1] == "+" then
			self.cells = simquery.rasterOrthogonalCross( self.sim, cellx, celly, self.params[2], self.params[3], self.params[4] )
		elseif self.params[1] == "X" then
			self.cells = simquery.rasterDiagonalCross( self.sim, cellx, celly, self.params[2], self.params[3], self.params[4] )
		end
	else
		self.cells = nil
	end

	if self.unitTargetFn then
        	local count = #self.unitTargets
        	if self.cells then
			for i = 1, #self.cells, 2 do
				local cell = self.sim:getCell( self.cells[i], self.cells[i+1] )
				if cell then
					for j, cellUnit in ipairs(cell.units) do
						if self.unitTargetFn( cellUnit ) then
							-- Hey: does this already exist in the list?
							local idx = array.find( self.unitTargets, cellUnit:getID() )
							if idx then
								table.insert( self.unitTargets, table.remove( self.unitTargets, idx ))
								count = count - 1
							else
								table.insert( self.unitTargets, cellUnit:getID() )
								self._game.boardRig:getUnitRig( cellUnit:getID() ):getProp():setRenderFilter( cdefs.RENDER_FILTERS["focus_target"] )
							end
						end
					end
				end
			end
		end

		while count > 0 do
			local unitRig = self._game.boardRig:getUnitRig( table.remove( self.unitTargets, 1 ))
			unitRig:refreshRenderFilter()
			count = count - 1
		end
	end
end

function customAreaTarget:endTargeting()
	self:setTargetCell( nil, nil )
end

function customAreaTarget:onInputEvent( event )
	if event.eventType == mui_defs.EVENT_MouseDown and event.button == mui_defs.MB_Left then
		self.ux, self.uy = event.wx, event.wy
	elseif event.eventType == mui_defs.EVENT_MouseUp and event.button == mui_defs.MB_Left then
		if mathutil.dist2d( self.ux or event.wx, self.uy or event.wy, event.wx, event.wy ) >= 16 then
			return nil
		else
			local x, y = self._game:wndToSubCell( event.wx, event.wy )	
			local cellx, celly = math.floor(x), math.floor(y)
			if self:isValidTargetLoc(cellx, celly) then
				return {cellx, celly}
			else
				return nil
			end
		end
	elseif event.eventType == mui_defs.EVENT_MouseMove then
		local x, y = self._game:wndToSubCell( event.wx, event.wy )
		x, y = math.floor(x), math.floor(y)
		if x ~= self.mx or y ~= self.my then
			self.tooltip = self:generateTooltip( x, y )
			self:setTargetCell( x, y )
		end
	end
end

function customAreaTarget:onDraw()
	MOAIGfxDevice.setPenColor(unpack(self._hiliteClr))
	if self.cells then
		for i = 1, #self.cells, 2 do
			local x, y = self.cells[i], self.cells[i+1]
			local x0, y0 = self._game:cellToWorld( x + 0.4, y + 0.4 )
			local x1, y1 = self._game:cellToWorld( x - 0.4, y - 0.4 )
			MOAIDraw.fillRect( x0, y0, x1, y1 )
		end
        	return true
	end

	return false
end

function customAreaTarget:generateTooltip( x, y )
	if not self:isValidTargetLoc(x, y) then
		return "<c:ff0000>" .. STRINGS.UI.REASON.BAD_TARGET_VIZ .."</>"
	end
end

function customAreaTarget:getTooltip( x, y )
	return self.tooltip
end

function customAreaTarget:getDefaultTarget()
	return nil
end

targeting.customAreaTarget = customAreaTarget