-- local util = include("modules/util")
-- local array = include("modules/array")
local inventory = include("sim/inventory")
local simunit = include("sim/simunit")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local simfactory = include("sim/simfactory")

local simgrenade = {ClassType = "simgrenade_stickystun"}

function simgrenade:onSpawn(sim)
    sim:addTrigger(simdefs.TRG_GRENADES_PRIMED, self)
    local ownerUnit = self:getUnitOwner()
    if ownerUnit then
        self:setPlayerOwner(ownerUnit:getPlayerOwner())
        self:getTraits().throwingUnit = ownerUnit:getID()
    end
end

function simgrenade:throw(throwingUnit, targetCell)
    local sim = self:getSim()
    local player = throwingUnit:getPlayerOwner()
    local x0, y0 = throwingUnit:getLocation()

    assert(player)
    self:setPlayerOwner(player)

    sim:dispatchEvent(simdefs.EV_UNIT_THROWN, {unit = self, x = targetCell.x, y = targetCell.y})

    if x0 ~= targetCell.x or y0 ~= targetCell.y then
        sim:warpUnit(self, targetCell)
    end

    self:getTraits().throwingUnit = throwingUnit:getID()

    self:getTraits().deployed = true
    local cells = self:getExplodeCells()
    sim:dispatchEvent(simdefs.EV_UNIT_ACTIVATE, {unit = self, cells = cells})

    -- Assumption: There will only be one launcher of this type. Else, we'd have to modify the Throw ability.
    for i, weaponUnit in pairs(throwingUnit:getChildren()) do
        if weaponUnit:getTraits().throwUnit == self:getUnitData().id and weaponUnit:getTraits().grenadesPrimed then
            self:explode()
            break
        end
    end

    if self:isValid() then
        for i, cellUnit in ipairs(targetCell.units) do
            if cellUnit:getInventoryCount() < (cellUnit:getTraits().inventoryMaxSize or 0) then
                inventory.pickupItem(sim, cellUnit, self)
                break
            end
        end
    end

    sim:processReactions()
end

function simgrenade:explode()
    local sim = self:getSim()

    if self:getUnitOwner() then
        local x, y = self:getUnitOwner():getLocation()
        self:getUnitOwner():removeChild(self)
        sim:warpUnit(self, sim:getCell(x,y))
    end

    local throwingUnit = sim:getUnit(self:getTraits().throwingUnit)
    if throwingUnit then
        self:setPlayerOwner(throwingUnit:getPlayerOwner())
    end

    local cells = self:getExplodeCells()
    
    sim:dispatchEvent(simdefs.EV_GRENADE_EXPLODE, {unit = self, cells = cells})
    if self.onExplode then
        self:onExplode(cells)
    end

    sim:dispatchEvent(simdefs.EV_UNIT_DEACTIVATE, {unit = self})
    local unitID = self:getID()
    sim:warpUnit(self, nil)
    sim:despawnUnit(self)

    for i, player in ipairs(sim:getPlayers()) do
        player:glimpseUnit(sim, unitID)
    end
end

function simgrenade:onExplode(cells)
    local sim = self:getSim()
    local player = self:getPlayerOwner()

    sim:startTrackerQueue(true)
    sim:startDaemonQueue()
    sim:dispatchEvent(simdefs.EV_KO_GROUP, true)

    if self:getTraits().baseDamage then
        for i, cell in ipairs(cells) do
            for i, cellUnit in ipairs(cell.units) do
                if simquery.isEnemyAgent(player, cellUnit) and not cellUnit:getTraits().isDrone then
                    local damage = self:getTraits().baseDamage
                    damage = cellUnit:processKOresist(damage)
                    cellUnit:setKO(sim, damage)
                end
            end
        end
    end

    sim:dispatchEvent(simdefs.EV_KO_GROUP, false)
    sim:startTrackerQueue(false)
    sim:processDaemonQueue()
end

function simgrenade:getExplodeCells()
    local x0, y0 = self:getLocation()
    local currentCell = self:getSim():getCell(x0, y0)
    local cells = {currentCell}
    if self:getTraits().range then
        local fillCells = simquery.fillCircle(self._sim, x0, y0, self:getTraits().range, 0)
        for i, cell in ipairs(fillCells) do
            if cell ~= currentCell then
                local raycastX, raycastY = self._sim:getLOS():raycast(x0, y0, cell.x, cell.y)
                if raycastX == cell.x and raycastY == cell.y then
                    table.insert(cells, cell)
                end
            end
        end
    end
    return cells
end

function simgrenade:onWarp(sim, oldcell, cell)
    if not oldcell and cell then
        sim:addTrigger(simdefs.TRG_UNIT_WARP, self)
    elseif not cell and oldcell then
        sim:removeTrigger(simdefs.TRG_UNIT_WARP, self)
    end
end

function simgrenade:onTrigger(sim, evType, evData)
    if evType == simdefs.TRG_UNIT_WARP and evData.unit ~= self then
        local x0, y0 = self:getLocation()
        local selfCell = sim:getCell(self:getLocation())
        if evData.to_cell == selfCell or evData.from_cell == selfCell then
            if evData.unit:getTraits().isAgent then
                sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = self})
            end
        end
    end
    if evType == simdefs.TRG_GRENADES_PRIMED then
        if evData.unit:getTraits().throwUnit == self:getUnitData().id and evData.unit:getTraits().grenadesPrimed then
            sim:removeTrigger(simdefs.TRG_GRENADES_PRIMED, self)
            self:explode()
        end
    end
end

local function createGrenade(unitData, sim)
    return simunit.createUnit(unitData, sim, simgrenade)
end

simfactory.register(createGrenade)

return {
    createGrenade = createGrenade
}
