local resources = include( "resources" )
local cdefs = include( "client_defs" )
local array = include( "modules/array" )
local util = include( "modules/util" )
local mathutil = include( "modules/mathutil" )
local simquery = include( "sim/simquery" )
local simdefs = include( "sim/simdefs" )

local function lerp( a, b, t )
	return a+(b-a)*t
end

local function hueToColor( hue )
	local i, f = math.modf( (hue % 1) * 6 )
	
	if i == 1 then
		return {1,f,0,1}
	elseif i == 2 then
		return {1-f,1,0,1}
	elseif i == 3 then
		return {0,1,f,1}
	elseif i == 4 then
		return {0,1-f,1,1}
	elseif i == 5 then
		return {f,0,1,1}
	else
		return {1,0,1-f,1}
	end
end

----------------------------------------------------------------

local hilite_radius = class()

function hilite_radius:init( evData )
    self.x, self.y, self.range = evData.x, evData.y, evData.range
	self.startOffset = 0
	if self.range % 2 == 0 then
		self.x, self.y = self.x + 0.5, self.y + 0.5
		self.startOffset = 1
	end
	
	self.cells = {}
    for i,x,y in util.xypairs(evData.cells) do
        self.cells[ simquery.toCellID( x, y ) ] = {
			x = x,
			y = y,
			manhattanDist = math.max(math.abs(x - self.x), math.abs(y - self.y)),
			trueDist = mathutil.dist2d(x, y, self.x, self.y),
			angle = math.atan2(y - self.y, x - self.x) / (2 * math.pi),
		}
    end
	
    self.currentTime = 0
	self.clr = hueToColor( math.random() )
	self.hueAngle = math.random()
end

function hilite_radius:updateCells( cells )
     -- This is the # of frames between rings. (eg. at self.period, radius=1 has max alpha, at self.period*2, radius=2 has max alpha, etc.)
    -- This is the lifetime for each individual radius. The ring fades in self.period * (radius-1), and lasts RADIUS_LIFETIME frames.
	local TOTAL_PERIOD = self.period * self.radius + self.lifetime
	local EPSILON = 0.3
    
    for radius = 1, self.radius do
        local r, g, b, a = unpack(self.clr or DEFAULT_RING_CLR)
        local minTime = self.period * (radius-1)
        local t = (self.currentTime - minTime) / self.lifetime
		if t > 0 and t <= 1 then
            -- t is in (0, 1].  Remap t to 'fade' to give a specific alpha easing.
            local fade
            if t < EPSILON then
                fade = math.sqrt(t/EPSILON)
			else
				fade = 1 - (t-EPSILON-self.startdecaytime) / (1-EPSILON-self.startdecaytime)
				fade = math.min(1, math.max(0,fade))
            end
            r, g, b, a = r * fade, g * fade, b * fade, a * fade
            local deltas = getCircleCells( radius )
            for i = 1, #deltas, 2 do
                local cellId = simquery.toCellID( self.x + deltas[i], self.y + deltas[i+1] )
                if (cells[ cellId ] == nil or cells[ cellId ][4] < a) and (self.cells == nil or self.cells[ cellId ] ~= nil ) then
                    cells[ cellId ] = { r, g, b, a }
		        end
	        end
        end
    end

    self.currentTime = self.currentTime + 1
    return self.currentTime <= TOTAL_PERIOD
end

function hilite_radius:updateCells( cells )
	local moveOutTime = 12
	local fadeInTime = 12
	local stayTime = 12
	local fadeOutTime = 12
	local colorTransitionSpeed = 0.15
	local hueRotationSpeed = -0.02
	local hueRotationByDist = 0.01
	
	local startOffset = self.startOffset
	local timeToPhase2 = math.max( self.range * 0.5 * moveOutTime + fadeInTime, (self.range * 0.5 - 1) * moveOutTime + fadeInTime + stayTime + fadeOutTime )
	
	if self.currentTime > timeToPhase2 then
		-- Phase 2, move towards center while rotating colours
		for cellId, cell in pairs( self.cells ) do
			local t = 2 * timeToPhase2 - self.currentTime - (cell.manhattanDist - startOffset) * moveOutTime
			local fade = 0
			local cLerp = 1
			local toColor = hueToColor( self.hueAngle + hueRotationByDist * cell.trueDist + self.currentTime * hueRotationSpeed + cell.angle )
			
			if cell.manhattanDist == (self.range - 1) * 0.5 and t > fadeInTime then
				fade = 1
				cLerp = (self.currentTime - timeToPhase2) * colorTransitionSpeed
				if cLerp > 1 then
					cLerp = 1
				end
			elseif t > fadeInTime + stayTime + fadeOutTime then
			
			elseif t > fadeInTime + stayTime then
				fade = (fadeOutTime + fadeInTime + stayTime - t) / fadeOutTime
			elseif t > fadeInTime then
				fade = 1
			elseif t > 0 then
				fade = t / fadeInTime
			end
			
			local r0, g0, b0, a0 = unpack(self.clr)
			local r1, g1, b1, a1 = unpack(toColor)
			
			local r, g, b, a = lerp(r0,r1,cLerp) * fade, lerp(g0,g1,cLerp) * fade, lerp(b0,b1,cLerp) * fade, lerp(a0,a1,cLerp) * fade
			
			if (cells[ cellId ] == nil or cells[ cellId ][4] < a) then
				cells[ cellId ] = { r, g, b, a }
			end
		end
	else
		-- Phase 1, move from center to perimiter
		for cellId, cell in pairs( self.cells ) do
			local t = self.currentTime - (cell.manhattanDist - startOffset) * moveOutTime
			local fade = 0
			
			if cell.manhattanDist == (self.range - 1) * 0.5 and t > fadeInTime then
				fade = 1
			elseif t > fadeInTime + stayTime + fadeOutTime then
			
			elseif t > fadeInTime + stayTime then
				fade = (fadeOutTime + fadeInTime + stayTime - t) / fadeOutTime
			elseif t > fadeInTime then
				fade = 1
			elseif t > 0 then
				fade = t / fadeInTime
			end
			
			local r, g, b, a = unpack(self.clr)
			r, g, b, a = r * fade, g * fade, b * fade, a * fade
			
			if (cells[ cellId ] == nil or cells[ cellId ][4] < a) then
				cells[ cellId ] = { r, g, b, a }
			end
		end
	end
	
    self.currentTime = self.currentTime + 1
	return self.currentTime <= 2 * timeToPhase2
end

local function hiliteEv( event, evType, evData, boardRig, hud, vizThread )
	local hilite = hilite_radius( evData )
	MOAIFmodDesigner.playSound( "SpySociety/Actions/EMP_explo", nil, nil, {hilite.x,hilite.y,0} )
	boardRig:hiliteRadius( hilite )
end

return hiliteEv