local simdefs = include("sim/simdefs")
local simfactory = include( "sim/simfactory" )
local abilitydefs = include( "sim/abilitydefs" )
local itemdefs = include( "sim/unitdefs/itemdefs" )
local util = include("modules/util")

-- This is the ability that spawns spell scrolls
local wizard_spellbook = {
	name = "Spell Book",
	getName = function( self, sim, unit )
		return self.name
	end,
		
	onSpawnAbility = function( self, sim, unit )
		sim:addTrigger( simdefs.TRG_START_TURN, self, unit )
	end,
		
	onDespawnAbility = function( self, sim, unit )
		sim:removeTrigger( simdefs.TRG_START_TURN, self )
	end,

	canUseAbility = function( self, sim, abilityOwner, abilityUser )
		return false -- Passive ability
	end,
	
	onTrigger = function( self, sim, evType, evData, book_item )
		if evType == simdefs.TRG_START_TURN then
			if sim:getTurnCount() == 0 then
				self:addSpellAbilities( sim, book_item )
			end
			sim:removeTrigger( simdefs.TRG_START_TURN, self )
		end
	end,
	
	getRandomSpells = function ( self, sim, book_item, count, maxLevel, minLevel )
		local potentialSpells = {}
		local availableSpellCount = 0
		
		-- Find all spells of an appropriate level
		for name, spell in pairs( abilitydefs._abilities ) do
			if spell.spellLevel and spell.spellLevel <= maxLevel and spell.spellLevel >= (minLevel or 1) then
				potentialSpells[name] = spell
				availableSpellCount = availableSpellCount + 1
			end
		end
		
		-- Remove spells already in the book
		for i, spell in pairs( book_item:getTraits().spellsKnown ) do
			if potentialSpells[spell] then
				potentialSpells[spell] = nil
				availableSpellCount = availableSpellCount - 1
			end
		end
		
		if book_item:getTraits().pastRewards then
			for i, group in ipairs( book_item:getTraits().pastRewards ) do
				for j, spell in ipairs(group) do
					if potentialSpells[spell] then
						potentialSpells[spell] = nil
						availableSpellCount = availableSpellCount - 1
					end
				end
			end
		end
		
		-- Remove spells already owned in scroll form
		local function checkItem( item )
			if type(item) == "table" and item.upgradeParams and item.upgradeParams.traits and item.upgradeParams.traits.storedSpell and potentialSpells[item.upgradeParams.traits.storedSpell] then
				potentialSpells[item.upgradeParams.traits.storedSpell] = nil
				availableSpellCount = availableSpellCount - 1
			end
		end
		
		-- Scrolls stored on the jet
		local agency = sim:getParams().agency
		for i, item in pairs( agency.upgrades or {} ) do
			checkItem( item )
		end
		
		-- Scrolls carried by agents
		for i, agent in pairs( agency.unitDefs ) do
			for j, item in pairs( agent.upgrades or {} ) do
				checkItem( item )
			end
		end
		
		-- Scrolls carried by captured agents
		for i, agent in pairs( agency.unitDefsPotential ) do
			for j, item in pairs( agent.upgrades or {} ) do
				checkItem( item )
			end
		end
		
		local returnSpells = {}
		
		-- Guarantee consistent iteration order
		ALPHABETICAL_PAIRS[potentialSpells] = true
		
		for i = 1, count do
			if availableSpellCount > 0 then
				local spellIdx = sim:nextRand(1, availableSpellCount)
				for name, spell in pairs(potentialSpells) do
					spellIdx = spellIdx - 1
					if spellIdx == 0 then
						table.insert(returnSpells, name)
						potentialSpells[name] = nil
						availableSpellCount = availableSpellCount - 1
						break
					end
				end
			end
		end
		
		ALPHABETICAL_PAIRS[potentialSpells] = nil
		
		return returnSpells
	end,
	
	addSpellAbilities = function( self, sim, book_item )
		local safeUnits = {}
		
		for unitID,unit in pairs( sim._units ) do
			if unit:getTraits().safeUnit and unit:getUnitData().lootTable then
				table.insert( safeUnits, unit )
			end
		end
		
		if #safeUnits > 0 then
			local user = savefiles.getCurrentGame()
			local campaign = user.data.saveSlots[ user.data.currentSaveSlot ]
			
			local safe = safeUnits[ sim:nextRand( 1, #safeUnits ) ]
			local spell = self:getRandomSpells( sim, book_item, 1, campaign.currentDay + 1 )[ 1 ]
			
			if spell then
				local item = itemdefs[ "item_wiz_spellscroll" ]
				local newItem = simfactory.createUnit( item, sim )
				newItem:getTraits().storedSpell = spell
				sim:spawnUnit( newItem )
				safe:addChild( newItem )
			end
		end
	end
}

return wizard_spellbook