local worldgen = include( "sim/worldgen" )
local serverdefs = include( "modules/serverdefs" )
local util = include( "modules/util" )
local weighted_list = include( "modules/weighted_list" )
local unitdefs = include( "sim/unitdefs" )
local array = include( "modules/array" )
local simdefs = include( "sim/simdefs" )
local prefabs = include( "sim/prefabs" )
local mathutil = include( "modules/mathutil" )

local EXTRA_CAMERAS = 
{
	[1] = 0, 
	[2] = 0, 
	[3] = 1, 
	[4] = 3,
	[5] = 3,
	[6] = 3, 
	[7] = 4, 
	[8] = 4, 
	[9] = 4, 
	[10] = 4, 

    [11] = 4, 
    [12] = 4, 
    [13] = 4, 
    [14] = 4,
    [15] = 4,
    [16] = 5, 
    [17] = 5, 
    [18] = 5, 
    [19] = 5, 
    [20] = 5,     
}

local TURRET_NUMBERS = 
{
	[1] = 0, 
	[2] = 1, 
	[3] = 2, 
	[4] = 3,
	[5] = 3,
	[6] = 3, 
	[7] = 4, 
	[8] = 4, 
	[9] = 4, 
	[10] = 4, 

    [11] = 4, 
    [12] = 4, 
    [13] = 4, 
    [14] = 4,
    [15] = 4,
    [16] = 5, 
    [17] = 5, 
    [18] = 5, 
    [19] = 5, 
    [20] = 5,     
} 

local MINE_NUMBERS = 
{
	[1] = 0, 
	[2] = 0, 
	[3] = 3, 
	[4] = 3,
	[5] = 4,
	[6] = 4, 
	[7] = 5, 
	[8] = 5, 
	[9] = 6, 
	[10] = 6, 

    [11] = 7, 
    [12] = 7, 
    [13] = 7, 
    [14] = 7,
    [15] = 7,
    [16] = 8, 
    [17] = 8, 
    [18] = 8, 
    [19] = 8, 
    [20] = 8,     
}

 local function generateLaserGenerators( cxt, candidates )
    local laserCount = 0
    local allCandidates = util.tmerge( {}, candidates, cxt.candidates )
    for i, candidate in ipairs( allCandidates ) do
        if array.find( candidate.prefab.tags, "laser" ) then
            laserCount = laserCount + 1
        end
    end
    prefabs.generatePrefabs( cxt, candidates, "laser_generator", math.ceil( laserCount / 2 ) )
 end

 local function generateMainframes( cxt, candidates )
    if cxt.params.campaignDifficulty == simdefs.NORMAL_DIFFICULTY and cxt.params.missionCount == 0 then
        -- Suppress all mainframes in beginner on the first mission.
        return
    end

    -- This conditional should match the situation where daemons are spawned into the level.
     if cxt.difficulty > 1 then
	    prefabs.generatePrefabs( cxt, candidates, "mainframedb", cxt.rnd:nextInt( 0, 1 ) )
	    prefabs.generatePrefabs( cxt, candidates, "daemondb", 1 )
    else
	    prefabs.generatePrefabs( cxt, candidates, "mainframedb_nodaemon", cxt.rnd:nextInt( 0, 1 ) )
    end
end

local function getCameraNumbers(params, difficulty)

	local cameraNum = 3 + EXTRA_CAMERAS[difficulty]

	
	local alarmList = params.difficultyOptions.alarmTypes
	if params.missionEvents and params.missionEvents.advancedAlarm then
		if params.difficultyOptions.alarmTypes == "EASY" then
			alarmList = "ADVANCED_EASY"
		elseif params.difficultyOptions.alarmTypes == "NORMAL" then
			alarmList = "ADVANCED_NORMAL"
		end
    end

	for i,alarm in ipairs( simdefs.ALARM_TYPES[alarmList] ) do
		if alarm == "cameras" then 
			cameraNum= cameraNum + simdefs.TRACKER_CAMERAS[i]
		end
	end

	return cameraNum
end

 local function generateMiniServers( cxt, candidates )
    if cxt.params.campaignDifficulty == simdefs.NORMAL_DIFFICULTY and cxt.params.missionCount == 0 then
        return
    end

 	local toSpawn = false 

 	if cxt.rnd:nextInt( 0, 2 ) == 2 then 
 		toSpawn = true 
 	else 
 		toSpawn = false 
 	end 

	local miniserverRatio = cxt.params.miniserversSeen / cxt.params.missionCount 

	if miniserverRatio < 0.5 then 
		toSpawn = true 
	end 

	if toSpawn == true then 
		prefabs.generatePrefabs( cxt, candidates, "miniserver", 1 )
	end 
end

 local function generateMainframes( cxt, candidates )
    if cxt.params.campaignDifficulty == simdefs.NORMAL_DIFFICULTY and cxt.params.missionCount == 0 then
        return
    end

     if cxt.difficulty > 1 then
	    prefabs.generatePrefabs( cxt, candidates, "mainframedb", cxt.rnd:nextInt( 0, 1 ) )
	    prefabs.generatePrefabs( cxt, candidates, "daemondb", 1 )
    else
	    prefabs.generatePrefabs( cxt, candidates, "mainframedb_nodaemon", cxt.rnd:nextInt( 0, 1 ) )
    end
end

local function lootFitness( cxt, prefab, tx, ty )
	if prefab.rooms[1] then
		local x, y = prefab.rooms[1].rects[1].x0 + tx, prefab.rooms[1].rects[1].y0 + ty
		local room = cxt:roomContaining( x, y )
		if room then
			return math.pow( room.lootWeight, 2 )
		end
	end
	return 1
end

local function cameraFitness( candidates )
    return function( cxt, prefab, tx, ty )
        -- Fine the coordinates where the camera will spawn (tile coords + match offset + prefab rotation offset)
        local MIN_CAMERA_DISTANCE = 5
        local x0, y0 = prefab.camera.x + tx + prefab.tx, prefab.camera.y + ty + prefab.ty
        -- Count how many cameras 
        local room = cxt:roomContaining( x0, y0 )
        local count = 0
        for i, candidate in ipairs(candidates) do
            if candidate.prefab.filename == prefab.filename then
                local x1, y1 = candidate.prefab.camera.x + candidate.tx + candidate.prefab.tx, candidate.prefab.camera.y + candidate.ty + prefab.ty
                local r2 = cxt:roomContaining( x1, y1 )
                if r2.roomIndex == room.roomIndex then
                    if mathutil.dist2d( x0, y0, x1, y1 ) < MIN_CAMERA_DISTANCE then
                        return 0
                    end
                    count = count + 1
                end
            end
        end
        if count >= 2 then
            return 0
        else
            return 1
        end
    end
end

function worldgen.worlds.ftm:generatePrefabs( )
	local candidates = {}
	
	self.NUM_CAMERAS = getCameraNumbers(self.params, self.difficulty)

    	if isVersion(self.params, "0.17.9") then
        	for i=1,simdefs.FTM_PREFABS_TABLE[self.difficulty] do
            		prefabData = worldgen.FTM.PREFABS[ self.rnd:nextInt(1,#worldgen.FTM.PREFABS)  ]
            		prefabs.generatePrefabs( self, candidates, prefabData.prefab, prefabData.num)        
        	end          
    	else        
        	if self.difficulty >= 2 then
            		prefabs.generatePrefabs( self, candidates, "scanner", 1)    
        	end
    	end

	if self.params.agency.blocker == true then 
		prefabs.generatePrefabs( self, candidates, "inhibitor", 2 )
	end 

	local diff = self.difficulty
	if (self.params.difficultyOptions.W93_AI or -1) > 0 then
		diff = diff - self.params.difficultyOptions.W93_AI + 1
	end

	if (self.params.difficultyOptions.W93_AI or -1) > 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
		prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 4, math.ceil((diff-1)/3) ) )
	end

	prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )
	prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
	prefabs.generatePrefabs( self, candidates, "console", self.NUM_CONSOLES )
	prefabs.generatePrefabs( self, candidates, "null_console", self.NUM_NULL_CONSOLES )
    	generateMainframes( self, candidates )
    	generateMiniServers( self, candidates )
	prefabs.generatePrefabs( self, candidates, "store", 1 )
	prefabs.generatePrefabs( self, candidates, "safe", self.NUM_SAFES, lootFitness )
	prefabs.generatePrefabs( self, candidates, "secret", 2 )
	prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS - EXTRA_CAMERAS[self.difficulty], cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "camera_2", EXTRA_CAMERAS[self.difficulty], cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "decor")

    	generateLaserGenerators( self, candidates )

	return candidates
end

function worldgen.worlds.ko:generatePrefabs( )
	local candidates = {}

	self.NUM_CAMERAS = math.max(getCameraNumbers(self.params,self.difficulty) - self.difficulty,1)

	self.NUM_TURRETS = TURRET_NUMBERS[ self.difficulty ]
	self.NUM_MINES = MINE_NUMBERS[ self.difficulty ]

	prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )

	for i,prefabData in ipairs(worldgen.KO.PREFABS) do 
		if self.difficulty >= prefabData.difficulty then
			prefabs.generatePrefabs( self, candidates, prefabData.prefab, prefabData.num)	
		end
	end

	if self.params.agency.blocker == true then 
		prefabs.generatePrefabs( self, candidates, "inhibitor", 2 )
	end 

	local diff = self.difficulty
	if (self.params.difficultyOptions.W93_AI or -1) > 0 then
		diff = diff - self.params.difficultyOptions.W93_AI + 1
	end

	if (self.params.difficultyOptions.W93_AI or -1) > 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
		prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 4, math.ceil((diff-1)/3) ) )
	end

	prefabs.generatePrefabs( self, candidates, "turret", self.NUM_TURRETS )

	prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
	prefabs.generatePrefabs( self, candidates, "console", self.NUM_CONSOLES )
	prefabs.generatePrefabs( self, candidates, "null_console", self.NUM_NULL_CONSOLES )
	generateMiniServers( self, candidates )
    	generateMainframes( self, candidates )
	prefabs.generatePrefabs( self, candidates, "store", 1 )
	prefabs.generatePrefabs( self, candidates, "shockmine", self.NUM_MINES )
	prefabs.generatePrefabs( self, candidates, "safe", self.NUM_SAFES, lootFitness )
	prefabs.generatePrefabs( self, candidates, "secret", 2 )
	prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS, cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "decor" )

    	generateLaserGenerators( self, candidates )

	return candidates
end

function worldgen.worlds.plastech:generatePrefabs()
	local candidates = {}
	
	for i,prefabData in ipairs(worldgen.PLASTECH.PREFABS) do 
		if self.difficulty >= prefabData.difficulty then
			prefabs.generatePrefabs( self, candidates, prefabData.prefab, prefabData.num)	
		end
	end

	self.NUM_CAMERAS = getCameraNumbers(self.params,self.difficulty)

	prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )
	if self.difficulty >= 3 then
		prefabs.generatePrefabs( self, candidates, "daemon_server", math.max(1,math.floor(self.difficulty/5)))
	end

	local diff = self.difficulty
	if (self.params.difficultyOptions.W93_AI or -1) > 0 then
		diff = diff - self.params.difficultyOptions.W93_AI + 1
	end

	if (self.params.difficultyOptions.W93_AI or -1) > 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
		prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 4, math.ceil((diff-1)/3) ) )
	end

	prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
	prefabs.generatePrefabs( self, candidates, "console", self.NUM_CONSOLES )
	prefabs.generatePrefabs( self, candidates, "null_console", self.NUM_NULL_CONSOLES )
	generateMiniServers( self, candidates )
	generateMainframes( self, candidates )
	prefabs.generatePrefabs( self, candidates, "store", 1 )
	prefabs.generatePrefabs( self, candidates, "safe", self.NUM_SAFES, lootFitness )
	prefabs.generatePrefabs( self, candidates, "secret", 2 )
	prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS, cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "decor" )

	generateLaserGenerators( self, candidates )

	return candidates
end

function worldgen.worlds.sankaku:generatePrefabs(  )
	local candidates = {}

	for i,prefabData in ipairs(worldgen.SANKAKU.PREFABS) do 
		if self.difficulty >= prefabData.difficulty then
			prefabs.generatePrefabs( self, candidates, prefabData.prefab, prefabData.num)	
		end
	end

	if self.params.agency.blocker == true then 
		prefabs.generatePrefabs( self, candidates, "inhibitor", 2 )
	end 

	self.NUM_CAMERAS = getCameraNumbers(self.params,self.difficulty)

	prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )

	prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
	prefabs.generatePrefabs( self, candidates, "console", 6 )

    	if self.difficulty >= 2 then
		prefabs.generatePrefabs( self, candidates, "soundbug", 6 )
	end
    	if self.difficulty >= 3 then
		prefabs.generatePrefabs( self, candidates, "drone_uplink", math.max(1, math.floor(self.difficulty/5)) )
	end

	local diff = self.difficulty
	if (self.params.difficultyOptions.W93_AI or -1) > 0 then
		diff = diff - self.params.difficultyOptions.W93_AI + 1
	end

	if (self.params.difficultyOptions.W93_AI or -1) > 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
		prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 4, math.ceil((diff-1)/3) ) )
	end

    	generateMainframes( self, candidates )
    	generateMiniServers( self, candidates )

	prefabs.generatePrefabs( self, candidates, "store", 1)
	prefabs.generatePrefabs( self, candidates, "safe", self.NUM_SAFES )
	prefabs.generatePrefabs( self, candidates, "secret", 2 )
	prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS, cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "decor" )

    	generateLaserGenerators( self, candidates )
	return candidates
end

if worldgen.worlds.neptune then
	function worldgen.worlds.neptune:generatePrefabs( )
		local candidates = {}
	
		self.NUM_CAMERAS = worldgen.getCameraNumbers(self.params,self.difficulty) - math.ceil(math.min(self.difficulty,4) * 0.5)
	
		prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )
		prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
		prefabs.generatePrefabs( self, candidates, "console", self.NUM_CONSOLES )
		prefabs.generatePrefabs( self, candidates, "null_console", self.NUM_NULL_CONSOLES )
	
		if self.difficulty >= 3 then
			prefabs.generatePrefabs( self, candidates, "frost_spire", math.max(1,math.floor(self.difficulty/5)) )
		end

		local diff = self.difficulty
		if (self.params.difficultyOptions.W93_AI or -1) > 0 then
			diff = diff - self.params.difficultyOptions.W93_AI + 1
		end

		if (self.params.difficultyOptions.W93_AI or -1) >= 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
			prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 4, math.ceil((diff-1)/3) ) )
		end
    
		worldgen.generateMainframes( self, candidates )
    		worldgen.generateMiniServers( self, candidates )
	
		prefabs.generatePrefabs( self, candidates, "store", 1 )
		prefabs.generatePrefabs( self, candidates, "safe", self.NUM_SAFES, worldgen.lootFitness )
		prefabs.generatePrefabs( self, candidates, "secret", 2 )
		prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS, worldgen.cameraFitness( candidates ) )
		prefabs.generatePrefabs( self, candidates, "decor")

    		worldgen.generateLaserGenerators( self, candidates )

		return candidates
	end
end

function worldgen.worlds.omni:generatePrefabs(  )
	local candidates = {}
	local OMNI ={
    		PREFABS_DEFAULT = {},
    		PREFABS={},
	}

	for i,prefabData in ipairs(OMNI.PREFABS) do 
		if self.difficulty >= prefabData.difficulty then
			prefabs.generatePrefabs( self, candidates, prefabData.prefab, prefabData.num)	
		end
	end

	self.NUM_CAMERAS = getCameraNumbers(self.params,self.difficulty)

	prefabs.generatePrefabs( self, candidates, "entry_guard", 2 )

	prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
	prefabs.generatePrefabs( self, candidates, "console", 6 )

	prefabs.generatePrefabs( self, candidates, "store", 1)

	local diff = self.difficulty
	if (self.params.difficultyOptions.W93_AI or -1) > 0 then
		diff = diff - self.params.difficultyOptions.W93_AI + 1
	end

	if (self.params.difficultyOptions.W93_AI or -1) >= 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
		prefabs.generatePrefabs( self, candidates, "W93_AI_Core", math.min( 2, math.ceil((diff-1)/5) ) )
	end

    	generateMainframes( self, candidates )

	prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS - 2, cameraFitness( candidates ) )
	prefabs.generatePrefabs( self, candidates, "decor" )

    	generateLaserGenerators( self, candidates )
	prefabs.generatePrefabs( self, candidates, "drone_uplink", 1 )
	prefabs.generatePrefabs( self, candidates, "daemon_server", 1 )
	prefabs.generatePrefabs( self, candidates, "frost_spire", 1 )
	prefabs.generatePrefabs( self, candidates, "shockmine", 6 )
	prefabs.generatePrefabs( self, candidates, "camera_2", 3, cameraFitness( candidates ) )
	return candidates
end

if worldgen.worlds.omni2 then
	function worldgen.worlds.omni2:generatePrefabs()
		local candidates = {}

		self.NUM_CAMERAS = getCameraNumbers(self.params, self.difficulty)

		prefabs.generatePrefabs( self, candidates, "entry_guard", 3 )
		prefabs.generatePrefabs( self, candidates, "barrier", math.floor( self.NUM_BARRIERS/2 ) )
		prefabs.generatePrefabs( self, candidates, "console", 6 )
		prefabs.generatePrefabs( self, candidates, "store", 1)

		local diff = self.difficulty
		if (self.params.difficultyOptions.W93_AI or -1) > 0 then
			diff = diff - self.params.difficultyOptions.W93_AI + 1
		end

		if (self.params.difficultyOptions.W93_AI or -1) >= 0 and diff >= (self.params.difficultyOptions.W93_AI or -1) and diff > 1 then
			prefabs.generatePrefabs( self, candidates, "W93_AI_Core", 1 )
		end

		generateMainframes( self, candidates )

		prefabs.generatePrefabs( self, candidates, "camera", self.NUM_CAMERAS, worldgen.cameraFitness( candidates ) )
		prefabs.generatePrefabs( self, candidates, "decor" )

		generateLaserGenerators( self, candidates )
		prefabs.generatePrefabs( self, candidates, "shockmine", 3 )
		prefabs.generatePrefabs( self, candidates, "camera_2", 1, cameraFitness( candidates ) )

		return candidates
	end
end