local array = include( "modules/array" )
local util = include( "modules/util" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local abilityutil = include( "sim/abilities/abilityutil" )
local hudFile = include( "hud/hud" )
local unitdefs = include("sim/unitdefs")
local simunit = include("sim/simunit")
local simplayer = include("sim/simplayer")
local pcplayer = include("sim/pcplayer")
local unitghost = include("sim/unitghost")

-- transitions to and from void-view
local function linearEase(var_name)
    local timer = MOAITimer.new()
    timer:setSpan( 0.5 )
    timer:setMode( MOAITimer.NORMAL )
    timer:start()

    return function( uniforms )
        local var_name = var_name --up value optimization
        local timer = timer --up value optimization
        local t = timer:getTime()
        uniforms:setUniformFloat( var_name, 2*t ) -- 2 * (0 .. 0.5)
	end
end
local function undesatEase(var_name)
	local timer = MOAITimer.new()
	timer:setSpan( 0.5 )
	timer:setMode( MOAITimer.NORMAL )
    timer:start()

    return function(uniforms)
    	local var_name = var_name
    	local timer = timer
    	local t = timer:getTime()
    	uniforms:setUniformFloat( var_name, 1-2*t )
    end
end

-- bypass for void units to be undetectable
local couldUnitSeeOld = simquery.couldUnitSee
function simquery.couldUnitSee( sim, unit, targetUnit, ignoreCover, targetCell, ... )
	-- Rifted units are invisible (except to themselves, which is how the PC is able to see it)

	if unit and targetUnit and unit ~= targetUnit and (targetUnit:getTraits().apex_rifted or targetUnit:getTraits().apex_marker) then
		return false
	end
	
	return couldUnitSeeOld( sim, unit, targetUnit, ignoreCover, targetCell, ... )
end

-- hud changes to restrict void view only to current game and only to current agent
local oldCreateHud = hudFile.createHud
function hudFile.createHud(...)
	local hud = oldCreateHud(...)

	local oldOnSelectUnit = hud.onSelectUnit

	function hud.onSelectUnit( self, prevUnit, selectedUnit, ... )
		-- if selecting a rifted unit and last unit (did not exist OR existed and was not rifted), desaturate
		if selectedUnit and selectedUnit:getTraits().apex_rifted and (prevUnit and not prevUnit:getTraits().apex_rifted or not prevUnit) then
			KLEIRenderScene:setDesaturation( linearEase("desat_ease") )
			KLEIRenderScene:setAscii(1)
		-- if selecting a nonrifted unit and last unit was rifted, resaturate
		elseif selectedUnit and prevUnit and prevUnit:getTraits().apex_rifted and not selectedUnit:getTraits().apex_rifted then
			KLEIRenderScene:setDesaturation( undesatEase("desat_ease") )
		end

		return oldOnSelectUnit( self, prevUnit, selectedUnit, ... )
	end

	local game = hud._game

	local oldQuit = game.quitToMainMenu

	function game.quitToMainMenu(self, ...)
		KLEIRenderScene:setDesaturation( undesatEase("desat_ease") )

		return oldQuit(self, ...)
	end

	return hud
end

function pcplayer.updateSeenUnits(self, ...)
	simplayer.updateSeenUnits(self, ...)

	for i, unit in pairs(self:getUnits()) do
		if unit:getTraits().apex_marker then
			local sim = unit:getSim()
			-- while we're here, make sure that there is, in fact, a trap
			local NPC, cell, seen, seer = sim:getNPC(), sim:getCell(unit:getLocation()), false, nil
			for i, unit in pairs(NPC:getUnits()) do
				if sim._los:hasSight( unit, cell.x, cell.y ) then
					seen = true
					seer = unit
					break
				end
			end
			if seen and simquery.couldUnitSee(sim, seer, nil) and (not seer:getTraits().mainframe_item or seer:getTraits().mainframe_status == "active") then
				-- the location is still seen by the NPC
				-- proceed!
				-- make a ghost out of the marker, cause it's not usually seen directly
				self:addSeenUnit( unit )
				local cell = unit:getSim():getCell(unit:getLocation())
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = unit })
			else
				-- the location was seen, but is not anymore (seer moved or got disabled)
				-- clear the extra glimpsed cell
				self:markSeen(sim,cell.x,cell.y)
			end
		end
	end
end

local createGhostOld = unitghost.createUnitGhost
function unitghost.createUnitGhost( unit, ... )
	local result = createGhostOld( unit, ... )

	result.getName = unit.getName

	return result
end

local abilities = {
	APEX_MakeRift = {
		name = STRINGS.APEX.ABILITIES.MAKERIFT.NAME,
		profile_icon = "gui/icons/skills_icons/skills_icon_small/icon-item_speed_small.png",
		iconColor = util.color( 75/255, 0/255, 130/255 ),
		iconColorHover = util.color( 139/255, 0/255, 139/255 ),
		alwaysShow = true,
		getName = function( self, sim, unit )
			return self.name
		end,

		rifted = false,

		onSpawnAbility = function(self, sim, abilityOwner)
			self.abilityOwnerItem = abilityOwner
		end,

		createToolTip = function(self, sim, abilityOwnerItem, abilityUser, targetID )
			local tooltip = nil
			if not self.rifted then
				tooltip = abilityutil.formatToolTip(
					STRINGS.APEX.ABILITIES.MAKERIFT.NAME, 
					util.sformat(STRINGS.APEX.ABILITIES.MAKERIFT.DESC,self.abilityOwnerItem:getTraits().range),
					simdefs.DEFAULT_COST
					)
			else
				tooltip = abilityutil.formatToolTip(
					STRINGS.APEX.ABILITIES.MAKERIFT.UNDO.NAME, 
					STRINGS.APEX.ABILITIES.MAKERIFT.UNDO.DESC,
					simdefs.DEFAULT_COST
					)
			end
			return tooltip
		end,

		canUseAbility =  function( self, sim, abilityOwnerItem, userUnit )
			local abilityOwner = abilityOwnerItem:getUnitOwner()
			if not abilityOwner or not simquery.isAgent(userUnit) or abilityOwner ~= userUnit or not userUnit:getLocation() or not userUnit:canAct() then
				return false
			end

			-- if rifted, can use to return. Otherwise, can only use if charged.
			if not self.rifted and abilityOwnerItem:getTraits().cooldown > 0 then
				return false, STRINGS.APEX.ABILITIES.MAKERIFT.ERRORS.UNCHARGED
			end

			return abilityutil.checkRequirements( abilityOwner, userUnit )
		end,

		executeAbility = function( self, sim, abilityOwnerItem, userUnit )
			local abilityOwner = abilityOwnerItem:getUnitOwner()
			if not abilityOwner then
				return
			end

			if self.rifted then
				-- recast early
				self:onTrigger(sim, "VOIDCEASE")
			else
				-- pay cost: consume charges (for now, don't)
				self.abilityOwnerItem:getTraits().cooldown = self.abilityOwnerItem:getTraits().cooldownMax
				-- enter the void!
				self.abilityOwner = abilityOwner
				self.abilityOwner:getTraits().apex_rifted = true
				self.abilityOwner:getTraits().cloakDistance = abilityOwnerItem:getTraits().range + 1
				self.abilityOwner:getTraits().dashSoundRangeOld = self.abilityOwner:getTraits().dashSoundRange
				self.abilityOwner:getTraits().dashSoundRange = 0
				self.rifted = abilityOwnerItem:getTraits().range
				sim:notifySeers()
				sim:addTrigger( simdefs.TRG_UNIT_WARP, self )
				sim:addTrigger( simdefs.TRG_END_TURN, self )
				KLEIRenderScene:setDesaturation( linearEase("desat_ease") )
			end

			sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = abilityOwner, fx = "emp" } )
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_UNIT_WARP and evData.unit == self.abilityOwner then
				self.rifted = self.rifted - 1
			end
			if evType == "VOIDCEASE" or evType == simdefs.TRG_END_TURN or self.rifted == 0 then
				KLEIRenderScene:setDesaturation( undesatEase("desat_ease") )
				self.rifted = false
				sim:notifySeers()
				self.abilityOwner:getTraits().apex_rifted = nil
				self.abilityOwner:getTraits().cloakDistance = nil
				self.abilityOwner:getTraits().dashSoundRange = self.abilityOwner:getTraits().dashSoundRangeOld
				self.abilityOwner:getTraits().dashSoundRangeOld = nil
				sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = self.abilityOwner, fx = "emp" } )
				self.abilityOwner = nil
				sim:removeTrigger( simdefs.TRG_UNIT_WARP, self )
				sim:removeTrigger( simdefs.TRG_END_TURN, self )
			end
		end,
	},

	APEX_VoidWarn = {
		name = "Voices of the Void",
		profile_icon = "gui/icons/skills_icons/skills_icon_small/icon-item_reflex_small.png",
		alwaysShow = false,
		getName = function( self, sim, unit )
			return self.name
		end,

		reveal_marks = {},
		cleanup_marks = {},

		onSpawnAbility = function(self, sim, abilityOwner)
			self.abilityOwner = abilityOwner
			sim:addTrigger( simdefs.TRG_START_TURN, self )
			sim:addTrigger( simdefs.TRG_END_TURN, self )
		end,

		onDespawnAbility = function(self, sim)
			sim:removeTrigger( simdefs.TRG_LOS_REFRESH, self )
			sim:removeTrigger( simdefs.TRG_UNIT_WARP, self )
		end,

		selectCells = function( self, sim, cells )
			-- given a list of cells on the other sides of doors, reveal ones that are watched
			local watchedCells = {}
			--log:write("[APEXLOG] CELL SCAN BEGINS")
			for _, cell in pairs(cells) do
				--log:write("Received cell ("..tostring(cell.x)..","..tostring(cell.y)..")")
				for i, unit in pairs(sim:getAllUnits()) do
					local t = unit:getTraits()
					if t.isGuard or ((t.mainframe_camera or t.mainframe_turret or t.isDrone) and t.mainframe_status == "active") then

						--[[if t.isGuard and unit._seenUnits then
							for i, u in pairs(unit._seenUnits) do
								log:write("- "..unit:getName().."("..unit:getID().."): "..u:getName().."("..u:getID()..")")
							end
						end]]

						--log:write("- "..unit:getName().."("..unit:getID().."): "..tostring(simquery.couldUnitSee(sim, unit, nil)).." / "..tostring(sim._los:hasSight( unit, cell.x, cell.y )).." / "..tostring(sim._los:hasPeripheral( unit, cell.x, cell.y )))
						if simquery.couldUnitSee(sim, unit, nil) and (sim._los:hasSight( unit, cell.x, cell.y ) or sim._los:hasPeripheral( unit, cell.x, cell.y )) and unit:getPlayerOwner() ~= sim:getPC() then
							table.insert(watchedCells, cell)
						end
					end
				end
			end
			return watchedCells
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_END_TURN then
				for i = #self.cleanup_marks, 1, -1 do
					local mark = self.cleanup_marks[i]

					--[[for j, unit in pairs(sim:getAllUnits()) do
						if unit and unit._seenUnits then
							for k = #unit._seenUnits, 1, -1 do
								local seen = unit._seenUnits[k]
								if seen == mark then
									table.remove(unit._seenUnits, k)
									break
								end
							end
						end
					end]]
					sim:warpUnit(mark)
					sim:despawnUnit(mark)
					table.remove(self.cleanup_marks, i)
				end
			end

			if evType == simdefs.TRG_START_TURN then
				sim:removeTrigger( simdefs.TRG_START_TURN, self )
				sim:addTrigger( simdefs.TRG_LOS_REFRESH, self )
				sim:addTrigger( simdefs.TRG_UNIT_WARP, self )
				return
				-- first, run this on turn start
				-- then on warp of units
			end
			if self.updating then
				return
			end

			self.updating = true

			local unit = self.abilityOwner

			-- we do not clear warnings each turn indiscriminately, because they are randomly named and I want the name to stay the same
			-- edit: fixed now to use cell-based pseudorandomness. Still not clearing because supposedly less intensive

			-- note down all previous warning locations
			local last_locs = {}
			for _, mark in pairs(self.reveal_marks) do
				table.insert(last_locs, sim:getCell(mark:getLocation()))
			end

			-- PART I: find suitable locations to place warnings
			local pre_cells = {}
			if not unit:getLocation() then return end
			for _, cell in pairs(sim._los:calculateUnitLOS(sim:getCell(unit:getLocation()), unit)) do
				if simquery.couldUnitSeeCell(sim, unit, cell) then
					-- if cell is seen by Wraith, check its walls
					for _, exit in pairs(cell.exits) do
						if exit.door and exit.closed and not sim:canPlayerSee(unit:getPlayerOwner(), exit.cell.x, exit.cell.y) then
							table.insert(pre_cells, exit.cell)	
						end
					end
				end
			end

			-- PART II: place warnings on all locations where they weren't already
			for _, cell in pairs(self:selectCells(sim, pre_cells)) do
				local preexisting = false
				for i, last_cell in pairs(last_locs) do
					if cell == last_cell then
						preexisting = true
						table.remove(last_locs,i)
						break
					end
				end
				if not preexisting then
					local markerdef = unitdefs.lookupTemplate( "APEX_Marker" )
					local marker = simunit.createUnit( markerdef, sim )
					local namegen = marker:getUnitData().namegen
					-- pick out a pseudorandom number from the namegen table
					-- we cant rly use sim:nextRand or else simhistory blows up
					-- so instead use existing data - the cell the marker spawns on will have pseudorandom coords
					-- (since it depends on procgen, unit vision, etc etc)
					local pseudorand = ((cell.x*10+cell.y) % #namegen) + 1
					marker.getName = function( self )
						return namegen[pseudorand]
					end
					local markAbility = self
					sim:spawnUnit( marker )
					local player = self.abilityOwner:getPlayerOwner()
					marker:setPlayerOwner( player )
					sim:warpUnit( marker, cell )
					-- if cell is unseen, glimpse it
					if not sim:canPlayerSee(player, cell.x, cell.y) then
						player:glimpseCell(sim, cell)
					end
					player:addSeenUnit(marker)
					marker:createTab(marker:getName(),"")
					--[[log:write("[APEXLOG] Displaying seen units by PC:")
					for i, unit in pairs(marker:getPlayerOwner()._seenUnits) do
						log:write("--- "..unit:getName().." ["..unit:getID().."]")
					end]]
					table.insert(self.reveal_marks, marker)
				end
			end

			-- PART III: clean up previous warnings that should no longer be there
			for _, cell in pairs(last_locs) do
				for i = #self.reveal_marks, 1, -1 do
					local mark = self.reveal_marks[i]
					if cell and cell == sim:getCell(mark:getLocation()) then
						table.remove(self.reveal_marks, i)
						mark:destroyTab()
						-- hide the thing man just TODO it
						table.insert(self.cleanup_marks, mark)
					end
				end
			end

			self.updating = nil
		end,
	},
}

return abilities